/* Prototype of an assembler syntax for expressing bytecode.
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

%{
#include <stdio.h>
%}

SPACE [\ \t\f]
DIG [0-9]
INT (\~?{DIG}+)
FMT [\ \t\n\f]
LETTER [A-Za-z]
ALPHA [A-Za-z0-9\'\_]
SYMB [\!\%\&\$\#\+\-\/\:\<\=\>\?\@\\\~\'\^\|\*\`]
ALPHA_ID ({LETTER}{ALPHA}*)
SYMB_ID ({SYMB}+)
ID ({SYMB_ID}|{ALPHA_ID})
PREFIX ({ID}{FMT}*\.{FMT}*)

%%

box { return kw_box; }
end { return kw_end; }
fun { return kw_fun; }
goto { return kw_goto; }
int { return kw_int; }
nil { return kw_nil; }
real { return kw_real; }
unbox { return kw_unbox; }
\, { return kw_comma; }
\; { return kw_semicolon; }
\( { return kw_open_paren; }
\) { return kw_close_paren; }
\= { return kw_equal; }
\(\*	{ skipComment(); }
{INT}	{ return integer_constant; }
{INT}E{INT} { return real_constant; }
{INT}\.{DIG}+   { return real_constant; }
{INT}\.{DIG}+E{INT} { return real_constant; }
\+inf { return kw_pinf; }
\-inf { return kw_minf; }
NAN\(0\)	{ return kw_nan0; }
NAN\(1\)	{ return kw_nan1; }
NAN\(2\)	{ return kw_nan2; }
NAN\(3\)	{ return kw_nan3; }
NAN\(4\)	{ return kw_nan4; }
NAN\(5\)	{ return kw_nan5; }
NAN\(6\)	{ return kw_nan6; }
NAN\(7\)	{ return kw_nan7; }
\~NAN\(1\)  { return kw_mnan1; }
\~NAN\(2\)  { return kw_mnan2; }
\~NAN\(3\)  { return kw_mnan3; }
\~NAN\(4\)  { return kw_mnan4; }
\~NAN\(5\)  { return kw_mnan5; }
\~NAN\(6\)  { return kw_mnan6; }
\~NAN\(7\)  { return kw_mnan7; }
{INT}(\.{DIG}+)?(E{INT})?{SPACE}*\:{SPACE}*{INT}(\.{DIG}+)?(E{INT})?	{
	return complex_constant; }
\"([^\"\\]|(\\([\"\\ntvf]|(\^[\@A-Z\[\\\]\^\_])|({DIG}{DIG}{DIG})|({FMT}*\\))))*\"  {
	return string_constant; }
{PREFIX}*{ALPHA_ID}  { return identifier; }
{PREFIX}*{SYMB_ID}   { return identifier; }
{SPACE}+	{}
\n+	{}
.	{ return yytext[0]; }

%%

PUBLIC void yyinit()
{
  YY_INIT
}

enum commentState { SKIPPING, AFTER_OPEN, AFTER_STAR };

PRIVATE void skipComment()
{
  int c;
  int level,state;

  level = 1;
  state = SKIPPING;
  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  while (c=input(), c!=-1)
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	{
	  hputc(c,coreTraceFile);
	}
#endif
      if (c=='\n')
	{
	  yylloc.last_line++;
	  yylloc.last_column = 0;
	}
      else
	yylloc.last_column++;
      switch (state)
	{
	case SKIPPING:
	  switch (c)
	    {
	    case '(':
	      state = AFTER_OPEN;
	      break;
	    case '*':
	      state = AFTER_STAR;
	      break;
	    default:
	      break;
	    }
	  break;
	case AFTER_OPEN:
	  if (c=='*')
	    level++;
	  state = SKIPPING;
	  break;
	case AFTER_STAR:
	  if (c==')')
	    if (--level==0)
	      return;
	  if (c!='*')
	    state = SKIPPING;
	  break;
	}
    }
  yy_error(yylloc.first_line,yylloc.first_column,
	   yylloc.last_line,yylloc.last_column,
	   message(132));
}
