/* Bytecode compiler (currently not used).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "translate.h"
#include "optimize.h"
#include "bytecode.h"
#include "interp.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "evlproto.h"
#include "errproto.h"
#include "typproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "rbcode.h"
#ifdef HAS_PROTOTYPES
#include "sbcode.h"
#endif

#define __SEG__ BYTECODEGEN

EXTERN int errno;

PRIVATE void varNamesSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  varNames *vn;

  vn = (varNames *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE: /* save */
      {
	COUNTER i;
	CHAR c;

	for (i=0; i<V_COUNT; i++)
	  {
	    outInt(vn->maxn[i]);
	    if (errno) return;
	  }
	outInt(vn->maxbconst);
	if (errno) return;
	outInt(vn->maxuconst);
	if (errno) return;
	c = vn->has_calls;
	outChars(&c,1);
	if (errno) return;
	c = vn->has_jumps;
	outChars(&c,1);
	if (errno) return;
      }
      /*fallthrough*/
    case X_MARK: /* mark */
    case X_MARKDATA:
      (*doext->action)(vn->names);
      (*doext->action)(vn->consts);
      (*doext->action)(vn->bconsts);
      (*doext->action)(vn->uconsts);
      (*doext->action)(vn->zapx);
      (*doext->action)(vn->zapm);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr varNamesLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  varNames *vn;
  ExprPtr e;
  COUNTER i;
  CHAR c;
  ENTER;

  vn = (varNames *)MALLOC(sizeof(varNames));
  for (i=0; i<V_COUNT; i++)
    {
      vn->maxn[i] = inInt();
      if (errno)
	{
err:
	  MARKFREE(vn);
	  return NIL;
	}
    }
  vn->maxbconst = inInt();
  if (errno) goto err;
  vn->maxuconst = inInt();
  if (errno) goto err;
  inChars(&c,1);
  if (errno) goto err;
  vn->has_calls = c;
  inChars(&c,1);
  if (errno) goto err;
  vn->has_jumps = c;
  vn->names = (*load)(sharedref);
  if (errno) goto err;
  vn->consts = (*load)(sharedref);
  if (errno) goto err;
  vn->bconsts = (*load)(sharedref);
  if (errno) goto err;
  vn->uconsts = (*load)(sharedref);
  if (errno) goto err;
  vn->zapx = (*load)(sharedref);
  if (errno) goto err;
  vn->zapm = (*load)(sharedref);
  if (errno) goto err;
  e = (shared?Kexternal:MakeExternal)(desc,(char *)vn);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc vnInfoDesc = {
  (CHAR *)"variable-name-collection",
  'v','n','#',' ', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  varNamesSweep,
  varNamesLoad
  };

PRIVATE ExprPtr varNamesNew()
{
  varNames *vn;
  ExprPtr e;
  COUNTER i;

  vn = (varNames *)MALLOC(sizeof(varNames));
  vn->names = NewMap();
  for (i=0; i<V_COUNT; i++)
    {
      vn->maxn[i] = 0;
    }
  vn->maxbconst = 0;
  vn->consts = NewMap();
  vn->bconsts = NewMap();
  vn->maxuconst = 0;
  vn->uconsts = NewMap();
  vn->has_calls = false;
  vn->has_jumps = false;
  vn->zapx = NewMap();
  vn->zapm = NewMap();
  e = MakeExternal(&vnInfoDesc,(char *)vn);
  return KEEP(e);
}

PRIVATE void varFreeListSweep(extdata,doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  varFreeList *vf;
  COUNTER i;

  vf = (varFreeList *)extdata;
  switch (doext->xtag)
    {
    case X_SAVE:
      {
	COUNTER i;

	for (i=0; i<V_COUNT; i++)
	  {
	    outInt(vf->n[i]);
	    if (errno) return;
	  }
      }
      /*fallthrough*/
    case X_MARK:
    case X_MARKDATA:
      (*doext->action)(vf->pendinguses);
      for (i=0; i<V_COUNT; i++)
	(*doext->action)(vf->free[i]);
      break;
    case X_FREE:
      MARKFREE(extdata); /* free the C structure */
      break;
    }
}

PRIVATE ExprPtr varFreeListLoad(desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  varFreeList *vf;
  ExprPtr e;
  COUNTER i;
  ENTER;

  vf = (varFreeList *)MALLOC(sizeof(varFreeList));
  for (i=0; i<V_COUNT; i++)
    {
      vf->n[i] = inInt();
      if (errno)
	{
err:
	  MARKFREE(vf);
	  return NIL;
	}
    }
  vf->pendinguses = (*load)(sharedref);
  if (errno) goto err;
  for (i=0; i<V_COUNT; i++)
    {
      vf->free[i] = (*load)(sharedref);
      if (errno) goto err;
    }
  e = (shared?Kexternal:MakeExternal)(desc,(char *)vf);
  KEEPONLY(e);
  return e;
}

PUBLIC extDesc vfInfoDesc = {
  (CHAR *)"variable-free-list",
  'v','f','#',' ', NULL,
  NULL, /* not shared */
  NULL, /* not shared */
  NULL, /* print as default */
  varFreeListSweep,
  varFreeListLoad
  };

PRIVATE ExprPtr varFreeListNew()
{
  varFreeList *vf;
  ExprPtr e;
  COUNTER i;

  vf = (varFreeList *)MALLOC(sizeof(varFreeList));
  vf->pendinguses = NewMap();
  for (i=0; i<V_COUNT; i++)
    {
      vf->free[i] = NewMap();
      vf->n[i] = 0;
    }
  e = MakeExternal(&vfInfoDesc,(char *)vf);
  return KEEP(e);
}

PRIVATE ExprPtr varFreeListCopy(vf)
IN CONST ExprPtr vf;
{
  ExprPtr vfn;
  varFreeList *v,*vv;
  COUNTER i;

  vfn = varFreeListNew();
  v = VF(vf);
  vv = VF(vfn);
  SETQ(vv->pendinguses,v->pendinguses);
  for (i=0; i<V_COUNT; i++)
    {
      SETQ(vv->free[i],v->free[i]);
      vv->n[i] = v->n[i];
    }
  return vfn;
}

PRIVATE void varFreeListMerge(vf,vff)
INOUT ExprPtr vf;
IN CONST ExprPtr vff;
{
  COUNTER i;
  COUNTER n,nn;
  ExprPtr var,*loc;

  setMapDomRestrTo(&VF(vf)->pendinguses,VF(vff)->pendinguses);
  HITBARRIER(vf);
  for (i=0; i<V_COUNT; i++)
    { /* take union of free lists and maximum of counters n[] */
      loc = &VF(vf)->free[i];
      setMapMerge(loc,VF(vff)->free[i]);
      HITBARRIER(vf);
      n = VF(vf)->n[i];
      nn = VF(vff)->n[i];
      VF(vf)->n[i] = (n>nn)?n:nn;
    }
}

PRIVATE COUNTER storageClass(storage,def)
IN COUNTER storage;
IN CONST ExprPtr def;
{
  if (storage & SCF_BOXED)
    return V_EXPRPTR;
  if (storage & SCF_BOOL)
    return V_BOOL;
  if (storage & SCF_INT)
    return V_INT;
  if (storage & SCF_STRING)
    return V_STRING;
  if (storage & SCF_REAL)
    return V_REAL;
  if (storage & SCF_CODE)
    return V_CODE;
  if (storage & SCF_LOCAL) /* local data: virtual couples, as returned by mapget,
			      are composite and unboxed;
			      but so are local refs, iterator variables, and so on. */
    {
      if (def->kind!=(CHAR)E_CONS)
	kinderror(message(738),def->kind);
      def = CAR(def);
      if (def->kind!=(CHAR)E_COUNTED)
	kinderror(message(738),def->kind);
      switch (COUNTEDN(def))
	{
	case Z_MAPGET:
	  return V_MAPGET;
	case Z_START_LISTSWEEP:
	  return V_EXPRPTR;
	case Z_START_MAPSWEEP:
	  return V_MAPSWEEP;
	case Z_START_SUBMAPSWEEP:
	  return V_EXPRPTR;
	case Z_BOX_LOCAL_REF:
	  return V_EXPRPTR;
	default:
	  kinderror(message(738),COUNTEDN(def));
	  /*NOTREACHED*/
	  break;
	}
    }
  return V_VOID; /* void */
}

PRIVATE ExprPtr makeVarName(x,deco,vn,vf)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
{
  ppInfo *pp;
  ExprPtr e;
  COUNTER stor;
  varFreeList *vfn;
  INT n;

  vfn = VF(vf);
  pp = getPpInfo(x,deco);
  if (pp->uses==NewMap())
    return NIL;
  stor = storageClass(pp->storage,pp->def);
  e = ChooseInDom(vfn->free[stor]);
  if (e!=NIL)
    {
      vfn->free[stor] = MapRemove(e,vfn->free[stor]);
      return e;
    }
  n = ++vfn->n[stor];
  if (VN(vn)->maxn[stor]<n)
    VN(vn)->maxn[stor] = n;
  e = Kvarref(n,stor);
  return KEEP(e);
}

PRIVATE void addVarName(x,name,vn)
IN CONST ExprPtr x;
IN CONST ExprPtr name;
INOUT ExprPtr vn;
{
  ExprPtr e;

  e = MapPut(x,name,VN(vn)->names);
  SETQ(VN(vn)->names,e);
  HITBARRIER(vn);
}

ARGSWEEPFUNC(do_updatependinguses)
{
  ExprPtr vn,vf,counter,uses,name,deco;
  COUNTER stor;
  Boolean found;

  vn = MAPDATA(struct upu *)->vn;
  vf = MAPDATA(struct upu *)->vf;
  deco = MAPDATA(struct upu *)->deco;
  counter = MapGet(x,VF(vf)->pendinguses,&found);
  if (found && --IVAL(counter)==0)
    {
      uses = MapRemove(x,VF(vf)->pendinguses);
      SETQ(VF(vf)->pendinguses,uses);
      HITBARRIER(vf);
      name = MapGetSure(x,VN(vn)->names);
      stor = VARLEVEL(name); /* storageClass(getPpInfo(x,deco)->storage,x); */
      uses = SetAdd(name,VF(vf)->free[stor]);
      SETQ(VF(vf)->free[stor],uses);
      HITBARRIER(vf);
    }
  return false;
}

PRIVATE updatePendingUses(e,deco,vn,vf)
IN CONST ExprPtr e;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
{
  struct upu u;

  u.vn = vn;
  u.vf = vf;
  u.deco = deco;
  if (e->kind==(CHAR)E_COUNTED)
    (void) argSweep(COUNTEDVAL(e),do_updatependinguses,&u);
  else (void) do_updatependinguses(e,&u);
}

PRIVATE Boolean isConstant P((IN CONST ExprPtr x,IN CONST ExprPtr deco,
			      OUT Boolean *compp,OUT ExprPtr *valp,
			      OUT COUNTER *storagep));

MAPSWEEPFUNC(do_isconstant)
{
  return x==NIL || !isConstant(x,MAPDATA(ExprPtr),NULL,NULL,NULL);
}

PRIVATE Boolean isConstant(x,deco,compp,valp,storagep)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
OUT Boolean *compp;
OUT ExprPtr *valp;
OUT COUNTER *storagep;
{
  ppInfo *pp;
  Boolean constp;

  pp = getPpInfo(x,deco);
  constp = ((pp->storage & SCF_NOINIT)==0 && pp->valset->kind==(CHAR)E_MAPLET);
  if (constp)
    {
      if (valp!=NULL)
	*valp = MAPLETX(pp->valset);
      if (storagep!=NULL)
	*storagep = storageClass(pp->storage,x);
      if (compp!=NULL)
	*compp = MapSweep(pp->uses,do_isconstant,deco);
    }
  return constp;
}

PRIVATE void addConstant(x,val,vn,storageclass)
IN CONST ExprPtr x,val;
INOUT ExprPtr vn;
IN COUNTER storageclass;
{
  Boolean found;
  ExprPtr idx;
  COUNTER offset,newoffset;
  ENTER;

  setMapPut(x,val,&VN(vn)->consts);
  HITBARRIER(vn);
  switch (storageclass)
    {
    case V_EXPRPTR:
    case V_MAPGET:
    case V_MAPSWEEP:
      (void) MapGet(x,VN(vn)->bconsts,&found);
      if (!found)
	{
	  setMapPut(x,KINT(VN(vn)->maxbconst++),&VN(vn)->bconsts);
	  HITBARRIER(vn);
	}
      break;
    case V_BOOL:
      offset = VN(vn)->maxuconst;
      newoffset = offset+1;
uconst:
      (void) MapGet(x,VN(vn)->uconsts,&found);
      if (!found)
	{
	  idx = Kvarref(offset,storageclass);
	  (void) KEEP(idx);
	  setMapPut(x,idx,&VN(vn)->uconsts);
	  HITBARRIER(vn);
	  VN(vn)->maxuconst = newoffset;
	}
      break;
    case V_INT:
      offset = VN(vn)->maxuconst;
      offset = ((offset+sizeof(INT)-1) / sizeof(INT))*sizeof(INT);
      newoffset = offset + sizeof(INT);
      goto uconst;
    case V_STRING:
      offset = VN(vn)->maxuconst;
      newoffset = offset+SLEN(val);
      goto uconst;
    case V_REAL:
      offset = VN(vn)->maxuconst;
      offset = ((offset+ALIGNMENT-1)/ALIGNMENT)*ALIGNMENT;
      newoffset = offset + sizeof(EXTENDED);
      goto uconst;
    case V_CODE:
      offset = VN(vn)->maxuconst;
      offset = ((offset+sizeof(char *)-1) / sizeof(char *))*sizeof(char *);
      newoffset = offset + sizeof(char *);
      goto uconst;
    }
  QUIT;
}

PRIVATE void createPendingUses(x,deco,vf)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vf;
{
  ExprPtr uses,card;
  ppInfo *pp;
  INT n;

  pp = getPpInfo(x,deco);
  uses = pp->uses;
  n = MapCard(uses);
  MAKEINT(n,card);
  (void) KEEP(card);
  uses = MapPut(x,card,VF(vf)->pendinguses);
  SETQ(VF(vf)->pendinguses,uses);
  HITBARRIER(vf);
}

PRIVATE void addZap(x,vn,vf,oldx,oldm)
IN CONST ExprPtr x;
INOUT ExprPtr vn,vf;
IN CONST MapPtr oldx,oldm;
{
  ExprPtr diff;
  ENTER;

  diff = SetDiff(VF(vf)->free[V_EXPRPTR],oldx);
  if (diff!=NewMap())
    {
      setMapPut(x,diff,&VN(vn)->zapx);
      HITBARRIER(vn);
    }
  diff = SetDiff(VF(vf)->free[V_MAPGET],oldm);
  if (diff!=NewMap())
    {
      setMapPut(x,diff,&VN(vn)->zapm);
      HITBARRIER(vn);
    }
  QUIT;
}

PRIVATE MapPtr getVarNames1 P((IN CONST ExprPtr x,IN CONST ExprPtr deco,
			       INOUT ExprPtr vn,INOUT ExprPtr vf,
			       IN CONST MapPtr frontier,IN CONST ExprPtr last));

DEF_PRIVATE_ACTION(getvarnames)
{
  struct getvarnames *data;

  data = (struct getvarnames *)env;
  return getVarNames1(data->x,data->deco,data->vn,data->vf,data->frontier,data->last);
}

PRIVATE MapPtr getVarNames1k(x,deco,vn,vf,frontier,last)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
IN CONST MapPtr frontier;
IN CONST ExprPtr last;
{
  struct getvarnames *data = (struct getvarnames *)alloctmp(sizeof(struct getvarnames));

  data->x = x;
  data->deco = deco;
  data->vn = vn;
  data->vf = vf;
  data->frontier = frontier;
  data->last = last;
  return CallCC(ACTION_NAME(getvarnames),(ExprPtr)data,true);
}

PRIVATE MapPtr getVarNames1(x,deco,vn,vf,frontier,last)
IN CONST ExprPtr x;
IN CONST ExprPtr deco;
INOUT ExprPtr vn,vf;
IN CONST MapPtr frontier;
IN CONST ExprPtr last;
{
  ExprPtr oldExprNames,oldMapGetNames;
  Boolean found;
  ENTER;

  if (STKFULL(found))
    return getVarNames1k(x,deco,vn,vf,frontier,last);
again:
  KEEPONLY(frontier);
  (void) MapGet(x,VN(vn)->names,&found);
  if (found)
    return frontier;
  oldExprNames = VF(vf)->free[V_EXPRPTR];
  (void) KEEP(oldExprNames);
  oldMapGetNames = VF(vf)->free[V_MAPGET];
  (void) KEEP(oldMapGetNames);
  switch (x->kind)
    {
    case E_NIL:
      break;
    case E_PROG:
      {
	ExprPtr ll,name;
	COUNTER i;

	for (i=0, ll=CAR(x); ll!=NIL; ll=CDR(ll))
	  {
	    name = makeVarName(ll,deco,vn,vf);
	    createPendingUses(ll,deco,vf);
	    addVarName(ll,name,vn);
	  }
      }
      last = x;
      x = CDR(x);
      goto again;
    case E_CONS:
      {
	ExprPtr val;
	Boolean compile;
	COUNTER storageclass;

	if (isConstant(x,deco,&compile,&val,&storageclass))
	  {
	    if (compile)
	      {
		addConstant(x,val,vn,storageclass);
		goto norm;
	      }
	    /* otherwise, do nothing; therefore, for each statement x, we
	       have 3 cases:
	       - VN(vn)->names maps x to nothing: x is a constant that is
	         used to build other constants only (so don't compile any
		 bytecode for it)
	       - VN(vn)->names maps x to some name, and VN(vn)->bconsts
	         (or VN(vn)->uconsts in the case of unboxed constants)
		 also maps x to some integer: x is a constant that we
		 have to store in the local array of constants; we then
		 have to compile an instruction that loads the constant
		 into the designated variable.
	       - or VN(vn)->names maps x to some name, but x is neither in
	         VN(vn)->bconsts or in VN(vn)->uconsts: x is not a constant.
	    */
	  }
	else
	  {
	    ExprPtr name,e;

	    e = CAR(x);
	    updatePendingUses(e,deco,vn,vf);
	    if (e->kind==(CHAR)E_COUNTED)
	      switch (COUNTEDN(e))
		{
		  /* case Z_CODE: suppressed, but should be redone as V_CONST */
		case Z_APPLY:
		case Z_CALL:
		  VN(vn)->has_calls = true;
		  /*fallthrough*/
		default:
norm:
		  name = makeVarName(x,deco,vn,vf);
		  createPendingUses(x,deco,vf);
		  break;
	    }
	    else
	      name = MapGetSure(e,VN(vn)->names);
	    addVarName(x,name,vn);
	  }
      }
      addZap(x,vn,vf,oldExprNames,oldMapGetNames);
      last = x;
      x = CDR(x);
      goto again;
    case E_IF:
      {
	ExprPtr test;

	test = CAR(CAR(x));
	updatePendingUses(test,deco,vn,vf);
	if (test->kind==(CHAR)E_COUNTED)
	  {
	    if (COUNTEDN(test)==Z_START_HANDLE)
	      VN(vn)->has_jumps = true;
	  }
      }
      addZap(x,vn,vf,oldExprNames,oldMapGetNames);
      last = CAR(x);
      frontier = getVarNames1(CDR(last),deco,vn,varFreeListCopy(vf),frontier,last);
      last = x;
      x = CDR(x);
      goto again;
    case E_GOTO:
      {
	struct upu u;

	u.vn = vn;
	u.vf = vf;
	u.deco = deco;
	(void) do_updatependinguses(MapGetSure(last,COUNTEDVAL(CAR(x))),&u);
      }
      {
	ExprPtr vff;
	Boolean found;

	vff = MapGet(x,frontier,&found);
	if (found)
	  varFreeListMerge(vff,vf);
	else
	  frontier = MapPut(x,vf,frontier);
      }
      break;
    default:
      kinderror(message(536),x->kind);
      /*NOTREACHED*/
      break;
    }
  KEEPONLY(frontier);
  return frontier;
}

PUBLIC ExprPtr getVarNames(c)
IN CONST ExprPtr c;
{
  Code *code;
  ExprPtr vn,vf,x,last,name;
  MapPtr frontier,before;

  code = (Code *)EXTDATA(c);
  vn = varNamesNew();
  vf = varFreeListNew();
  frontier = NewMap();
  before = NewMap();
  x = code->prgm;
  last = NIL;
  while (true)
    {
      frontier = getVarNames1(x,code->deco,vn,vf,frontier,last);
      if (frontier==NewMap())
	break;
      x = getHighestMergeNode(frontier,code->deco,&before);
      vf = MapGetSure(x,frontier);
      frontier = SetRemove(x,frontier);
      name = makeVarName(x,code->deco,vn,vf);
      addVarName(x,name,vn);
      createPendingUses(x,code->deco,vf);
      switch (storageClass(getPpInfo(x,code->deco)->storage,x))
	{ /* indicate that some values can be zapped (used in outputCode1(),
	     case of E_GOTOs. */
	case V_MAPGET: setMapPut(x,UNIT,&VN(vn)->zapm); HITBARRIER(vn); break;
	case V_EXPRPTR: setMapPut(x,UNIT,&VN(vn)->zapx); HITBARRIER(vn); break;
	}
      last = x;
      x = CDR(x);
    }
  return vn;
}

#ifdef OBSOLETE

/***** Splitting codes into run-time and init-time codes. *****/

ARGMAPFUNC(do_rewrite)
{
  return MapGetSure(x,MAPDATA(MapPtr));
}

PRIVATE ExprPtr rewriteExpr(e,done)
IN CONST ExprPtr e;
IN CONST MapPtr done;
{
  ENTER;

  if (e->kind==(CHAR)E_COUNTED)
    e = zApply(COUNTEDN(e),argMap(COUNTEDVAL(e),do_rewrite,done));
  else
    e = do_rewrite(e,done);
  KEEPONLY(e);
  return e;
}

ARGMAPFUNC(do_irewrite)
{
  return getPpInfo(x,MAPDATA(ExprPtr))->def;
}

PRIVATE ExprPtr rewriteIExpr(e,ideco)
IN CONST ExprPtr e;
IN CONST ExprPtr ideco;
{
  ENTER;

  if (e->kind==(CHAR)E_COUNTED)
    e = zApply(COUNTEDN(e),argMap(COUNTEDVAL(e),do_irewrite,ideco));
  else
    e = do_irewrite(e,ideco);
  KEEPONLY(e);
  return e;
}

MAPSWEEPFUNC(do_existsnoninituse)
{
  ExprPtr deco;

  deco = MAPDATA(ExprPtr);
  return (getPpInfo(x,deco)->storage & SCF_NOINIT)!=0;
}

PRIVATE Boolean existsNonInitUse(pp,deco)
IN CONST ppInfo *pp;
IN CONST ExprPtr deco;
{
  return MapSweep(pp->uses,do_existsnoninituse,deco);
}

PRIVATE INT getKnownName(val)
IN CONST ExprPtr val;
{
  if (val==trueValue)
    return I_TRUEVALUE;
  if (val==falseValue)
    return I_FALSEVALUE;
  if (val==NIL)
    return I_NIL;
  if (val==NewMap())
    return I_EMPTY;
  if (val==tuple0)
    return I_TUPLE0;
  if (val==record0)
    return I_RECORD0;
  if (val==array0)
    return I_ARRAY0;
  if (val==zzero)
    return I_ZZERO;
  if (val==one)
    return I_ONE;
  if (val==mone)
    return I_MONE;
  if (val==two)
    return I_TWO;
  if (val==half)
    return I_HALF;
  if (val==zpi)
    return I_ZPI;
#if FLOAT_FORMAT==IEEE754
  if (val==pinf)
    return I_PINF;
  if (val==minf)
    return I_MINF;
#endif
  if (val==izero)
    return I_IZERO;
  if (val==ione)
    return I_IONE;
  if (val==itwo)
    return I_ITWO;
  return -1;
}

PRIVATE MapPtr splitCode1 P((IN CONST ExprPtr x,IN CONST ExprPtr last,
			     INOUT ExprPtr *runlastp,INOUT ExprPtr *initlastp,
			     IN CONST ExprPtr deco,IN OUT ExprPtr ideco,
			     IN CONST MapPtr ieqns,
			     IN CONST MapPtr done,INOUT INT *constn));

DEF_PRIVATE_ACTION(splitcode)
{
  struct splitcode *data;

  data = (struct splitcode *)env;
  return splitCode1(data->x,data->last,&data->runlast,&data->initlast,data->deco,data->ideco,data->ieqns,data->done,&data->constn);
}

PRIVATE MapPtr splitCode1k(x,last,runlastp,initlastp,deco,ideco,ieqns,done,constn)
IN CONST ExprPtr x;
IN CONST ExprPtr last;
INOUT ExprPtr *runlastp,*initlastp;
IN CONST ExprPtr deco;
IN OUT ExprPtr ideco;
IN CONST MapPtr ieqns;
IN CONST MapPtr done;
INOUT INT *constn;
{
  struct splitcode *data = (struct splitcode *)alloctmp(sizeof(struct splitcode));
  ExprPtr res;

  data->x = x;
  data->last = last;
  data->runlast = *runlastp;
  data->initlast = *initlastp;
  data->deco = deco;
  data->ideco = ideco;
  data->ieqns = ieqns;
  data->done = done;
  data->constn = *constn;
  res = CallCC(ACTION_NAME(splitcode),(ExprPtr)data,true);
  *runlastp = data->runlast;
  *initlastp = data->initlast;
  *constn = data->constn;
  return res;
}

PRIVATE MapPtr splitCode1(x,last,runlastp,initlastp,deco,ideco,ieqns,done,constn)
IN CONST ExprPtr x;
IN CONST ExprPtr last;
INOUT ExprPtr *runlastp,*initlastp;
IN CONST ExprPtr deco;
IN OUT ExprPtr ideco;
IN CONST MapPtr ieqns;
IN CONST MapPtr done;
INOUT INT *constn;
{
  ppInfo *pp;
  ExprPtr e,y;
  Boolean found;

  if (STKFULL(found))
    return splitCode1k(x,last,runlastp,initlastp,deco,ideco,ieqns,done,constn);
again:
  switch (x->kind)
    {
    case E_NIL:
      break;
    case E_CONS:
      (void) MapGet(x,done,&found);
      if (found)
	return done;
      pp = getPpInfo(x,deco);
      if ((pp->storage & SCF_NOINIT) ||
	  ((pp->storage & SCF_UNBOXED) &&
	   (pp->storage & (SCF_BOOL | SCF_INT | SCF_STRING | SCF_CODE))))
	{
	  e = rewriteExpr(CAR(x),done);
	  *runlastp = tackAssign(e,*runlastp);
	}
      else
	{
	  if ((pp->storage & SCF_BOXED) &&
	      pp->valset->kind==(CHAR)E_MAPLET)
	    { /* constant boxed expression */
	      INT i;

	      i = getKnownName(MAPLETX(pp->valset));
	      if (i>=0)
		{
		  *runlastp = tackApplyBin(Z_C_GLOBAL,KINT(i),
					   KINT(pp->storage),*runlastp);
		  goto next;
		}
	    }
	  e = rewriteIExpr(CAR(x),ideco);
	  y = MapGet(e,ieqns,&found);
	  if (!found)
	    {
	      ppInfo *ppp;

	      *initlastp = y = tackAssign(e,*initlastp);
	      ieqns = MapPut(e,y,ieqns);
	      ppp = getPpInfo(y,ideco); /* create decoration */
	      SETQ(ppp->def,y);
	      SETQ(ppp->eqns,ieqns);
	      SETQ(ppp->fragile,pp->fragile);
	      SETQ(ppp->valset,pp->valset);
	      ppp->storage = pp->storage;
	      memcpy(ppp->tags,pp->tags,NTAGS);
	    }
	  if (existsNonInitUse(pp,deco))
	    { /* if it is used at run-time, generate a const C name for it. */
	      ExprPtr name,instr;

	      if (found && CDR(y)->kind==(CHAR)E_CONS &&
		  (instr = CAR(CDR(y)),
		   COUNTEDN(instr)==Z_SET_C_CONST &&
		   CDR(COUNTEDVAL(instr))==y))
		name = CAR(COUNTEDVAL(instr));
	      else
	      {
		ppInfo *ppp;

		name = KINT((*constn)++);
		e = zApplyBin(Z_SET_C_CONST,name,y);
		*initlastp = y = tackAssign(e,*initlastp);
		ieqns = MapPut(e,y,ieqns);
		ppp = getPpInfo(y,ideco); /* create decoration */
		SETQ(ppp->def,y);
		SETQ(ppp->eqns,ieqns);
		(void) newUse(y,NIL,ideco);
	      }
	      *runlastp = tackApplyBin(Z_C_CONST,name,KINT(pp->storage),*runlastp);
	    }
	}
next:
      done = MapPut(x,*runlastp,done);
      last = x;
      x = CDR(x);
      goto again;
    case E_IF:
      (void) MapGet(x,done,&found);
      if (found)
	return done;
      {
	ExprPtr thenlast;

	thenlast = compileIf(rewriteExpr(CAR(CAR(x)),done),*runlastp,runlastp);
	/* split then part */
	done = splitCode1(CDR(CAR(x)),CAR(x),&thenlast,initlastp,deco,ideco,ieqns,done);
	/* then split else part */
	last = x;
	x = CDR(x);
	goto again;
      }
    case E_GOTO:
      {
	ExprPtr l,r;

	l = do_rewrite(last,done);
	r = do_rewrite(MapGetSure(last,COUNTEDVAL(CAR(x))),done);
	e = MapGet(x,done,&found);
	if (found)
	  *runlastp = mergeContinuations(r,l,e);
	else
	  {
	    *runlastp = mergeContinuations(r,l,NIL);
	    goto next;
	  }
      }
      break;
    default:
      kinderror(message(576),x->kind);
      /*NOTREACHED*/
      break;
    }
  return done;
}

PRIVATE MapPtr copyTranslation(arglist,trans)
IN CONST ExprPtr arglist;
OUT struct translation *trans;
{
  ExprPtr list,ll;
  MapPtr done;
  ENTER;

  for (list=NIL, ll=arglist; ll!=NIL; ll=CDR(ll))
    {
      list = MakePair(E_CONS,CAR(ll),list);
      KEEPONLY(list);
    }
  list = Nreverse1(list,NIL);
  KEEPONLY(list);
  done = NewMap();
  for (ll=arglist; ll!=NIL; ll=CDR(ll), list=CDR(list))
    done = MapPut(ll,list,done);
  KEEPONLY(done);
  trans->prgm = MakePair(E_PROG,list,NIL);
  (void) KEEP(trans->prgm);
  return done;
}

PRIVATE ExprPtr splitCode(c,initc,constn)
IN CONST ExprPtr c;
INOUT ExprPtr initc;
INOUT INT *constn;
{
  Code *code,*initcode;
  struct translation *fun = (struct translation *)alloctmp(sizeof(struct translation));
  ExprPtr runlast,initlast;
  MapPtr done;
  ExprPtr runc;
  Boolean found;
  ExprPtr deco;
  ENTER;

  code = (Code *)EXTDATA(c);
  done = copyTranslation(CAR(code->prgm),fun);
  initcode = (Code *)EXTDATA(initc);
  initlast = initcode->last;
  runlast = fun->prgm;
  done = splitCode1(CDR(code->prgm),code->prgm,&runlast,&initlast,code->deco,
		    initcode->deco,getPpInfo(initlast,initcode->deco)->eqns,
		    done,constn);
  runlast = MapGet(code->last,done,&found);
  if (!found)
    runlast = NIL;
  runc = makeCode(fun,runlast);
  deco = ppDecoInit();
  ppDecorate(fun->prgm,runlast,deco,NewMap(),NewMap());
  fillCode(runc,deco);
  ((Code *)EXTDATA(runc))->flags = code->flags;
  KEEPONLY(runc);
  return runc;
}

#endif

/***** Function that output bytecode. *****/

PRIVATE bcFile *bcOpen()
{
  bcBuffer *buf;
  bccBuffer *cbuf;
  bcuBuffer *ubuf;
  bcFile *f;

  buf = (bcBuffer *)MALLOC(sizeof(bcBuffer));
  buf->next = NULL;
  buf->n = 0;
  cbuf = (bccBuffer *)MALLOC(sizeof(bccBuffer));
  cbuf->next = NULL;
  cbuf->n = 0;
  ubuf = (bcuBuffer *)MALLOC(sizeof(bcuBuffer));
  ubuf->next = NULL;
  ubuf->n = 0;
  f = (bcFile *)MALLOC(sizeof(bcFile));
  f->buf = f->current = buf;
  f->cbuf = f->ccurrent = cbuf;
  f->ubuf = f->ucurrent = ubuf;
  return f;
}

PRIVATE INT byteCodeSize(n,cn,un,rawdata,instr)
IN INT n,cn,un;
OUT COUNTER *rawdata,*instr;
{
  INT size;

  size = VARSIZE(ByteCodeChunk,consts,cn);
  if (rawdata!=NULL)
    *rawdata = (COUNTER)size;
  size += un;
  size = (size+sizeof(OPCODE)-1)%sizeof(OPCODE); /* align on OPCODE */
  if (instr!=NULL)
    *instr = size;
  size += n*sizeof(OPCODE);
  return size;
}

MAPSWEEPFUNC(do_bconsts)
{
  struct bconsts *data;

  data = MAPDATA(struct bconsts *);
  data->expr[IVAL(y)] = MapGetSure(x,VN(data->vn)->consts);
  return false;
}

MAPSWEEPFUNC(do_uconsts)
{
  struct uconsts *data;
  COUNTER offset,storageClass;
  ExprPtr val;

  data = MAPDATA(struct uconsts *);
  offset = VARRANK(y);
  storageClass = VARLEVEL(y);
  val = MapGetSure(x,VN(data->vn)->consts);
  switch (storageClass)
    {
    case V_EXPRPTR:
    case V_MAPGET:
    case V_MAPSWEEP:
      *(ExprPtr *)data->rawdata[offset] = val;
      break;
    case V_BOOL:
      *(char *)data->rawdata[offset] = (val==trueValue);
      break;
    case V_INT:
      *(INT *)data->rawdata[offset] = IVAL(val);
      break;
    case V_STRING:
      memcpy(data->rawdata[offset],STEXT(val),SLEN(val));
      break;
    case V_CODE:
      *(char **)data->rawdata[offset] = <contents of CODE part, should be a bFunction *>
	break;
    case V_REAL:
      *(EXTENDED *)data->rawdata[offset] = XVAL(val);
      break;
    default: kinderror(message(544),storageClass); break;
    }
  return false;
}

PRIVATE byteCodeChunk *bcClose(f,vn)
INOUT bcFile *f;
IN CONST ExprPtr vn;
{
  bcBuffer *buf,*next;
  INT n;
  COUNTER rawdata,instr,size;
  byteCode bc;
  OPCODE *to;
  ExprPtr *cto;
  char *uto;

  for (n=0, buf=f->buf; buf!=NULL; buf=buf->next)
    n+= buf->n;
  size = byteCodeSize(n,VN(vn)->maxbconst,VN(vn)->maxuconst,&rawdata,&instr);
  bc = (byteCode)MALLOC(size);
  bc->refcount = 0;
  bc->nConsts = VN(vn)->maxbconst;
  bc->rawDataOff = rawdata;
  bc->instrOff = instr;
  bc->size = size;
  {
    struct bconsts *data = (struct bconsts *)alloctmp(sizeof(struct bconsts));

    data->vn = vn;
    data->expr = bc->consts;
    (void) MapSweep(VN(vn)->bconsts,do_bconsts,(char *)data);
  }
  {
    struct uconsts *data = (struct uconsts *)alloctmp(sizeof(struct uconsts));

    data->vn = vn;
    data->rawdata = (char *)(((char *)bc)+rawdata);
    (void) MapSweep(VN(vn)->uconsts,do_uconsts,(char *)data);
  }
  for (to=(OPCODE *)(((char *)bc)+instr), buf=f->buf; buf!=NULL; buf=next)
    {
      memcpy(to,buf->instr,buf->n*sizeof(OPCODE));
      next = buf->next;
      to += buf->n;
      MARKFREE(buf);
    }
  MARKFREE(f);
  return bc;
}

PRIVATE void putCode(c,f)
IN OPCODE c;
INOUT bcFile *f;
{
  REGISTER bcBuffer *current;

  current = f->current;
  if (current->n==BC_BUFSIZE)
    {
      bcBuffer *next;

      next = (bcBuffer *)MALLOC(sizeof(bcBuffer));
      next->n = 0;
      next->next = NULL;
      current->next = next;
      f->current = current = next;
    }
  current->instr[current->n++] = c;
}

PRIVATE INT bcTell(f)
IN CONST bcFile *f;
{
  REGISTER bcBuffer *buf;
  REGISTER INT n;

  for (n=0, buf=f->buf; buf!=NULL; n += buf->n, buf = buf->next);
  return n;
}

PRIVATE void bcPatchJump(addr,off,f)
IN REGISTER INT addr; /* address just after the opcode for a branching instruction */
IN OPCODE off; /* offset after addr where we find the [offset] to patch */
INOUT bcFile *f;
{
  REGISTER INT here;
  REGISTER OPCODE offset;
  REGISTER bcBuffer *buf;

  here = bcTell(f);
  offset = here-addr;
  if (addr+offset!=here)
    berror(message(94));
  addr += off; /* where to patch */
  for (buf=f->buf; buf!=NULL; buf = buf->next)
    {
      if (addr<buf->n)
	break;
      addr -= buf->n;
    }
  buf->instr[addr] = offset;
}

PRIVATE void outputFunStart(prgm,f,vn,hasexprsp)
IN CONST ExprPtr prgm;
INOUT bcFile *f;
IN ExprPtr vn;
OUT Boolean *hasexprsp;
{
  ExprPtr ll,name;
  COUNTER storage;
  varNames *vnn;
  COUNTER ne,ng,nm,n;
  OPCODE code;
  COUNTER i;

  vnn = (varNames *)EXTDATA(vn);
  if (vnn->has_calls)
    putCode(K_TEST_STACK,f); /* should be the first instruction in the code
				to avoid useless replay of the first instructions
				in case the stack overflows */
  ne = vnn->maxn[V_EXPRPTR];
  ng = vnn->maxn[V_MAPGET];
  nm = vnn->maxn[V_MAPSWEEP];
  n = ne+ng+nm*(MAP_CHAIN_LENGTH+1);
  /* declare number of entries in the local array of protected variables;
     ne entries for ExprPtrs, ng entries for MapGets,
     nm*(MAP_CHAIN_LENGTH+1) entries for MapSweeps
     (those are hand-coded stacks). */
  if (*hasexprsp = (n!=0))
    {
      putCode(K_ALLOC_EXPRS,f);
      putCode(n,f);
      putCode(nm,f);
    }
  else
    putCode(K_ALLOC_NOEXPR,f);
  putCode(vnn->maxn[V_STRING]);
  putCode(vnn->maxn[V_CODE]);
  putCode(vnn->maxn[V_REAL]);
  putCode(vnn->maxn[V_INT]);
  putCode(vnn->maxn[V_BOOL]+vnn->maxn[V_MAPGET]);
  for (n = 1, ll = CAR(prgm); ll!=NIL; ll=CDR(ll), n++)
    {
      storage = storageClass(IVAL(CAR(ll)),CAR(ll));
      switch (storage)
	{
	case V_EXPRPTR:
	case V_MAPGET:
	case V_MAPSWEEP:
	  code = K_PARAM_EXPR; break;
	case V_BOOL: code = K_PARAM_BOOL; break;
	case V_INT: code = K_PARAM_INT; break;
	case V_STRING: code = K_PARAM_STRING; break;
	case V_CODE: code = K_PARAM_CODE; break;
	case V_REAL: code = K_PARAM_REAL; break;
	default: kinderror(message(578),storage); break;
	}
      putCode(code,f);
      i = VARRANK(MapGetSure(ll,vnn->names));
      putCode((OPCODE)i,f);
    }
  putCode((OPCODE)K_END_PARAM,f);
}

PRIVATE void outputCode1 P((IN CONST ExprPtr x,IN CONST ExprPtr last,
			    IN CONST ExprPtr c,IN CONST ExprPtr vn,
			    INOUT bcFile *f,INOUT ExprPtr doneref,
			    IN Boolean hasexprs));

DEF_PRIVATE_ACTION(outputcode)
{
  struct outputcode *data;

  data = (struct outputcode *)env;
  outputCode1(data->x,data->last,data->c,data->vn,data->f,
	      data->doneref,data->hasexprs);
  return NIL;
}

PRIVATE void outputCode1k(x,last,c,vn,f,doneref,hasexprs)
IN CONST ExprPtr x,last;
IN CONST ExprPtr c;
IN CONST ExprPtr vn;
INOUT bcFile *f;
INOUT ExprPtr doneref;
IN Boolean hasexprs;
{
  struct outputcode *data = (struct outputcode *)alloctmp(sizeof(struct outputcode));

  data->x = x;
  data->last = last;
  data->c = c;
  data->vn = vn;
  data->f = f;
  data->doneref = doneref;
  data->hasexprs = hasexprs;
  (void) CallCC(ACTION_NAME(outputcode),(ExprPtr)data,true);
  return;
}

PRIVATE Boolean loadIfConstant(x,vn,f)
IN CONST ExprPtr x;
IN CONST ExprPtr vn;
INOUT bcFile *f;
{
  ExprPtr idx;
  Boolean found;
  OPCODE code;

  idx = MapGet(x,VN(vn)->bconsts,&found);
  if (found)
    {
      putCode(K_CONST_EXPR,f);
      putCode((OPCODE)IVAL(idx),f);
      return true;
    }
  else
    {
      idx = MapGet(x,VN(vn)->uconsts,&found);
      if (found)
	{
	  switch (VARLEVEL(idx))
	    {
	    case V_EXPRPTR:
	    case V_MAPGET:
	    case V_MAPSWEEP:
	      code = K_CONST_EXPR; break;
	    case V_BOOL: code = K_CONST_BOOL; break;
	    case V_INT: code = K_CONST_INT; break;
	    case V_STRING: code = K_CONST_STRING; break;
	    case V_CODE: code = K_CONST_CODE; break;
	    case V_REAL: code = K_CONST_REAL; break;
	    default:
	      kinderror(message(542),VARLEVEL(idx));
	      break;
	    }
	  putCode(code,f);
	  putCode((OPCODE)VARRANK(idx),f);
	  return true;
	}
    }
  return false;
}

PRIVATE void outputAssignVar(x,y,vn,f)
IN CONST ExprPtr x,y;
IN CONST ExprPtr vn;
INOUT bcFile *f;
{
  ExprPtr varx,vary;
  OPCODE code;
  Boolean found1,found2;

  varx = MapGet(x,VN(vn)->names,&found1);
  vary = MapGet(y,VN(vn)->names,&found2);
  if (found1 && found2 && varx!=vary)
    {
      switch (VARLEVEL(varx))
	{
	case V_EXPRPTR:
	case V_MAPGET:
	case V_MAPSWEEP:
	  code = K_ASSIGN_EXPR; break;
	case V_BOOL: code = K_ASSIGN_BOOL; break;
	case V_INT: code = K_ASSIGN_INT; break;
	case V_STRING: code = K_ASSIGN_STRING; break;
	case V_CODE: code = K_ASSIGN_CODE; break;
	case V_REAL: code = K_ASSIGN_REAL; break;
	default:
	  kinderror(message(543),VARLEVEL(varx));
	  break;
	}
      putCode(code,f);
      putCode((OPCODE)VARRANK(varx),f);
      putCode((OPCODE)VARRANK(vary),f);
    }
}

PRIVATE void outputAssignExpr(x,e,vn,f)
IN CONST ExprPtr x;
IN CONST ExprPtr e;
IN CONST ExprPtr vn;
INOUT bcFile *f;
{
  ExprPtr arg;

  arg = COUNTEDVAL(e);
  switch (COUNTEDN(e))
    { /* deal with special cases first */
    case Z_UNBOXED_REAL:
      /* should not happen (constants are compiled by loadIfConstant()) */
/*!!! output code for all the following expressions: */
    case Z_BOX_UNSHARED_TUPLE_1:
      fun = "  MKSING(E_TUPLE_1,";
using:
      (void) hfprintf(of->runCode,fun);
      outputNameOf(arg,names,of);
      goto uend;
    case Z_BOX_UNSHARED_RECORD_1:
      fun = "  MKPAIR(E_RECORD_1,";
ucons:
      (void) hfprintf(of->runCode,fun);
      outputNameOf(CAR(arg),names,of);
      (void) hfputc(',',of->runCode);
      outputNameOf(CDR(arg),names,of);
uend:
      (void) hfputc(',',of->runCode);
      outputName(name,of->runCode);
      (void) hfputc(')',of->runCode);
      break;
    case Z_CODE: /* codes and globals are given the right name by
		    getVarNames1() already, so don't do anything. */
    case Z_C_GLOBAL:
    case Z_ASSERT_EQUAL: /* this is just an assertion, don't produce code for it */
      return;
    case Z_QUEUE_TO_LIST: fun = "  Q2L("; goto ucons;
    case Z_BOX_UNSHARED_CONS: fun = "  MKPAIR(E_CONS,"; goto ucons;
    case Z_BOX_UNSHARED_TUPLE_2: fun = "  MKPAIR(E_TUPLE_2,"; goto ucons;
    case Z_BOX_TUPLE_GNRL:
      outputVector(name,arg,names,of,true,true);
      break;
    case Z_BOX_UNSHARED_TUPLE_GNRL:
      outputVector(name,arg,names,of,true,false);
      break;
    case Z_BOX_RECORD_GNRL:
      outputVector(name,arg,names,of,false,true);
      break;
    case Z_BOX_UNSHARED_RECORD_GNRL:
      outputVector(name,arg,names,of,false,false);
      break;
    case Z_CONSENV:
      (void) hfputs("  MKENV(",of->runCode);
      outputNameOf(CAR(arg),names,of);
      (void) hfprintf(of->runCode,",%d,",IVAL(CDR(arg)));
      outputName(name,of->runCode);
      (void) hfputc(')',of->runCode);
      break;
    case Z_FAKEENV:
      (void) hfputs("  FAKEENV(",of->runCode);
      outputNameOf(arg,names,of);
      (void) hfputc(')',of->runCode);
      break;
    case Z_LET:
      (void) hfputs("ELET(",of->runCode);
      outputNameOf(CAR(arg),names,of);
      (void) hfputc(',',of->runCode);
      arg = CDR(arg);
      outputNameOf(CAR(arg),names,of);
      (void) hfprintf(of->runCode,",%ld)",IVAL(CDR(arg)));
      break;
    case Z_MAPGET: fun = "  MAPGET("; goto ucons;
    case Z_BOX_QUEUE: fun = MKQ(; goto using;
    case Z_BOX_1_QUEUE: fun = "MK1Q("; goto using;
    case Z_START_LISTSWEEP: fun = "START_LISTSWEEP("; goto using;
    case Z_START_MAPSWEEP: fun = "START_MAPSWEEP("; goto using;
    case Z_START_SUBMAPSWEEP: fun = "START_SUBMAPSWEEP("; goto using;
    case Z_CONCAT_QUEUE: fun = "CONCATQ("; goto ucons;
    default:
      outputAssign1(name,of);
      switch (COUNTEDN(e))
	{
	case Z_NIL:
	  fun = "NIL";
cst:
	  (void) hfputs(fun,of->runCode);
	  break;
	case Z_UNBOXED_INT:
	  (void) hfprintf(of->runCode,"%ld",IVAL(arg));
	  break;
	case Z_UNBOX_INT:
	  fun = "IVAL(";
unary:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(arg,names,of);
paren:
	  (void) hfputc(')',of->runCode);
	  break;
	case Z_BOX_INT: fun = "Kint("; goto unary;
	case Z_UNBOX_REAL: fun = "(EXTENDED)XVAL("; goto unary;
	case Z_BOX_REAL: fun = "Kext("; goto unary;
	case Z_UNBOX_TUPLE_NTH:
	  fun = "AVAL(";
indexing:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfprintf(of->runCode,")->expr[%ld]",IVAL(CDR(arg)));
	  break;
	case Z_BOX_COMPLEX:
	  fun = "Kcomplex(";
binary:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputc(',',of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  goto paren;
	case Z_UNBOXED_STRING:
	  (void) hfprintf(of->runCode,"_S_%s+%ld",modulename,of->offset);
	  outputString(arg,of);
	  break;
	case Z_UNBOX_STRING_LEN: fun = "SLEN("; goto unary;
	case Z_UNBOX_STRING_TEXT: fun = "STEXT("; goto unary;
	case Z_BOX_STRING: fun = "Kstring("; goto binary;
	case Z_EMPTY: fun = "empty_map"; goto cst;
	case Z_TUPLE_0: fun = "tuple0"; goto cst;
	case Z_RECORD_0: fun = "record0"; goto cst;
	case Z_BOX_TYPE_APPL: fun = "KtypeAppl("; goto binary;
	case Z_BOX_REF: fun = "Kref("; goto unary;
	case Z_BOX_TUPLE_1: fun = "Ktuple1("; goto unary;
	case Z_BOX_RECORD_1: fun = "Krecord1("; goto binary;
	case Z_BOX_CLOSURE: fun = "MakeCompiled(E_COMPILED_CLOSURE,"; goto binary;
	case Z_BOX_CONS: fun = "Kons("; goto binary;
	case Z_BOX_TUPLE_2: fun = "Ktuple2("; goto binary;
	case Z_BOX_SINGLETON: fun = "MapletKons("; goto binary;
	case Z_OVERWRITE: fun = "MapOverwrite("; goto binary;
	case Z_ENEXT: fun = "ENEXT("; goto unary;
	case Z_VAR: fun = "EVARS("; goto indexing;
	case Z_RAISE: fun = "(void) Raise("; goto unary;
	case Z_MAPPUT:
	  fun = "MapAdd(";
ternary:
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputc(',',of->runCode);
	  arg = CDR(arg);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputc(',',of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  goto paren;
	case Z_EQUAL:
	  fun = " == ";
cmp:
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputs(fun,of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  break;
	case Z_CAR: fun = "CAR("; goto unary;
	case Z_CDR: fun = "CDR("; goto unary;
	case Z_UNBOX_REF: fun = "REFVAL("; goto unary;
	case Z_HAS_TAG:
	  outputNameOf(CAR(arg),names,of);
	  (void) hfprintf(of->runCode,"->kind==(CHAR)%ld",IVAL(CDR(arg)));
	  break;
	case Z_GET_VECTOR_LEN:
	  (void) hfputs("AVAL(",of->runCode);
	  outputNameOf(arg,names,of);
	  (void) hfputs("->n",of->runCode);
	  break;
	case Z_TYPE_MATCH: fun = "TypeMatch("; goto binary;
	case Z_MAP_CARD_DEC:
	  (void) hfputs("MapCardDec(",of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfprintf(of->runCode,",%ld)",IVAL(CDR(arg)));
	  break;
	case Z_INT_LEQ: fun = "<="; goto cmp;
	case Z_GET_SHAPE: fun = "ASHAPE("; goto unary;
	case Z_GET_MAPGET_Y: fun = "MAPGETY("; goto unary;
	case Z_GET_MAPGET_FOUND: fun = "MAPGETF("; goto unary;
	case Z_TUPLE_SUB_NTH:
	  (void) hfputs("AVAL(",of->runCode);
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputs(")->expr[",of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  (void) hfputs("-1]",of->runCode);
	  break;
	case Z_MAPPUTBEHIND: fun = "MapSlip("; goto ternary;
	case Z_MAPREMOVE: fun = "MapKill("; goto binary;
	case Z_UNBOXED_FALSE: fun = "false"; goto cst;
	case Z_UNBOXED_TRUE: fun = "true"; goto cst;
	case Z_REF_SET: fun = "SETREFVAL("; goto binary;
	case Z_BOX_NUMTYPE: fun = "KnumType("; goto unary;
	case Z_BOX_TYPEFX: fun = "KtypeFx("; goto binary;
	case Z_APPLY:
	case Z_CALL:
	  outputApply(arg,names,of);
	  break;
	case Z_UNBOX_CLOSURE_FN: fun = "ACTION("; goto unary;
	case Z_UNBOX_CLOSURE_ENV: fun = "QENV("; goto unary;
	case Z_APPEND: fun = "Append0("; goto binary;
	case Z_MAPINTER:
	case Z_MAPDOMRESTRTO: fun = "MapDRT("; goto binary;
	case Z_MAPDOMRESTRBY: fun = "MapDRB("; goto binary;
	case Z_UNBOX_LISTSWEEP: fun = "UNBOX_LISTSWEEP("; goto unary;
	case Z_UNBOX_MAPSWEEP: fun = "UNBOX_MAPSWEEP("; goto unary;
	case Z_UNBOX_SUBMAPSWEEP: fun = "UNBOX_SUBMAPSWEEP("; goto unary;
	case Z_BOX_LOCAL_REF:
	case Z_UNBOX_LOCAL_REF:
	  outputNameOf(arg,names,of);
	  break;
	case Z_LOCAL_REF_SET:
	  outputNameOf(CAR(arg),names,of);
	  (void) hfputs(" = ",of->runCode);
	  outputNameOf(CDR(arg),names,of);
	  break;
	case Z_REVERSE1: fun = "Reverse1("; goto binary;
	case Z_BOX_UNSHARED_EXCEPTION: fun = "MakeException("; goto unary;
	case Z_BOX_SHARED_EXCEPTION: fun = "Kref(E_EXCEPTION,"; goto unary;
	case Z_BOX_BOOL: fun = "Kbool("; goto unary;
	case Z_UNBOX_BOOL: fun = "BVAL("; goto unary;
	case Z_END_HANDLE:
	  (void) hfputs("END_HANDLE()",of->runCode);
	  break;
	case Z_RMU1: fun = "RMU1("; goto ternary;
	case Z_UNSHARED_RMU1: fun = "UNSHARED_RMU1("; goto ternary;
	case Z_RMUX: fun = "RMUX("; goto binary;
	case Z_UNSHARED_RMUX: fun = "UNSHARED_RMUX("; goto binary;
	case Z_MAKE_PROMISE: fun = "MakePromise("; goto unary;
	case Z_BOX_PACK: fun = "MakePack(E_PACK,"; goto binary;
	case Z_CMPSTR: fun = "CompareStrings("; goto binary;
	case Z_INC_LISTSWEEP:
	case Z_INC_MAPSWEEP:
	case Z_INC_SUBMAPSWEEP:
	case Z_START_HANDLE:
	case Z_FAKE_UNBOX_TUPLE_LEN:
	case Z_FAKE_MAPGET:
	case Z_FAKE_NOT_MAPGET:
	case Z_FAKE_MAPGET_FOUND:
	default:
	  kinderror("outputAssignExpr",COUNTEDN(e));
	  /*NOTREACHED*/
	  break;
	}
    }
  (void) hfputs(";\n",of->runCode);
}

PRIVATE void outputCode1(x,last,c,vn,f,doneref,hasexprs)
IN CONST ExprPtr x,last;
IN CONST ExprPtr c;
IN CONST ExprPtr vn;
INOUT bcFile *f;
INOUT ExprPtr doneref;
IN Boolean hasexprs;
{
  Code *code;
  ExprPtr name,e;
  MapPtr names;
  Boolean found;
  ENTER;

  if (STKFULL(found))
    {
      outputCode1k(x,last,c,vn,f,doneref,hasexprs);
      return;
    }
  code = ((Code *)EXTDATA(c));
  names = VN(vn)->names;
again:
  switch (x->kind)
    {
    case E_NIL:
      if (hasexprs)
	putCode(K_STOP_PROTECTION,f);
      if (code->last==last)
	{
	  if (code->storage!=SCF_TOP &&
	      (name = MapGet(last,names,&found), found))
	    {
	      OPCODE cc;

	      switch (storageClass(code->storage,last))
		{
		case V_EXPRPTR:
		case V_MAPGET:
		case V_MAPSWEEP:
		  cc = K_RETURN_EXPR; break;
		case V_BOOL: cc = K_RETURN_BOOL; break;
		case V_INT: cc = K_RETURN_INT; break;
		case V_STRING: cc = K_RETURN_STRING; break;
		case V_REAL: cc = K_RETURN_REAL; break;
		case V_CODE: cc = K_RETURN_CODE; break;
		default:
		  kinderror(message(537),code->storage);
		  break;
		}
	      putCode(cc,f);
	      putCode((OPCODE)VARRANK(name),f);
	    }
	  else
	    putCode(K_RETURN_VOID,f);
	}
      break;
    case E_IF:
      {
	ExprPtr car,test;
	INT pos;

	car = CAR(x);
	test = CAR(car);
	if (test->kind!=(CHAR)E_COUNTED) /* ordinary if */
	  {
	    putCode(K_IF,f);
	    pos = bcTell(f);
	    putCode(VARRANK(MapGetSure(test,names)),f);
	    putCode(DUMMY_OFFSET,f); /* prepare for patch */
	    outputCode1(CDR(x),x,c,vn,f,doneref,hasexprs);
	    bcPatchJump(pos,1,f); /* patch forward reference */
	    last = car;
	    x = CDR(last);
	    goto again;
	  }
	else /* E_COUNTED */
	  switch (COUNTEDN(test))
	    {
	    case Z_INC_LISTSWEEP:
	      putCode(K_INC_LISTSWEEP);
inc:
	      pos = bcTell(f);
	      putCode(VARRANK(MapGetSure(COUNTEDVAL(test),names)),f);
	      putCode(DUMMY_OFFSET,f); /* prepare for patch */
	      outputCode1(CDR(car),car,c,vn,f,doneref,hasexprs);
	      bcPatchJump(pos,1,f); /* patch forward reference */
	      last = x;
	      x = CDR(last);
	      goto again;
	    case Z_INC_MAPSWEEP:
	      putCode(K_INC_MAPSWEEP);
	      goto inc;
	    case Z_INC_SUBMAPSWEEP:
	      putCode(K_INC_SUBMAPSWEEP);
	      goto inc;
	    case Z_START_HANDLE:
	      putCode(K_START_HANDLE);
	      pos = bcTell(f);
	      putCode(DUMMY_OFFSET,f); /* prepare for patch */
	      outputCode1(CDR(car),car,c,vn,f,doneref,hasexprs);
	      bcPatchJump(pos,0,f); /* patch forward reference */
	      last = x;
	      x = CDR(last);
	      goto again;
	    default:
	      kinderror(message(521),COUNTEDN(test));
	      break;
	    }
      }
      break;
    case E_GOTO: /* merge points */
      {
	ExprPtr offset;
	INT off;

	/* assign the variable corresponding to the merge node to the last result */
	outputAssignVar(x,MapGetSure(last,COUNTEDVAL(CAR(x))),vn,f);
	offset = MapGet(x,CDR(doneref),&found);
	if (found) /* compile a goto */
	  {
	    putCode(K_GOTO,f);
	    off = IVAL(offset)-bcTell(f);
	    if (off!=(OPCODE)off)
	      berror(message(94));
	    putCode((OPCODE)off,f);
	  }
	else /* record the address of the label */
	  {
	    off = bcTell();
	    MAKEINT(off,offset);
	    (void) KEEP(offset);
	    setcdrMapPut(x,offset,doneref);
	    /* then compile the rest of the code */
	    last = x;
	    x = CDR(last);
	    goto again;
	  }
      }
      break;
    case E_CONS:
      name = MapGet(x,names,&found);
      if (found && !loadIfConstant(last,vn,f))
	{
	  e = CAR(x);
	  if (e->kind!=(CHAR)E_COUNTED)
	    outputAssignVar(x,e,vn,f);
	  else outputAssignExpr(x,e,vn,f);
	}
      last = x;
      x = CDR(last);
      goto again;
    default:
      kinderror(message(656),x->kind);
      break;
    }
  KEEPONLY(done);
  return done;
}

PUBLIC void outputCode(c,f)
IN CONST ExprPtr c;
INOUT bcFile *f;
{
  ExprPtr deco,done,vn,doneref;
  Code *code;
  COUNTER i;
  Boolean hasexprs;
  ENTER;

  code = (Code *)EXTDATA(c);
  deco = code->deco;
  done = NewMap();
  vn = getVarNames(c);
  outputFunStart(code->prgm,f,vn,&hasexprs);
  /* output code */
  MAKECONS(NewMap(),NewMap(),doneref);
  (void) KEEP(doneref);
  (void) outputCode1(CDR(code->prgm),code->prgm,c,vn,f,doneref,hasexprs);
  /*!!!*/

  /* then, if we has to check the stack, emit here a call to the beamer function,
     which is responsible for changing stacks when the current one is full. */
  (void) hfputs("}\n",of->runCode);
  if (VN(vn)->has_calls)
    outputBeamerJump(code->name,CAR(code->prgm),code->storage,of);
  (void) hfputs("#undef _QUIT\n#undef _N__X\n#undef _N_MAPGET\n#undef _N_EXPR\n}\n\n",of->runCode);
  QUIT;
}

#endif
