/* Initialization functions.
   Copyright (C) 1992-1999 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "mem.h"
#include "applmem.h"
#include "actions.h"
#include "execinit.h"
#include "timing.h"
#include "iniproto.h"
#include "stkproto.h"
#include "exeproto.h"
#include "memproto.h"
#include "conproto.h"
#include "mapproto.h"
#include "symproto.h"
#include "grmproto.h"
#include "typproto.h"
#include "ariproto.h"
#include "evlproto.h"
#include "cmpproto.h"
#include "emtproto.h"
#include "dbgproto.h"
#include "rndproto.h"
#include "expproto.h"
#include "trlproto.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==AMIGA_OS
#include <exec/exec.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sinit.h"
#endif

PUBLIC actions *doafterinit=(actions *)0L;
PUBLIC actions *dobeforeend=(actions *)0L;

#ifdef CORE_TRACE
PUBLIC char coreTraceName[32];
PUBLIC hFILE *coreTraceFile;
#define MAKE_CORE_FILE_NAME() strcpy(coreTraceName,"SamML.trace")
PUBLIC Boolean replay_mode=false;
#endif
PUBLIC Boolean initializing=true;

char *coreEnd = "(* END-OF-PROLOG *)";

EXTERN char *codeStatus[];
EXTERN ExprPtr *NaNs;

PUBLIC long replay_size=0; /* 0 to get default size */

EXTERN long gc_utime,gc_stime,gc_rtime,gc_count;
PRIVATE struct tms start_tms;

EXTERN COUNTER initCommandMode;

#if OS==UNIX_SYS_V || OS==UNIX_BSD
IMPORT char *sbrk P((int incr));
PRIVATE char *start_limit;
#endif

#define __SEG__ INIT

struct init {
  mainAction fun;
  int argc;
  char **argv;
};

PRIVATE NORETURN void lower_init(_data)
IN char *_data;
{
  struct init *data;

  data = (struct init *)_data;
  InitExprs();
  InitGC();
  InitAllocTables();
  InitMap();
  InitSymbols();
  InitLex();
  InitGram();
  InitTypes();
  InitArith();
  InitEval();
  InitCompile();
#ifdef COMPILER
  InitEmit();
#endif
  InitExec();
  InitRandom();
#if DEBUGGING!=0
  InitDebug();
#endif
  InitOpt();
#ifdef CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfputs(coreEnd,coreTraceFile);
      hfputc('\n',coreTraceFile);
      hfflush(coreTraceFile);
    }
#endif
  for (doinit = doafterinit; doinit!=(actions *)0L; doinit = doinit->next)
    (*doinit->exec)();
  if (initCommandMode!=2)
    /* if no external initialization, put initials away */
    {
      initializing = false;
      cutoffInitials();
    }
  (void) times(&start_tms);
  (*data->fun)(data->argc,data->argv);
}

PUBLIC NORETURN void Init(fun)
IN mainAction fun;
{
  long size,replay_begin;
  char line[128],*s;
  int c;
  int intsize,cntsize,ptrsize;
  int Ver,ver,rev;
  char status[128];
  actions *doinit;

  InitHIO();
#ifdef SPECIAL_MALLOC
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  start_limit = sbrk(0);
#endif
  if (replay_size==0)
    {
#if OS==UNIX_SYS_V || OS==UNIX_BSD
      size = MIN_MEMSIZE;
#else
#if OS==AMIGA_OS
      {
	char *var;
	long maxsize;

	var = getenv("SAMML_SIZE");
	if (var==NULL || sscanf(var,"%ld",&size)<=0)
	  size = MEM_SIZE;
	maxsize = AvailMem(MEMF_LARGEST);
	if (size>maxsize)
	  size = maxsize;
      }
#else
      size = MEM_SIZE;
#endif
#endif
    }
  else
    size = replay_size;
#else
  size = -1;
#endif
  replay_begin = 0L;
  intsize = cntsize = ptrsize = 0;
  Ver = ver = rev = 0;
  status[0] = 0;
#ifdef CORE_TRACE
  if (!replay_mode)
    {
      MAKE_CORE_FILE_NAME();
      coreTraceFile = hfopen(coreTraceName,"w");
      if (coreTraceFile==NULL)
	berror(message(133),coreTraceName);
      if (initCommandMode!=1) /* don't print banner if in -cmd mode */
	{
	  hfprintf(hstdout,message(134),coreTraceName);
	}
    }
  else
    {
      coreTraceFile = NULL;
      if (initCommandMode!=1) /* don't print banner if in -cmd mode */
	{
	  hfprintf(hstdout,message(135));
	}
    }
  hfflush(hstdout);
#endif
  if (replay_mode)
    {
      do {
	s = line;
	do {
	  c = hgetc(hstdin);
	  if (c==-1)
	    berror(message(136));
	  if (c=='\n')
	    {
	      *s = 0;
	      break;
	    }
	  *s++ = c;
	  if (s>=line+sizeof(line))
	    {
	      s[-1] = 0;
	      bmsg(message(137));
	      break;
	    }
	} while (true);
	if (sscanf(line,message(138),&replay_size)==1)
	  size = replay_size;
	else if (sscanf(line,message(139),&replay_begin)==1)
	  ;
	else if (sscanf(line,message(140),&intsize)==1)
	  ;
	else if (sscanf(line,message(141),&cntsize)==1)
	  ;
	else if (sscanf(line,message(142),&ptrsize)==1)
	  ;
	else if (sscanf(line,message(143),&Ver,&ver,status,&rev)==4)
	  ;
	else if (!strcmp(line,coreEnd))
	  break;
	else
	  bmsg(message(144),line);
      } while (true);
    }
#ifdef SPECIAL_MALLOC
#if OS==MAC || OS==AMIGA_OS
  for (; (s = (char *)NEWPTR(size))==(char *)0L; size = (7L*size)/8L);
  DISPOSPTR(s,size);
  size -= APPL_SIZE; /* we do this because the Mac loads segment by segment,
			and we don't want the application to fail because
			it cannot load a segment; the Amiga loads hunks,
			too.*/
  if (size<MIN_MEMSIZE)
    berror(message(145));
  theMZone = newMZone(size);
  theMZone->MZ_hook = defaultMemHook;
#else
  for (; size>=MIN_MEMSIZE && (theMZone = newMZone(size))==(MZPtr)0L;
       size = (7L*size)/8L);
  if (size<MIN_MEMSIZE)
    berror(message(145));
#endif
#endif
  if (replay_mode)
    {
      if (replay_size!=0L && size!=replay_size)
	bmsg(message(146),size,replay_size);
#ifdef SPECIAL_MALLOC
      if (replay_begin!=0L && replay_begin!=(long)theMZone)
	bmsg(message(147),(long)theMZone,replay_begin);
      else
	{
#endif
	  if (intsize!=0 && intsize!=sizeof(INT))
	    bmsg(message(148),sizeof(INT),intsize);
	  if (cntsize!=0 && cntsize!=sizeof(INT))
	    bmsg(message(149),sizeof(COUNTER),cntsize);
	  if (ptrsize!=0 && ptrsize!=sizeof(char *))
	    bmsg(message(150),sizeof(char *),ptrsize);
#ifdef SPECIAL_MALLOC
	}
#endif
      if (Ver!=0 && (Ver!=MAJOR_VERSION || ver!=MINOR_VERSION ||
		     rev!=REVISION ||
		     strcmp(status,codeStatus[CODE_STATUS-1])))
	bmsg(message(151),Ver,ver,status,rev,
	     MAJOR_VERSION,MINOR_VERSION,
	     codeStatus[CODE_STATUS-1],
	     REVISION);
    }
#ifdef CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfprintf(coreTraceFile,message(152),size);
#ifdef SPECIAL_MALLOC
      hfprintf(coreTraceFile,message(153),(long)theMZone);
#endif
      hfprintf(coreTraceFile,message(154),sizeof(INT));
      hfprintf(coreTraceFile,message(155),sizeof(COUNTER));
      hfprintf(coreTraceFile,message(156),sizeof(char *));
      hfprintf(coreTraceFile,message(157),MAJOR_VERSION,MINOR_VERSION,
	       codeStatus[CODE_STATUS-1],REVISION);
      hfflush(coreTraceFile);
    }
#endif
  InitStack(lower_init);
}

PRIVATE void ml_done_clean(doend)
actions *doend;
{
  if (doend!=(actions *)0L)
    {
      ml_done_clean(doend->next);
      (*doend->exec)();
    }
}

PUBLIC void showstats(out)
hFILE *out;
{
  struct tms now;
  long utime,stime,rtime;
  EXTENDED cent = 100.0;
  unsigned long space;

  (void) times(&now);
  utime = now.tms_utime-start_tms.tms_utime;
  stime = now.tms_stime-start_tms.tms_stime;
  rtime = utime+stime;
  hfputs("SamML timing statistics:\n",out);
  hfputs("========================\n\n",out);

  hfprintf(out,"total running time: %fs (user: %fs)\n",
	   ((EXTENDED)rtime/((EXTENDED)HZ)),
	   ((EXTENDED)utime)/((EXTENDED)HZ));
  hfprintf(out,"number of garbage collections (GC): %ld\n",gc_count);
  hfprintf(out,"total GC time: %fs (user: %fs)\n",
	   ((EXTENDED)gc_rtime)/((EXTENDED)HZ),
	   ((EXTENDED)gc_utime)/((EXTENDED)HZ));
  hfprintf(out,"GC overhead: %.1f%% (user: %.1f%%)\n",
	   cent*((EXTENDED)gc_rtime)/((EXTENDED)rtime),
	   cent*((EXTENDED)gc_utime)/((EXTENDED)utime));
#ifdef SPECIAL_MALLOC
  space = theMZone->MZ_block.MB_size;
#define SPACE_OK
#else
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  space = sbrk(0)-start_limit;
#define SPACE_OK
#endif
#endif
#ifdef SPACE_OK
  hfprintf(out,"total data space size = %.1f Kb\n",
	   ((EXTENDED)space)/1024.0);
#endif
#undef SPACE_OK
}

PUBLIC NORETURN void ml_done(n)
IN int n;
{
  ml_done_clean(dobeforeend);
#ifdef CORE_TRACE
  if (coreTraceFile!=NULL)
    {
      hfputs("\n(*\n",coreTraceFile);
      showstats(coreTraceFile);
      hfputs("*)\n",coreTraceFile);
      if (initCommandMode!=1) /* don't print banner if in -cmd mode */
	{
	  hfprintf(hstdout,message(158),coreTraceName);
	  hfflush(hstdout);
	}
      hfclose(coreTraceFile);
    }
#endif
  EndRandom();
  EndHIO();
#ifdef SPECIAL_MALLOC
  if (theMZone!=(MZPtr)0L)
    {
      EndMZone(theMZone);
      theMZone = (MZPtr)0;
    }
#endif
  exit(n);
}
