%{
/* HimML scanner.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "abssyn.h"
#include "gram.h"
#include "himml.tab.h"
#include "input.h"
#include "expr.h"
#include "grmproto.h"
#include "memproto.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#include <unistd.h>
#endif
#ifdef HAS_PROTOTYPES
#include "slex.h"
#endif

#define __SEG__ PARSE

#define YY_NO_UNPUT
/* We undefined the following, even though we have to include files (this is why
   scanning buffer is), and to read from strings. The reason is that we had to
   do everything by hand before flex offered these features. Now that flex has them,
   it would be difficult to change this behaviour, coded in the functions OpenFile()
   and CloseFile() below. Admittedly, the latter works only because we have intimate
   knowledge of the internals of the scanner flex generates, and may not work with
   other versions of flex.
*/
#define YY_NO_SCAN_BUFFER
#define YY_NO_SCAN_STRING
#define YY_NO_SCAN_BYTES

#if OS==MAC
extern void flex_mac_init P((void));
#define YY_USER_INIT flex_mac_init()
#endif

#define YY_USER_INIT 
EXTERN hFILE *yyin,*yyout;
EXTERN char *yytext;
EXTERN int yyleng;
EXTERN YYLTYPE yylloc; /* location data for the lookahead */
EXTERN Boolean wantCloseFile;
PUBLIC COUNTER initCommandMode=0;
EXTERN Boolean initializing;

EXTERN ExprPtr boolType,specialMarker;

EXTERN int condState;
EXTERN int condLevel;
EXTERN COUNTER minDebugLevel;

#if CORE_TRACE
EXTERN hFILE *coreTraceFile;
#endif

#define BUMP bump();
#define BUMPCR bumpcr();

PRIVATE void bump()
{
  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  yylloc.last_column += yyleng;
}

PRIVATE void bumpcr()
{
  register char *s,*send;

  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  for (s=yytext, send=s+yyleng; s<send; s++)
    {
      yylloc.last_column++;
      if (*s=='\n')
	{
	  yylloc.last_line++;
	  yylloc.last_column = 0;
	}
    }
}

PUBLIC int yy_input_device=YY_INPUT_STDIN;
PUBLIC CHAR *yyfilename=NULL;
PUBLIC CHAR *yystring=NULL;
PUBLIC long yystrlen=0;
PUBLIC long yyoffset=0;
PRIVATE CHAR *yy_end_buffer;

#if DEBUGGING
PUBLIC ExprPtr currentFile;

PUBLIC void setCurrentFile()
{
  ExprPtr file;
  ENTER;

  file = (yyfilename==NULL)?NIL:Ksymbol(yyfilename);
  SETQ(currentFile,file);
  QUIT;
}
#else
#define setCurrentFile()
#endif

PUBLIC Boolean unit_recognized=false;
PUBLIC COUNTER star_recognized=0,
	union_recognized=0;

PUBLIC ExprPtr lexEnv, /* maps ids to precedences as E_INT */
	consEnv; /* sets of constructors */
#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC ExprPtr strEnv,  /* map from structure names to lexical structures
		    (arrays of [lexMap,consMap,strMap,sigMap])
		    or to lexical functors (couples (lexical signature,lexical structure))
		    lexical signatures are exactly the same as lexical structures */
	sigEnv,keepStrEnv,keepSigEnv;
#endif
PUBLIC ExprPtr keepLexEnv,keepConsEnv;
PUBLIC SymPtr lexId,lexPath;
EXTERN SymPtr starSym,unionSym,concSym;
EXTERN SymPtr nilSym,consSym,refsym,trueSym,falseSym,
	bindSym,matchSym,emptySym,mapgetSym,
	parsweepSym,nomemSym,nonsenseSym,catchSym,
	stackSym,incompSym,noninvSym,nthSym,
	rangeSym,asciiSym,stringnthSym,
	subscriptSym,arithSym,ioSym,reSym,ufindSym,futureSym;

PRIVATE void skipComment P((void));

typedef struct remText {
  struct remText *next;
  int len;
  CHAR text[1];
}   remText;

PRIVATE remText *theRemText=(remText *)0L;

typedef struct fileStack {
  struct fileStack *prev;
  int input_device;
  int saveCondState,saveCondLevel;
  /*
  char *yy_c_buf_p;
  int yy_n_chars;
  */
  union {
    struct {
      hFILE *file;
      CHAR *name;
    } file;
    struct {
      long len;
      CHAR *text;
    } string;
  } what;
  int line,pos;
  long offset;
  remText *rem;
} fileStack;

PRIVATE fileStack *theFileStack=NULL;
void yyinit();

PRIVATE void openRemText1(text,len)
CHAR *text;
int len;
{
  remText *rem;
  
  if (len>0)
    {
      rem = (remText *)MALLOC(sizeof(remText)+len);
      rem->next = theRemText;
      rem->len = len;
      memcpy(rem->text,text,(size_t)len);
      theRemText = rem;
    }
}

PRIVATE void openRemText()
{
  int len; /* really deeply depends on how flex works internally! */
  CHAR *after;

#ifdef OLD
  after = (CHAR *)(yytext+yyleng-1); /* points to the '"' */
  after[0] = ';';
  after[1] = yy_hold_char;
#else
  after = (CHAR *)(yytext+yyleng); /* points to the NUL */
  after[0] = yy_hold_char;
#endif
  len = yy_end_buffer - after;
  openRemText1(after,len);
}

PRIVATE void closeRemText()
{
  remText *rem;

  if (theRemText!=(remText *)0L)
    {
      rem = theRemText->next;
      MARKFREE(theRemText);
      theRemText = rem;
    }
}

PRIVATE void resetRemText()
{
  while (theRemText!=(remText *)0L)
    closeRemText();
}

PUBLIC Boolean PushFile()
{
  fileStack *next;
  
  openRemText();
  next = (fileStack *)MALLOC_FAIL(sizeof(fileStack));
  if (next==(fileStack *)0L)
    return false;
  next->prev = theFileStack;
  theFileStack = next;
  next->input_device = yy_input_device;
  next->line = yylloc.last_line;
  next->pos = yylloc.last_column;
  next->offset = yyoffset;
  /*
  *yy_c_buf_p = yy_hold_char;
  next->yy_c_buf_p = yy_c_buf_p;
  next->yy_n_chars = yy_n_chars;
  */
  yylloc.last_line = 0;
  yylloc.last_column = 0;
  yyinit();
  yyoffset = 0;
  next->saveCondState = condState;
  condState = 0;
  next->saveCondLevel = condLevel;
  condLevel = 0;
  switch (next->input_device)
    {
    case YY_INPUT_FILE:
    case YY_INPUT_STDIN:
      next->what.file.file = yyin;
      next->what.file.name = yyfilename;
      break;
    case YY_INPUT_STRING0:
    case YY_INPUT_STRING:
      next->what.string.text = yystring;
      next->what.string.len = yystrlen;
      break;
    }
  next->rem = theRemText;
  theRemText = (remText *)0L;
  return true;
}

PUBLIC Boolean OpenFile(name)
IN CONST CHAR *name;
{
  hFILE *new;
  CHAR *realname;
  
  new = xfopen((char *)name,"r",(char **)&realname);
  if (new==NULL)
    return false;
  if (!PushFile())
    {
      MARKFREE(realname);
      return false;
    }
  yy_input_device = YY_INPUT_FILE;
  yyin = yy_current_buffer->yy_input_file = new;
  yyfilename = realname;
  setCurrentFile();
  openRemText1(";",1); /* to end the declarations after 'use' */
  return true;
}

PUBLIC Boolean OpenString(str,len)
IN CONST CHAR *str;
IN COUNTER len;
{
  if (!PushFile())
    return false;
  yy_input_device = YY_INPUT_STRING;
  yystring = str;
  yystrlen = len;
  return true;
}

PUBLIC void CloseFile()
{
  fileStack *prev;

  resetRemText();
  if (theFileStack==NULL)
    switch (initCommandMode)
      {
      case 0: /* happens when doing himml <file or ... | himml: exit */
      case 1: /* command file (-cmd) option was given: exit after execution */
	ml_done(0);
        break;
      case 2: /* init file (-i) option was given: switch to stdin */
	yy_input_device = YY_INPUT_STDIN;
	initCommandMode = 0;
	yylloc.last_line = 0;
	yylloc.last_column = 0;
	yyoffset = 0;
	yyin = yy_current_buffer->yy_input_file = hstdin;
	yyfilename = NULL;
	setCurrentFile();
	initializing  = false;
	cutoffInitials(); /* now put away initial values */
	break;
      case 3: /* compilation option (-c) was given: compile file compile_file */
	initCommandMode = 4;
        ml_done (HimMLCompile ()?0:10);
        break;
      case 4: /* happens when -c option was given but end-of-file was reached during
		 compilation; then bison tries to synchronize on semicolons, and
		 each synchronization calls CloseFile().  Then set wantCloseFile back
		 to true; this will be noticed by yywrap(), which will then issue
		 an 'unexpected end of file' error message. */
	wantCloseFile = true;
	break;
      }
  else
    {
      /*
      yy_c_buf_p = theFileStack->yy_c_buf_p;
      yy_n_chars = theFileStack->yy_n_chars;
      */
      if (yy_input_device==YY_INPUT_FILE)
        {
	  hfclose(yyin);
	  if (yyfilename!=NULL)
	    MARKFREE(yyfilename);
	}
      yylloc.last_line = theFileStack->line;
      yylloc.last_column = theFileStack->pos;
      yyoffset = theFileStack->offset;
      switch (yy_input_device = theFileStack->input_device)
	{
	case YY_INPUT_FILE:
	case YY_INPUT_STDIN:
	  yyin = yy_current_buffer->yy_input_file = theFileStack->what.file.file;
	  yyfilename = theFileStack->what.file.name;
	  setCurrentFile();
	  break;
	case YY_INPUT_STRING0:
	case YY_INPUT_STRING:
	  yystring = theFileStack->what.string.text;
	  yystrlen = theFileStack->what.string.len;
	  yyfilename = NULL;
	  setCurrentFile();
	  break;
	}
      prev = theFileStack->prev;
      theRemText = theFileStack->rem;
      MARKFREE(theFileStack);
      theFileStack = prev;
    }
  yyinit();
  /* unput(';'); to end list of declarations at end of file,
      but does not work. This is taken care of by the
      *after = ';' line in openRemText() */
}

PUBLIC char *getMarkFiles()
{
  return (char *)theFileStack;
}

PUBLIC void ResetFiles(mark,flushRemText)
IN CONST char *mark;
IN Boolean flushRemText;
{
  fileStack *fs;

  for (fs=theFileStack; mark!=(char *)fs; fs=fs->prev)
    if (fs==NULL) /* mark not found */
      goto end;
  while (mark!=(char *)theFileStack && theFileStack!=NULL) /*yy_input_device!=YY_INPUT_STDIN*/
    CloseFile();
end:
  if (flushRemText)
    resetRemText();
  wantCloseFile = false;
}

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) \
	if (theRemText!=(remText *)0L) \
	{ \
	if (max_size>=theRemText->len) \
	{ \
	result = theRemText->len; \
	memcpy(buf,(char *)theRemText->text,(size_t)result); \
	closeRemText(); \
	} \
	else \
	{ \
	result = max_size; \
	memcpy(buf,(char *)theRemText->text,(size_t)result); \
	theRemText->len -= result; \
	memcpy((char *)theRemText->text,(char *)theRemText->text+result, \
	(size_t)theRemText->len); \
	} \
	} \
	else { \
	switch (yy_input_device) \
	{ \
	case YY_INPUT_FILE: \
	  if ( (result = read(yyin->fildesc, buf, max_size )) < 0 ) \
	     yy_error(yylloc.first_line,yylloc.first_column, \
		     yylloc.last_line,yylloc.last_column, \
		     message(242)); \
	  break; \
	case YY_INPUT_STDIN: \
	  { \
		register int c; \
		register char *s; \
		for (result = 0, s = buf; result<max_size; result++) \
		{ \
		c = hgetc(hstdin); \
		if (c==EOF) \
		  break; \
		*s++ = c; \
		if (c=='\n') \
		  { \
			result++; \
			break; \
		  } \
		} \
		break; \
	  } \
	case YY_INPUT_STRING0: \
	case YY_INPUT_STRING: \
	  result = yystrlen-yyoffset; \
	  if (result>max_size) \
		result = max_size; \
	  memcpy(buf,(char *)(yystring+yyoffset),result); \
	  break; \
	} \
	yyoffset += result; \
	} \
	yy_end_buffer = (CHAR *)((buf)+(result));

#undef YY_FATAL_ERROR
#define YY_FATAL_ERROR(msg) berror(msg)

#undef yywrap
#define YY_SKIP_YYWRAP
int yywrap()
{
  if (condState!=0 || condLevel!=0 || YY_START!=INITIAL)
    {
      condState = condLevel = 0;
      BEGIN(INITIAL);
      yy_error(yylloc.first_line,yylloc.first_column,
	       yylloc.last_line,yylloc.last_column,
	       message(567));
    }
  if (wantCloseFile)
    {
      wantCloseFile = false;
      yy_error(yylloc.first_line,yylloc.first_column,
	       yylloc.last_line,yylloc.last_column,
	       message(129));
    }
  wantCloseFile = true;
  openRemText1(";",1);
  return 0; /* continue */
}

PUBLIC void pushEnv()
{
  ExprPtr aux;
  ENTER;

  MAKECONS1(NewMap(),lexEnv,aux);
  lexEnv = aux;
  MAKECONS1(NewMap(),consEnv,aux);
  consEnv = aux;
#if MODULE_SYSTEM & STD_ML_MODULE
  MAKECONS1(NewMap(),strEnv,aux);
  strEnv = aux;
  MAKECONS1(NewMap(),sigEnv,aux);
  sigEnv = aux;
#endif
  QUIT;
}

PUBLIC void popEnv()
{
  POPQ(lexEnv);
  POPQ(consEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  POPQ(strEnv);
  POPQ(sigEnv);
#endif
}

PUBLIC void setLocalEnv()
{
  ExprPtr ref;

  ENTER;
  ref = CDR(CDR(lexEnv));
  setcarMapMerge(ref,CAR(lexEnv));
  ref = CDR(CDR(consEnv));
  setcarMapMerge(ref,CAR(consEnv));
#if MODULE_SYSTEM & STD_ML_MODULE
  ref = CDR(CDR(strEnv));
  setcarMapMerge(ref,CAR(strEnv));
  ref = CDR(CDR(sigEnv));
  setcarMapMerge(ref,CAR(sigEnv));
#endif
  QUIT;
}

#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC void keepEnv()
{
  SETQ(keepLexEnv,lexEnv);
  SETQ(keepConsEnv,consEnv);
  SETQ(keepStrEnv,strEnv);
  SETQ(keepSigEnv,sigEnv);
}

PUBLIC void restoreEnv()
{
  SETQ(lexEnv,keepLexEnv);
  SETQ(consEnv,keepConsEnv);
#if MODULE_SYSTEM & STD_ML_MODULE
  SETQ(strEnv,keepStrEnv);
  SETQ(sigEnv,keepSigEnv);
#endif
  SETQ(keepLexEnv,NIL);
  SETQ(keepConsEnv,NIL);
  SETQ(keepStrEnv,NIL);
  SETQ(keepSigEnv,NIL);
}
#endif

PUBLIC void setTopEnv()
{
  while (CDR(lexEnv)!=NIL)
    {
      POPQ(lexEnv);
    }
  while (CDR(consEnv)!=NIL)
    {
      POPQ(consEnv);
    }
#if MODULE_SYSTEM & STD_ML_MODULE
  while (CDR(strEnv)!=NIL)
    {
      POPQ(strEnv);
    }
  while (CDR(sigEnv)!=NIL)
    {
      POPQ(sigEnv);
    }
#endif
  SETQ(keepLexEnv,NIL);
  SETQ(keepConsEnv,NIL);
#if MODULE_SYSTEM & STD_ML_MODULE
  SETQ(keepStrEnv,NIL);
  SETQ(keepSigEnv,NIL);
#endif
}

PRIVATE ExprPtr copyEnv1(env)
IN CONST ExprPtr env;
{
  env = Nreverse(env);
  return Nreverse(env);
}

PUBLIC ExprPtr copyLexEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(lexEnv);
  KEEPONLY(e);
  return e;
}

PUBLIC ExprPtr copyConsEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(consEnv);
  KEEPONLY(e);
  return e;
}

#if MODULE_SYSTEM & STD_ML_MODULE
PUBLIC ExprPtr copyStrEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(strEnv);
  KEEPONLY(e);
  return e;
}

PUBLIC ExprPtr copySigEnv()
{
  ExprPtr e;
  ENTER;

  e = copyEnv1(sigEnv);
  KEEPONLY(e);
  return e;
}
#endif

PRIVATE Boolean getConstructor(name)
IN CONST ExprPtr name;
{
  ExprPtr consE;
  Boolean found;
  
  for (found = false, consE=consEnv; consE!=NIL; consE=CDR(consE))
    {
      (void) MapGet(name,CAR(consE),&found);
      if (found)
	break;
    }
  return found;
}

PUBLIC void setConstructor(name)
IN CONST ExprPtr name;
{
  setcarMapPut(name,UNIT,consEnv);
}

PUBLIC int id_class(name)
IN CONST ExprPtr name;
{
  ExprPtr prec,lexE;
  Boolean found;
  Boolean cc;
  
  cc = getConstructor(name);
  for (lexE = lexEnv; lexE!=NIL; lexE=CDR(lexE))
    {
      prec = MapGet(name,CAR(lexE),&found);
      if (found)
	{
	  switch (IVAL(prec))
	    {
	    case 1:  return cc?kw_left_0_constructor:kw_left_0;
	    case -1: return cc?kw_right_0_constructor:kw_right_0;
	    case 2:  return cc?kw_left_1_constructor:kw_left_1;
	    case -2: return cc?kw_right_1_constructor:kw_right_1;
	    case 3:  return cc?kw_left_2_constructor:kw_left_2;
	    case -3: return cc?kw_right_2_constructor:kw_right_2;
	    case 4:  return cc?kw_left_3_constructor:kw_left_3;
	    case -4: return cc?kw_right_3_constructor:kw_right_3;
	    case 5:  return cc?kw_left_4_constructor:kw_left_4;
	    case -5: return cc?kw_right_4_constructor:kw_right_4;
	    case 6:  return cc?kw_left_5_constructor:kw_left_5;
	    case -6: return cc?kw_right_5_constructor:kw_right_5;
	    case 7:  return cc?kw_left_6_constructor:kw_left_6;
	    case -7: return cc?kw_right_6_constructor:kw_right_6;
	    case 8:  return cc?kw_left_7_constructor:kw_left_7;
	    case -8: return cc?kw_right_7_constructor:kw_right_7;
	    case 9:  return cc?kw_left_8_constructor:kw_left_8;
	    case -9: return cc?kw_right_8_constructor:kw_right_8;
	    case 10: return cc?kw_left_9_constructor:kw_left_9;
	    case -10:return cc?kw_right_9_constructor:kw_right_9;
	    }
	  break;
	}
    }
  return cc?kw_cons_id:kw_identifier;
}

PRIVATE int longid_class(text,op)
IN CONST CHAR *text;
IN Boolean op;
{
  REGISTER CHAR *s,*s0,*t,c;
  SymPtr strid;
  ExprPtr aux,path;
  int code;
  ENTER;
  
  path = NIL;
  keepEnv();
  s0 = text;
  while (true)
    {
      s = s0;
      while (c = *s, c!=0 && c!=' ' && c!='\t' && c!='\n' && c!='\v' && c!='\f' && c!='.')
	s++;
      for (t=s; (c = *t); t++)
	if (c==0 || c=='.')
	  break;
      if (c==0)
	break;
      strid = Kstring((INT)(s-s0),s0);
      (void) KEEP(strid);
      MAKECONS(strid,path,aux);
      path = KEEP(aux);
      lexOpen(&yylloc,strid,false,strEnv);
      for (s0=t+1; (c = *s0); s0++)
	if (c!=' ' && c!='\t' && c!='\n' && c!='\v' && c!='\f')
	  break;
    }
  SETQ(lexPath,Nreverse(path));
  aux = Ksymbol(s0);
  SETQ(lexId,aux);
  code = op?(getConstructor(lexId)?kw_cons_id:kw_identifier):id_class(lexId);
  restoreEnv();
  QUIT;
  return code;
}

PRIVATE CHAR *skipSpaces(text)
IN REGISTER CONST CHAR *text;
{
  REGISTER CHAR c;

  while (c = *text++, c==' ' || c=='\t' || c=='\n' || c=='\v' || c=='\f');
  return text-1;
}

#if IMPERATIVE_TYPES==STD_ML_TYVAR
PRIVATE void checkNoLevelType(s)
IN CONST CHAR *s;
{
  while (*s++ == '\'');
  if (*s>='1' && *s<='9')
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	hfputs(yytext,coreTraceFile);
#endif
      yy_error(yylloc.first_line,yylloc.first_column,
	       yylloc.last_line,yylloc.last_column,
	       message(130));
    }
}
#endif

#if FLOAT_FORMAT==IEEE754
#define IEEE_LEX(n) n
#else
PRIVATE NORETURN int infError()
{
#if CORE_TRACE
  if (coreTraceFile!=NULL)
    hfputs(yytext,coreTraceFile);
#endif
  yy_error(yylloc.first_line,yylloc.first_column,
	   yylloc.last_line,yylloc.last_column,
	   message(131));
}

#define IEEE_LEX(n) infError()
#endif

%}

SPACE [\ \t\f]
DIG [0-9]
INT (\~?{DIG}+)
FMT [\ \t\n\f]
LETTER [A-Za-z]
ALPHA [A-Za-z0-9\'\_]
SYMB [\!\%\&\$\#\+\-\/\:\<\=\>\?\@\\\~\'\^\|\*\`]
ALPHA_ID ({LETTER}{ALPHA}*)
SYMB_ID ({SYMB}+)
ID ({SYMB_ID}|{ALPHA_ID})
XID ({ID}|\_{ID})
PREFIX ({XID}{FMT}*\.{FMT}*)

%x iffalse
%x ifdone

%%

abstraction { BUMP return kw_abstraction; }
abstype { BUMP return kw_abstype; }
all	{ BUMP return kw_all; }
and { BUMP return kw_and; }
andalso	{ BUMP return kw_andalso; }
as { BUMP return kw_as; }
case { BUMP return kw_case; }
do	{ BUMP return kw_do; }
datatype { BUMP return kw_datatype; }
delay	{ BUMP return kw_delay; }
else { BUMP return kw_else; }
end { BUMP return kw_end; }
eqtype { BUMP return kw_eqtype; }
exception { BUMP return kw_exception; }
exists { BUMP return kw_exists; }
extern { BUMP return kw_extern; }
fn { BUMP return kw_fn; }
fun { BUMP return kw_fun; }
functor { BUMP return kw_functor; }
handle { BUMP return kw_handle; }
if { BUMP return kw_if; }
in { BUMP return kw_in; }
in{SPACE}list/{FMT}	{ BUMPCR return kw_in_list; }
in{SPACE}map/{FMT}	{ BUMPCR return kw_in_map; }
in{SPACE}set/{FMT}	{ BUMPCR return kw_in_set; }
include { BUMP return kw_include; }
infix	{ BUMP return kw_infix; }
infixr { BUMP return kw_infixr; }
iterate { BUMP return kw_iterate; }
let { BUMP return kw_let; }
local	{ BUMP return kw_local; }
memofn { BUMP return kw_memo_fn; }
memofun { BUMP return kw_memo_fun; }
nonfix	{ BUMP return kw_nonfix; }
of { BUMP return kw_of; }
open { BUMP return kw_open; }
open\* { BUMP return kw_openx; }
orelse { BUMP return kw_orelse; }
pack	{ BUMP return kw_pack; }
dimension	{ BUMP return kw_dimension; }
raise { BUMP return kw_raise; }
rec	{ BUMP return kw_rec; }
sharing { BUMP return kw_sharing; }
sig { BUMP return kw_sig; }
signature { BUMP return kw_signature; }
some { BUMP return kw_some; }
struct { BUMP return kw_struct; }
structure { BUMP return kw_structure; }
sub{SPACE}map/{FMT}	{ BUMPCR return kw_sub_map; }
such{SPACE}that/{FMT}  { BUMPCR return kw_such_that; }
then { BUMP return kw_then; }
type { BUMP return kw_type; }
scale { BUMP return kw_scale; }
val	{ BUMP return kw_val; }
^#{SPACE}*compile { BUMP return kw_compile; }
^#{SPACE}*if{SPACE}[^\n]*\n	{
	BUMPCR
 testcond:
	{
	  char *s=yytext+1; /* skip '#' */
	  ExprPtr val;

	  while (*s++!='i'); /* wait for 'i' */
	  s++;
	  val = readValueSimple(s,boolType,basisEnv);
	  condState++; /* counts the number of enclosing #if's */
	  /* condLevel counts the number of enclosing #if's while skipping. */
	  if (val!=specialMarker && PACKVAL(val)==trueValue)
	    condLevel = 0;
	  else { condLevel = 1; BEGIN(iffalse); } /* look for #else, #elsif or #endif */
	}
}
^#{SPACE}*elsif{SPACE}[^\n]*\n	|
^#{SPACE}*else{SPACE}?[^\n]*\n	{
	BUMPCR
	if (condState==0) /* no #if before */
	  {
 errcond:
	    yy_error(yylloc.first_line,yylloc.first_column,
		     yylloc.last_line,yylloc.last_column,
		     message(566));
	  }
	condLevel = 1;
	BEGIN(ifdone); /* skip past the closing #endif */
}
^#{SPACE}*endif{SPACE}?[^\n]*\n	{
	BUMPCR
	if (condState==0) /* no #if before */
	  goto errcond;
	condState--;
}
<iffalse,ifdone>^#{SPACE}*if{SPACE}[^\n]*\n	{
	BUMPCR
	condLevel++;
}
<iffalse>^#{SPACE}*elsif{SPACE}[^\n]*\n	{
	BUMPCR
	if (condLevel==1) /* looking for #elsif or #else */
	  {
	    condLevel = 0;
	    condState--;
	    BEGIN(INITIAL);
	    goto testcond;
	  }
}
<iffalse>^#{SPACE}*else{SPACE}?[^\n]*\n	{
	BUMPCR
	if (condLevel==1) /* looking for #elsif or #else */
	  {
	    condLevel = 0;
	    BEGIN(INITIAL);
	  }
}
<iffalse,ifdone>^#{SPACE}*endif{SPACE}?[^\n]*\n	{
	BUMPCR
	if (--condLevel == 0)
          { condState--; BEGIN(INITIAL); }
}
while { BUMP return kw_while; }
with	{ BUMP return kw_with; }
withtype { BUMP return kw_withtype; }
use { BUMP return kw_use; }
U	{ BUMP if (union_recognized)
     return kw_union;
     SETQ(lexId,unionSym);
     SETQ(lexPath,NIL);
     return id_class(lexId); }
\(	{ BUMP return kw_open_paren; }
\)	{ BUMP return kw_close_paren; }
\[	{ BUMP return kw_open_bracket; }
\]	{ BUMP return kw_close_bracket; }
\.\(	{ BUMP return kw_open_array_ref; }
\<\{	{ BUMP return kw_open_brace_under; }
\{	{ BUMP return kw_open_brace; }
\}	{ BUMP return kw_close_brace; }
\|\[	{ BUMP return kw_open_record; }
\]\|	{ BUMP return kw_close_record; }
\[\|	{ BUMP return kw_open_array; }
\|\]	{ BUMP return kw_close_array; }
\+\+\|\[	{ BUMPCR return kw_open_mu; }
\,	{ BUMP return kw_comma; }
\:	{ BUMP return kw_colon; }
\;+	{ BUMP return kw_semicolon; }
\.\.\.	{ BUMP return kw_ellipsis; }
\_	{ BUMP return kw_underscore; }
\|	{ BUMP return kw_vbar; }
\|\|	{ BUMP return kw_vbar_vbar; }
\=	{ BUMP return kw_equal; }
\.\:\=	{ BUMP return kw_array_assign; }
\=\>	{ BUMP return kw_big_arrow; }
\-\>	{ BUMP return kw_arrow; }
\-m\>	{ BUMP return kw_map_arrow; }
\*		{ BUMP
		    if (star_recognized)
			return kw_star;
			SETQ(lexId,starSym);
			SETQ(lexPath,NIL);
			return id_class(lexId); }
\`		{ BUMP if (star_recognized || unit_recognized)
			  return kw_scale_star;
			goto symbid; }
\^\~		{ BUMP if (star_recognized || unit_recognized)
			  { /* writing m^~2 instead of m^ ~2 is common,
				and should be detected. */
			    yyless(1); /* recognize only \^ */
			    return kw_scale_power;
			  }
			goto symbid; }
\^		{ BUMP if (star_recognized || unit_recognized)
					return kw_scale_power;
				SETQ(lexId,concSym);
				SETQ(lexPath,NIL);
				return id_class(lexId); }
\#	{ BUMP return kw_sharp; }
\(\*\$[DPEV][+-]\*\)	{ BUMP return kw_pragma; }
<INITIAL,iffalse>\(\*	{
	BUMP
#if CORE_TRACE
	if (coreTraceFile!=NULL)
	  hfputs(yytext,coreTraceFile);
#endif
	skipComment();
#if CORE_TRACE
	*yytext = 0;
#endif
	}

{DIG}	{ BUMP return precedence_constant; }
{INT}	{ BUMP return integer_constant; }
\~?0[xX][0-9a-fA-F]+ { BUMP return hex_integer_constant; }
\~?0[oO][0-7]+ { BUMP return octal_integer_constant; }
\~?0[bB][01]+ { BUMP return binary_integer_constant; }
{INT}E{INT} { BUMP return real_constant; }
{INT}\.{DIG}+   { BUMP return real_constant; }
{INT}\.{DIG}+E{INT} { BUMP return real_constant; }
\+inf { BUMP return IEEE_LEX(kw_pinf); }
\~inf { BUMP return IEEE_LEX(kw_minf); }
NAN\(0\)	{ BUMP return IEEE_LEX(kw_nan0); }
NAN\(1\)	{ BUMP return IEEE_LEX(kw_nan1); }
NAN\(2\)	{ BUMP return IEEE_LEX(kw_nan2); }
NAN\(3\)	{ BUMP return IEEE_LEX(kw_nan3); }
NAN\(4\)	{ BUMP return IEEE_LEX(kw_nan4); }
NAN\(5\)	{ BUMP return IEEE_LEX(kw_nan5); }
NAN\(6\)	{ BUMP return IEEE_LEX(kw_nan6); }
NAN\(7\)	{ BUMP return IEEE_LEX(kw_nan7); }
\~NAN\(1\)  { BUMP return IEEE_LEX(kw_mnan1); }
\~NAN\(2\)  { BUMP return IEEE_LEX(kw_mnan2); }
\~NAN\(3\)  { BUMP return IEEE_LEX(kw_mnan3); }
\~NAN\(4\)  { BUMP return IEEE_LEX(kw_mnan4); }
\~NAN\(5\)  { BUMP return IEEE_LEX(kw_mnan5); }
\~NAN\(6\)  { BUMP return IEEE_LEX(kw_mnan6); }
\~NAN\(7\)  { BUMP return IEEE_LEX(kw_mnan7); }
{INT}(\.{DIG}+)?(E{INT})?{SPACE}*\:{SPACE}*{INT}(\.{DIG}+)?(E{INT})?	{
	BUMP return complex_constant; }
\"([^\"\\]|(\\([\"\\ntvf]|(\^[\@A-Z\[\\\]\^\_])|({DIG}{DIG}{DIG})|({FMT}*\\))))*\"  {
	BUMPCR return string_constant; }
{PREFIX}*{XID}  { BUMPCR
 symbid:
	return longid_class(yytext,false); }
op{SPACE}+{PREFIX}*{XID}	{ BUMPCR
	return longid_class(skipSpaces((CHAR *)yytext+2),true); }
\'\'?(\_|[1-9]{DIG}*)?\#?{ALPHA_ID} { BUMP
#if IMPERATIVE_TYPES==STD_ML_TYVAR
	checkNoLevelType(yytext);
#endif
	{
	  ExprPtr aux;

	  aux = Ksymbol(yytext);
	  SETQ(lexId,aux);
	  SETQ(lexPath,NIL);
	}
	return kw_type_variable; }
{SPACE}+	{ BUMP
#if CORE_TRACE
	if (coreTraceFile!=NULL)
	  hfputc(' ',coreTraceFile);
#endif
	}
<*>\n+	{ BUMPCR
#if CORE_TRACE
	if (coreTraceFile!=NULL)
	  hfputc('\n',coreTraceFile);
#endif
	}
<iffalse,ifdone>.	{ BUMP }
.	{ CHAR buf[80];
	  sprintf((char *)buf,message(564),yytext[0]);
	  yy_error(yylloc.first_line,yylloc.first_column,
		   yylloc.last_line,yylloc.last_column,
		   buf);
	}

%%

PUBLIC void yyinit()
{
  YY_FLUSH_BUFFER; /* YY_INIT */
}

enum commentState { SKIPPING, AFTER_OPEN, AFTER_STAR };

PRIVATE void skipComment()
{
  int c;
  int level,state;

  level = 1;
  state = SKIPPING;
  yylloc.first_line = yylloc.last_line;
  yylloc.first_column = yylloc.last_column;
  while (c=input(), c!=-1)
    {
#if CORE_TRACE
      if (coreTraceFile!=NULL)
	{
	  hputc(c,coreTraceFile);
	}
#endif
      if (c=='\n')
	{
	  yylloc.last_line++;
	  yylloc.last_column = 0;
	}
      else
	yylloc.last_column++;
      switch (state)
	{
	case SKIPPING:
	  switch (c)
	    {
	    case '(':
	      state = AFTER_OPEN;
	      break;
	    case '*':
	      state = AFTER_STAR;
	      break;
	    default:
	      break;
	    }
	  break;
	case AFTER_OPEN:
	  if (c=='*')
	    level++;
	  state = SKIPPING;
	  break;
	case AFTER_STAR:
	  if (c==')')
	    if (--level==0)
	      return;
	  if (c!='*')
	    state = SKIPPING;
	  break;
	}
    }
  yy_error(yylloc.first_line,yylloc.first_column,
	   yylloc.last_line,yylloc.last_column,
	   message(132));
}

#if CORE_TRACE
PRIVATE int coreState=0;

PUBLIC void resetTracingYyLex()
{
  coreState = 0;
}

PUBLIC int tracingYylex()
{
  int n;

  n = yylex();
  if (coreTraceFile!=NULL)
    {
      if (n==kw_use)
        coreState = 1;
      else if ((n==string_constant && coreState==1) || coreState==2)
        coreState = 0;
      else
        hfputs(yytext,coreTraceFile);
    }
  return n;
}
#endif
