/* HimML main().
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2003 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "portable.h"
#include "eval.h"
#include "evlproto.h"
#include "errproto.h"
#include "grmproto.h"
#include "expproto.h"
#include "stkproto.h"
#include "prtproto.h"
#if OS==MAC
#ifdef MPW
#include <memory.h>
#include <segload.h>
#endif
#endif
#include "input.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <signal.h>
#endif
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef THINK_C
#include <console.h>
#endif
#ifdef HAS_PROTOTYPES
#include "sml.h"
#endif

#if OS==AMIGA_OS
long __stack = 10000L; /* minimum of stack space */
#include <signal.h>
#endif

#if OS==UNIX_SYS_V || OS==UNIX_BSD
#define OPT_EXPANDED
#else
#ifdef THINK_C
#define OPT_EXPANDED
#else
#undef OPT_EXPANDED
#endif
#endif

EXTERN COUNTER crtab_maxlines;

#ifdef CORE_TRACE
EXTERN char coreTraceName[];
EXTERN hFILE *coreTraceFile;
EXTERN Boolean replay_mode;
#endif

EXTERN struct Expression *features,*featuresType;
EXTERN char maintenance[],creator[];
EXTERN long replay_size;
EXTERN int yy_input_device;
EXTERN CHAR *yystring;
EXTERN long yystrlen;
EXTERN long yyoffset;
EXTERN COUNTER theRmargin;

EXTERN COUNTER initCommandMode;

EXTERN EXTENDED fone;

EXTERN struct Expression *nihil;

#if DEBUGGING!=0
EXTERN Boolean debugMode;
#endif

IMPORT void Init P((void));

struct Expression *MLParse(data)
char *data;
{
  if (yy_input_device==YY_INPUT_STDIN)
    {
#if DEBUGGING!=0
      hfputs(debugMode?message(90):message(91),hstdout);
#else
      hfputs(message(91),hstdout);
#endif
      hfflush(hstdout);
    }
  (void) yyparse();
  return nihil;
}

int theArgc,theArgi;
char **theArgv;

#define MAX_PATHS 50
int thePathc = 0;
char *thePathv[MAX_PATHS];

PRIVATE void getopt P((int argc,char **argv));
PRIVATE void usage P((IN Boolean fatalp));

main(argc,argv)
int argc;
char *argv[];
{
  int i;

#ifdef THINK_C
  argc = ccommand(&argv);
#endif
#if OS==AMIGA_OS
  argv[0] = "samml";
#endif
#if OS==MAC
  MaxApplZone();
#endif
  getopt(argc,argv);
  Init();
#if OS==MAC
  UnloadSeg(Init);
#if DEBUGGING!=0
  UnloadSeg(InitDebug);
#endif
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  signal(2,ml_break);
  signal(3,ml_done);
  signal(SIGTERM,ml_done);
#ifdef SIGFPE
  signal(SIGFPE,ml_fpe);
#endif
#endif
  if (initCommandMode!=1) /* don't print banner if in -cmd mode */
    {
#if (CODE_STATUS!=STATUS_RELEASE)
      hfprintf(hstdout,message(92),maintenance);
#ifdef CORE_TRACE
      if (coreTraceFile!=NULL)
	hfprintf(hstdout,message(93),coreTraceName);
#endif
      hfputs(message(94),hstdout);
      crtab_maxlines = MAX_SIGNED_INT;
      (void) valPretty(features,featuresType,do_printfile,
		       (char *)hstdout,0,11,theRmargin);
      hputc('\n',hstdout);
      hputc('\n',hstdout);
#endif
      hfprintf(hstdout,message(95),OS_STRING,creator);
      hfflush(hstdout);
    }
  for (i=0; i<thePathc; i++)
    add_path(thePathv[i],strlen(thePathv[i]));
  for (;;)
    (void) doML(MLParse,(char *)0L);
  /*ml_done(0);*/
}

typedef struct Command {
  char *key;
  void (*do_) P((void));
  char *doc;
  long xmask; /* the command (number #j) cannot be used if the command #i
		    has been used before, for 1<<i in xmask;
		    in particular, if j=i, the command may be reused
		    only if 1<<i is not in xmask */
}   Command;

PRIVATE void do_initreplay()
{
  char *replay_file;

  if (theArgi>=theArgc)
    bfatal(message(96));
  replay_file = theArgv[theArgi++];
  if (freopen(replay_file,"r",stdin)==NULL)
    /* this is used before Init, so we cannot use hfreopen on hstdin,
       but we can use the usual I/O library */
    bfatal(message(97),replay_file);
  replay_mode = true;
}

PRIVATE void do_initmemsize()
{
  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%ld",&replay_size)!=1)
    bfatal(message(98));
}

PRIVATE void do_initcommand()
{
  CHAR *s;

  if (theArgi>=theArgc
#ifndef OPT_EXPANDED
       || (s = (CHAR *)theArgv[theArgi], *s!='\"' || s[strlen((char *)s)-1]!='\"')
#endif
       )
    bfatal(message(99));
#ifndef OPT_EXPANDED
  (void) getString(NULL,NULL,s,false); /* on non-unix systems, the shell does not expand strings */
#else
  s = (CHAR *)theArgv[theArgi];
#endif
  initCommandMode = 1;
  yy_input_device = YY_INPUT_STRING;
  yystring = s;
  yystrlen = (long)strlen((char *)s);
  yyoffset = 0L;
  theArgi++;
}

PRIVATE void do_initinit()
{
  CHAR *s;

  if (theArgi>=theArgc
#ifndef OPT_EXPANDED
       || (s = (CHAR *)theArgv[theArgi], *s!='\"' || s[strlen((char *)s)-1]!='\"')
#endif
       )
    bfatal(message(100));
#ifndef OPT_EXPANDED
  (void) getString(NULL,NULL,s,false); /* on non-unix systems, the shell does not expand strings */
#else
  s = (CHAR *)theArgv[theArgi];
#endif
  initCommandMode = 2;
  yy_input_device = YY_INPUT_STRING;
  yystring = s;
  yystrlen = (long)strlen((char *)s);
  yyoffset = 0L;
  theArgi++;
}

PRIVATE void do_initusage()
{
  usage(false);
  ml_done(0);
}

PRIVATE void do_addpath()
{
  if (theArgi>=theArgc)
    bfatal(message(101));
  if (thePathc>=MAX_PATHS-1)
    bfatal(message(102),MAX_PATHS);
  thePathv[thePathc++] = theArgv[theArgi++];
  /* we cannot use add_path() directly because it uses Malloc(),
     but theMZone has not yet been initialized */
}

PRIVATE void do_rmargin()
{
  int n;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%d",&n)!=1)
    bfatal(message(103));
  theRmargin = (COUNTER)n;
}

PRIVATE void do_growfactor()
{
  double grow;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%lf",&grow)!=1 || grow<=fone)
    bfatal(message(104));
  installGrowFactor(&grow);
}

PRIVATE void do_maxgrowfactor()
{
  double grow;

  if (theArgi>=theArgc || sscanf(theArgv[theArgi++],"%lf",&grow)!=1 || grow<=fone)
    bfatal(message(104));
  installMaxGrowFactor(&grow);
}

PRIVATE Command command[] = {
/* 0 */ { "-replay", do_initreplay, " replay-file", 0xffffffffL, },
/* 1 */ { "-mem", do_initmemsize, " memory-size", 0x00000003L },
/* 2 */ { "-cmd", do_initcommand, " ML-command-string", 0x0000000dL },
/* 3 */ { "-init", do_initinit, " ML-init-string", 0x0000000dL },
/* 4 */ { "-path", do_addpath, " path", 0x00000001L },
/* 5 */	{ "-col", do_rmargin, " number-of-columns", 0x00000021L },
/* 6 */	{ "-grow", do_growfactor, " memory-grow-factor", 0x00000041L },
/* 7 */	{ "-maxgrow", do_maxgrowfactor, " max-memory-grow-factor", 0x00000081L },
/* 8 */ { "?", do_initusage, "", 0x000000001L },
	{ NULL, NULL }
};

PRIVATE long requiredMask = 0L; /* the xmask of mandatory options */

PRIVATE char errbuf[512];

PRIVATE void getopt(argc,argv)
int argc;
char *argv[];
{
  Command *com;
  char *s;
  long xmask,mask,m;
  int i;
  Boolean start;

  theArgc = argc;
  theArgv = argv;
  replay_mode = false;
  xmask = 0L;
  for (theArgi = 1; theArgi<theArgc; )
    {
      for (com=command, mask=1L; com->key!=NULL; com++, mask <<= 1)
	if (!strcmp(theArgv[theArgi],com->key))
	  {
	    if (m = (com->xmask & xmask))
	      {
		start = true;
		if (xmask & mask)
		  sprintf(errbuf,message(105),theArgv[theArgi]);
		else
		  {
		    sprintf(errbuf,message(106),theArgv[theArgi]);
		    s = errbuf+strlen(errbuf);
		    for (i=0; m!=0L; i++)
		      {
			if (m & 1)
			  {
			    if (start)
			      start = false;
			    else if (m!=0L && (m & (-m))==m)
			      strcpy(s,message(107));
			    else
			    strcpy(s,message(108));
			    s += strlen(s);
			    sprintf(s,"%s",command[i].key);
			    s += strlen(s);
			  }
			m >>= 1;
		      }
		  }
		bfatal(errbuf);
	      }
	    theArgi++;
	    xmask |= mask;
	    (*com->do_)();
	    break;
	  }
      if (com->key==NULL)
	{
	  bmsg(message(109),theArgv[theArgi]);
	  usage(false);
	  ml_done(10);
	}
    }
  if (m = (requiredMask & (~xmask)))
    {
      start = true;
      
      sprintf(errbuf,
	      ((m & (-m))==m)?message(110):message(111));
      s = errbuf+strlen(errbuf);
      for (i=0; m!=0L; i++)
	if (m & 1)
	  {
	    m >>= 1;
	    if (start)
	      start = false;
	    else if (m!=0L && (m & (-m))==m)
	      strcpy(s,message(107));
	    else
	      strcpy(s,message(108));
	    s += strlen(s);
	    sprintf(s,"%s",command[i].key);
	    s += strlen(s);
	  }
      bfatal(errbuf);
    }
}

PRIVATE void usage(fatalp)
IN Boolean fatalp;
{
  char *s;
  Command *com;
  long mask;

  sprintf(errbuf,message(112));
  s = errbuf+strlen(errbuf);
  for (com=command, mask = 1L; com->key!=NULL; com++, mask <<= 1)
    {
      sprintf(s,(requiredMask & mask)?" %s%s":" [%s%s]",com->key,com->doc);
      s += strlen(s);
    }
  *s++ = '\n';
  *s = 0;
  if (fatalp)
    bfatal(errbuf);
  else
    bmsg(errbuf);
}

#ifdef SIGFPE
EXTERN struct Expression *arithException;

void ml_fpe(code)
int code;
{
  while (true)
    Raise(arithException); /* cannot get back result,
			      so raise exception ad infinitum;
			      this is not very smart, but what
			      else can we do ? */
}
#endif
