/* Definitions for optimized (unused).
   Copyright (C) 1992-1997 Jean Goubault-Larrecq and Bull S.A.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* optimization data are coded as C structures,
   which we embed in HimML through the use of
   E_EXTERNAL tags. */

IMPORT void optimizeCode P((INOUT ExprPtr c));

typedef struct ppInfo { /* info on program points (alternatively, variables) */
  MapPtr uses; /* set of useful subsequent statements (or NIL, for return)
		  referring to the variable;
		  alternatively, co-reachable DEFUSE-like chains. */
  ExprPtr def; /* first definition of the value of the variable x; we never represent top,
		  x itself encodes bottom. */
  MapPtr eqns; /* valid equations a this program point */
  /* it is a map from depth-1 expressions to their simplifications
     (like Z_CONS(x,y) -> z). */
  MapPtr fragile; /* set of equations above that are fragile, i.e. that may not hold
		     after applying an impure function. */
  ExprPtr valset; /* set of all possible values for x, or NIL for bottom (whole universe) */
  COUNTER storage; /* storage class of the variable */
  CHAR tags[NTAGS]; /* possible tags */
} ppInfo;

#define PP(e) ((ppInfo *)EXTDATA(e))
#define HASTAG(pp,i) ((pp)->tags[i>>3] & (1 << ((i) & 7)))
#define SETTAG(pp,i) (pp)->tags[i>>3] |= (1 << ((i) & 7))

typedef Boolean (*argSweepFunc) P((IN CONST ExprPtr x,INOUT VOIDPTR _data));
#define ARGSWEEPFUNC(fun) \
PRIVATE Boolean fun P((IN CONST ExprPtr x,INOUT VOIDPTR _data)); \
PRIVATE Boolean fun(x,_data) \
IN CONST ExprPtr x; \
INOUT VOIDPTR _data;

typedef ExprPtr (*argMapFunc) P((IN CONST ExprPtr x,INOUT VOIDPTR _data));
#define ARGMAPFUNC(fun) \
PRIVATE ExprPtr fun P((IN CONST ExprPtr x,INOUT VOIDPTR _data)); \
PRIVATE ExprPtr fun(x,_data) \
IN CONST ExprPtr x; \
INOUT VOIDPTR _data;

IMPORT Boolean argSweep P((IN CONST ExprPtr arg,
			   IN CONST argSweepFunc fun,
			   INOUT VOIDPTR data));
IMPORT ExprPtr argMap P((IN CONST ExprPtr arg,
			 IN CONST argMapFunc fun,
			 INOUT VOIDPTR data));

IMPORT ppInfo *getPpInfo P((IN CONST ExprPtr stmt,INOUT ExprPtr deco));

typedef CONST ExprPtr (*crossFunc) P((ExprPtr,ExprPtr,char *));
typedef CONST ExprPtr (*ncrossFunc) P((INT,ExprPtr *,char *));

IMPORT ExprPtr getHighestMergeNode P((IN CONST MapPtr frontier,
				      IN CONST ExprPtr deco,
				      INOUT MapPtr *beforep));

typedef struct ppDecoration { /* results of analyzing a piece of code */
  MapPtr info; /* map from program points to ppInfos */
  MapPtr lambda; /* set of local Code functions (E_EXTERNALs) */
  INT fxtypes; /* flag bits for possible side-effects (see below, FXF_* flags) */
} ppDecoration;

#define PD(e) ((ppDecoration *)EXTDATA(e))

/* bit marks for use in the fxtypes field in a Code */
#define FXF_REF 0x01
#define FXF_ARRAY 0x02
#define FXF_ALL (FXF_REF | FXF_ARRAY)

#define Q_YES 1
#define Q_NO 2
#define Q_MAYBE (Q_YES | Q_NO)

IMPORT ExprPtr ppDecoInit P((void));
IMPORT void ppDecorate P((IN CONST ExprPtr x,IN CONST ExprPtr last,
			  INOUT ExprPtr deco,IN CONST MapPtr fragile,
			  IN CONST MapPtr eqns));
IMPORT void fillCode P((INOUT ExprPtr c,IN CONST ExprPtr deco));

IMPORT Boolean newUse P((IN CONST ExprPtr stmt,IN CONST ExprPtr use,INOUT ExprPtr deco));
