(* An implementation of message-passing functions between threads.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* requires mutex.ml *)

abstype ('a,''env,'rs) channel =
    CHANNEL of |[here : (''env,'rs) condition,
		 gone : (''env,'rs) condition,
		 value : 'a option ref
		 ]|
with
  fun channel_manager (tm : (''env thread,'rest,'rs) thread_manager,
		       new_sched : unit -> (''env thread,'rs) scheduler) =
      let val |[mutex,condition,wait,with_condition,broadcast,...]| =
		mutex_manager (tm,new_sched)
	  fun channel () =
	      let val m = mutex ()
	      in
		 CHANNEL |[here = condition m,
			   gone = condition m,
			   value = ref NONE
			   ]|
	      end
	  fun put (CHANNEL |[here,gone,value]|,v) =
	      let fun put' () =
		      case !value of
			   SOME _ => (wait gone; put' ())
			 | NONE => (value := SOME v;
				    broadcast here)
	      in
		 with_condition (gone, put')
	      end
	  fun get (CHANNEL |[here,gone,value]|) =
	      let fun get' () =
		      case !value of
			   SOME v => (value := NONE;
				      broadcast gone;
				      v)
			 | NONE => (wait here; get' ())
	      in
		 with_condition (here, get')
	      end
      in
	 |[channel = channel,
	   put = put,
	   get = get
	   ]|
      end
end
