(* Threads, mutexes and conditions, channels, futures, ...
   Simulated through catch/throw, not preemptive.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

(* This is an example on how you can declare thread functions;
   here, we use a simple fifo scheduler.
   Do beware that I have never tested these functions!
*)

(* if you just want some of the models, just erase the useless
   use "xxx.ml", val ... = xxx_manager ... and type xxx = ...
   declarations. *)

use "fifo.ml"

use "schedsig.ml"
use "simplesched.ml"
use "threadsig.ml"
use "thread.ml"
use "store.ml"
local
  val tm = thread_manager (current_store,new_store)
			  (simple_scheduler ())
in
  val |[abort = thread_abort,
	block = thread_block,
	yield,fork,reschedule,deschedule
	]| = tm
  use "mutex.ml"
  val |[mutex,try_acquire,acquire,release,with_mutex,
	condition,mutex_of,with_condition,signal,broadcast,wait,await
	]| = mutex_manager (tm,simple_scheduler)
  use "futures.ml"
  val |[future,touch,cobegin]| = future_manager (tm,simple_scheduler)
  use "channels.ml"
  val |[channel,put = send,get = recv]| =
      channel_manager (tm,simple_scheduler)
  use "rpc.ml"
  val |[rpc,accept = raccept,call = rcall]| =
      rpc_manager (tm,simple_scheduler)
  use "rw.ml"
  val |[rw,read_with_lock,write_with_lock]| = rw_manager (tm,simple_scheduler)
  use "chan.ml"
  val |[chan,put = async_put,get = async_get,get_wait,get_now]| =
      chan_manager (tm,simple_scheduler)
  use "rdv.ml" (* requires chan.ml *)
  val |[rendezvous,arm,select,accept = sel_accept,call = sel_call]| =
      rdv_manager (tm,simple_scheduler)
end

(* redefine all parameterized types to make the notation lighter *)
type thread = store thread
type mutex = (store,|[]|) mutex
type condition = (store,|[]|) condition
type 'a future = ('a,store,|[]|) future
type 'a channel = ('a,store,|[]|) channel
type ('a,'b) rpc = ('a,'b,store,|[]|) rpc
type rw = (store,|[]|) rw
type 'a chan = ('a,store,|[]|) chan
type ('a,'b) rendezvous = ('a,'b,store,|[]|) rendezvous
type ('a,'b) rdv_arm = ('a,'b,store,|[]|) rdv_arm
type ('a,'b) select = ('a,'b,store,|[]|) select
