(* Generic implementation of Union-Find structures.
   Copyright (C) 1992-1995 Jean Goubault and Bull S.A.
   All rights reserved.

   Redistribution and use in source and binary forms are permitted provided
   that: (1) source distributions retain this entire copyright notice and
   comment, and (2) distributions including binaries display the following
   acknowledgement:  ``This product includes software developed by Jean
   Goubault as a contributor of Bull S.A.'' in the
   documentation or other materials provided with the distribution and in
   all advertising materials mentioning features or use of this software.
   Neither the name of Bull S.A. or of its corporate affiliates,
   nor the names of its contributors may be used to endorse or promote
   products derived from this software without specific prior written permission.
   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
   WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
   MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*)

abstype
  ''a color = COLOR of ''a
with
  eqtype color
  exception UnColored
  datatype (''a,'info) ufind =
    UFIND of |[find : ''a -> ''a color, (* raises UnColored if not found *)
	       oneof : ''a color -> ''a, (* gets canonical element of color *)
	       info :  ''a color -> 'info, (* gets info associated with color *)
	       setinfo : ''a color * 'info -> unit,
	       merge : (''a color * ''a color * 'info * 'info -> 'info) ->
			''a color * ''a color -> ''a color,
	       colored : ''a -> bool,
	       copy : ('info -> 'info) -> (''a,'info) ufind,
	       newcolor : ''a * 'info -> ''a color (* add a new color; unspecified if already there *)
	       ]|
  local
  fun new (linkr,infor,numr) =
      let fun find x =
	      if x inset !linkr
		  then find (? (!linkr) x)
	      else COLOR x
      in UFIND
	  |[find = find,
	    oneof = fn (COLOR x) => x,
	    info = fn c => ? (!infor) c,
	    setinfo = fn (c,i) => infor := !infor ++ {c => i},
	    merge = fn extramerge =>
			fn (c1 as COLOR x1,c2 as COLOR x2) =>
			   let val n1 = ?(!numr) c1 and n2 = ?(!numr) c2
			   in
			      if n1<n2
				 then (linkr := !linkr ++ {x1 => x2};
				       numr := ({c1} <-| !numr) ++ {c2 => n1+n2};
				       infor := ({c1} <-| !infor) ++
					   {c2 => extramerge (c1,c2,?(!infor)c1,?(!infor)c2)};
				       c2)
			      else (linkr := !linkr ++ {x2 => x1};
				    numr := ({c2} <-| !numr) ++ {c1 => n1+n2};
				    infor := ({c2} <-| !infor) ++
					{c1 => extramerge (c2,c1,?(!infor)c2,?(!infor)c1)};
				    c1)
			   end,
	    colored = fn x => x inset !linkr orelse COLOR x inset !infor,
	    copy = fn infocopy => new (ref (!linkr),
				       ref {c => infocopy i | c => i in map !infor},
				       ref (!numr)),
	    newcolor = fn (x,i) =>
		       let val c = COLOR x
		       in (* no check that not (colored x) is performed! *)
			  infor := !infor ++ {c => i};
			  numr := !numr ++ {c => 1};
			  c
		       end
	    ]|
      end
  in
    fun ufindNew () = new (ref {},ref {},ref {})
  end
end

(*
  in STD_ML_NJ_TYVAR:
UFIND : |[colored : ''a -> bool,
	  oneof : ''a color -> ''a,
	  copy : ('b -> 'b) -> (''a,'b) ufind,
	  info : ''a color -> 'b,
	  merge : (''a color * ''a color * 'b * 'b -> 'b)
		  -> ''a color * ''a color -> ''a color,
	  find : ''a -> ''a color,
	  setinfo : ''a color * 'b -> unit,
	  newcolor : ''a * 'b -> ''a color]|
	-> (''a,'b) ufind
ufindNew :  unit -> (''1a,'1b) ufind
*)
