/* Bytecode compiler, himmrun version.
   Copyright (C) 2003-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#if TRANSLATOR
#include "expr.h"
#include "refcount.h"
#include "mem.h"
#include "bc.h"
#include "bcaux.h"
#include "timing.h"
#include "debug.h"
#include "transfer.h"
#include "errproto.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "iter.h"
#include "rbc.h"
#ifdef HAS_PROTOTYPES
#include "sbc.h"
#endif

#define __SEG__ BYTECODE

#define DEBUG

#if DEBUGGING!=0
PRIVATE ExprPtr currentDebugZone;
EXTERN Boolean debugEnabled;
#endif
#if PROFILING
EXTERN ExprPtr defaultProf;
#else
#define defaultProf NIL
#endif

PRIVATE void compileExpr P((bcFile *, compEnv *, ExprPtr));

struct argList {
  ExprPtr next; /* points to an extdesc with a struct argList inside. */
  compEnv *env;
  ExprPtr arg;
};

PRIVATE void arglistSweep (extdata, doext)
INOUT char *extdata;
IN CONST doExtDesc *doext;
{
  switch (doext->xtag)
    {
    case X_MARKDATA:
    case X_SAVE: /* no save option; arg lists are private to this file. */
      break;
    case X_MARK:
      (*doext->action) (((struct argList *)extdata)->arg);
      (*doext->action) (((struct argList *)extdata)->next);
      /* don't mark env, which should be reachable from somewhere else. */
      break;
    case X_FREE:
      MARKFREE (extdata);
      break;
    }
}

PRIVATE ExprPtr arglistLoad (desc,shared,sharedref,load)
IN CONST extDesc *desc;
IN Boolean shared;
INOUT ExprPtr sharedref;
IN CONST loadAction load;
{
  return NIL; /* no way to load arg lists; they are private to this file. */
}

/* By the way, since we never load or save arglists, it is not
 necessary to registerExtDesc(&arglistExtDesc). */

PUBLIC extDesc arglistExtDesc = {
  (CHAR *)"arglist",
  { 'a','r','g','l' }, NULL,
  NULL, /* no hash function, integer arrays are not shared */
  NULL, /* no comparison function, integer arrays are not shared */
  NULL, /* no print function */
  arglistSweep,
  arglistLoad
};

/* compilation to bytecode: */

PRIVATE void compileCGlobal(f,idx)
IN bcFile *f;
IN INT idx;
{
  bcPut(BC_CGLOBAL,f);
  bcPut((OPCODE)idx,f);
  bcPush(1,f);
}

PRIVATE INT findConstant(f,e)
IN bcFile *f;
IN CONST ExprPtr e;
{
  REGISTER bccBuffer *buf;
  REGISTER INT i;
  REGISTER ExprPtr *ep,*epend;

  for (buf = f->cbuf, i = 0; buf!=NULL; buf = buf->next)
    {
      ep = buf->data;
      epend = ep+buf->n;
      for (; ep<epend; i++)
	if (e == *ep++)
	  return i;
    }
  return -1;
}

PUBLIC INT getConstant(f,e)
IN bcFile *f;
IN CONST ExprPtr e;
{
  INT i;

  i = findConstant(f,e);
  if (i<0)
    {
      i = bccTell(f);
      bccPut(e,f);
    }
  return i;
}

PRIVATE void compileConstant(f,e)
IN bcFile *f;
IN CONST ExprPtr e;
{
  INT i;

  bcPut(BC_CONST,f);
  i = getConstant(f,e);
  bcPut((OPCODE)i,f);
  bcPush(1,f);
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileFetchOp(IN bcFile *f, IN OPCODE op, IN OPCODE specop,
			    IN INT n)
#else
PRIVATE void compileFetchOp(f,op,specop,n)
IN bcFile *f;
IN OPCODE op,specop;
IN INT n;
#endif
{
  n -= f->stackDepth;
  if (specop!=0 && n>=-8 && n<0)
    {
      bcPut(specop+8+n,f);
    }
  else
    {
      bcPut(op,f);
      bcPut((OPCODE)n,f);
    }
  bcPush(1,f);
}

PRIVATE void compileFetch(f,n)
IN bcFile *f;
IN INT n;
{
  compileFetchOp(f,BC_FETCH,BC_FETCH_m8,n);
}

PRIVATE void bcdPutInt(i,f)
IN INT i;
IN bcFile *f;
{
  bcdPut((OPCODE)(i>>(8*sizeof(OPCODE))),f);
  bcdPut((OPCODE)(i & ((1 << (8*sizeof(OPCODE))) - 1)),f);
}

#if DEBUGGING!=0
PRIVATE void compileCodePoint(f)
IN bcFile *f;
{
  INT pc;

  if (!debugEnabled)
    return;
  pc = bcTell(f);
  if ((INT)(OPCODE)pc == pc)
    {
      bcdPut(DBG_WHERE,f);
      bcdPut((OPCODE)pc,f);
    }
  else
    {
      bcdPut(DBG_WHERE_BIG,f);
      bcdPutInt(pc,f);
    }
  bcdPut((OPCODE)f->stackDepth,f);
  bcdPut((OPCODE)f->mapStackDepth,f);
  bcdPut ((OPCODE)0,f); /* was: bcdPut((OPCODE)f->excStackDepth,f);*/
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileDebugZone(IN Boolean startp, IN bcFile *f)
#else
PRIVATE void compileDebugZone(startp,f)
IN Boolean startp;
IN bcFile *f;
#endif
{
  if (currentDebugZone==NULL)
    return;
  if (!debugEnabled)
    return;
  {
    ExprPtr newfile;

    newfile = CAR(CDR(ARGS(currentDebugZone)));
    if (f->lastDebugZone==NULL ||
	CAR(CDR(ARGS(f->lastDebugZone)))!=newfile)
    { /* File name has changed: */
      INT i;

      i = getConstant (f, newfile);
      bcdPut(DBG_FILE,f);
      bcdPut((OPCODE)i,f);
    }
    f->lastDebugZone = currentDebugZone;
  }
  {
    COUNTER oldline, newline;
    COUNTER oldpos, newpos;

    oldline = f->lastLine;
    oldpos = f->lastPos;
    if (startp)
      {
	newline = ZONE(currentDebugZone)->line1;
	newpos = ZONE(currentDebugZone)->pos1;
      }
    else
      {
	newline = ZONE(currentDebugZone)->line2;
	newpos = ZONE(currentDebugZone)->pos2;
      }
    /* We add 1 to line2 (when at the end of an instruction),
       because we assume that every codepoint is on a different line.
       This means that in, say:
       let val x=3
       in
         x+1
       end
       the codepoint at the end of val x=3 will be shown on the 'in' line. */
    if (newline!=oldline || newpos!=oldpos)
      {
	bcdPut(DBG_LINE,f);
	bcdPut((OPCODE)newline,f);
	bcdPut((OPCODE)newpos, f);
      }
    f->lastLine = newline;
    f->lastPos = newpos;
  }
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileCodePointOp(IN OPCODE op,IN bcFile *f)
#else
PRIVATE void compileCodePointOp(op,f)
IN OPCODE op;
IN bcFile *f;
#endif
{
  compileCodePoint(f);
  bcPut(op,f);
}
#else
#define compileCodePoint(f)
#define compileCodePoint(op,f) bcPut(op,f)
#define compileDebugZone(startp,f)
#endif

PRIVATE INT getCurIEnv(env)
IN CONST compEnv *env;
{
  ExprPtr i_l,index;
  ENTER;

  i_l = KONS(NIL,KINT(env->ilevel));
  index = MapGetSure(i_l,CE_VARS(env));
  QUIT;
  return IVAL(index);
}

PRIVATE INT getCurEnv(env)
IN CONST compEnv *env;
{
  ExprPtr i_l,index;
  ENTER;

  i_l = KONS(NIL,KINT(env->level));
  index = MapGetSure(i_l,CE_VARS(env));
  QUIT;
  return IVAL(index);
}

PRIVATE void newCurEnv(env,n)
INOUT compEnv *env;
IN COUNTER n;
{
  ExprPtr i_l;
  ENTER;

  i_l = KONS(NIL,KINT(env->level));
#ifdef DEBUG
  {
    ExprPtr v;
    Boolean found;

    v = MapGet(i_l,CE_VARS(env),&found);
    if (found && IVAL(v)!=n)
      berror(message(661));
  }
#endif
  setMapPut(i_l,KINT(n),&CE_VARS(env)); /* record that we've built the new env */
  QUIT;
}

PRIVATE void compileCurEnv(f,env)
IN bcFile *f;
IN CONST compEnv *env;
{
  compileFetch(f,getCurEnv(env));
  /*
  bcPut(BC_FETCH,f);
  bcPut((OPCODE)(getCurEnv(env)-f->stackDepth),f);
  bcPush(1,f);
  */
}

PRIVATE bcFile *bcOpenD (env)
IN CONST compEnv *env;
{
  bcFile *f;

  f = bcOpenE (env);
#if DEBUGGING
  if (debugEnabled)
    {
      INT i;

      bcdPut (DBG_ENV, f);
      i = getConstant (f, CE_PARENT_DBGNAMES(env));
      bcdPut ((OPCODE)i, f);
    }
#endif
  return f;
}

PRIVATE ExprPtr bcCloseD(f)
IN bcFile *f;
{
#if DEBUGGING
  if (debugEnabled)
    {
      bcdPut (DBG_STOP, f);
    }
#endif
  return bcClose (f);
}

PRIVATE void compileVarref(f,env,level,rank)
IN bcFile *f;
IN CONST compEnv *env;
IN COUNTER level,rank;
{
  level = env->level-level;
  if (env->globenvs & (1L << level)) /* level is GLOBENV (or a copy of it) */
    {
#ifdef RISKY
      if (rank<EVARS(GLOBENV)->n)
	compileConstant(f,EVARS(GLOBENV)->expr[rank]);
      /* known constants are compiled right away */
      else
#endif
	{
	  bcPut(BC_GLOBVARREF,f);
	  bcPut((OPCODE)rank,f);
	  bcPush(1,f);
	}
    }
  else
    {
      ExprPtr i_l,index;
      Boolean found;
      ENTER;

      i_l = KINT(rank);
      if (level!=0)
	i_l = KONS(i_l,KINT(level));
      index = MapGet(i_l,CE_VARS(env),&found);
      if (found) /* variable already on stack */
	{
	  compileFetch(f,IVAL(index));
	  /*
	  bcPut(BC_FETCH,f);
	  bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	  */
	}
      else
	{
	  COUNTER i;

	  i = level;
	  do {
	    i_l = KONS(NIL,KINT(i));
	    index = MapGet(i_l,CE_VARS(env),&found);
	    if (found)
	      break;
	  } while (i++!=env->level);
	  /* level should always be >=0 here! */
	  switch (i-level)
	    {
	    case 0:
	      bcPut(BC_VARREF0,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    case 1:
	      bcPut(BC_VARREF1,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    case 2:
	      bcPut(BC_VARREF2,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    case 3:
	      bcPut(BC_VARREF3,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      break;
	    default:
	      bcPut(BC_VARREF,f);
	      bcPut((OPCODE)(IVAL(index)-f->stackDepth),f);
	      bcPut((OPCODE)(i-level),f);
	      break;
	    }
	  bcPut((OPCODE)rank,f);
	  bcPush(1,f);
	}
      QUIT;
    }
}

#if DEBUGGING!=0
PRIVATE Boolean debugVarref(f,env,level,rank)
IN bcFile *f;
IN CONST compEnv *env;
IN COUNTER level,rank;
{
  if (!debugEnabled)
    return false;
  /*level = env->level-level;*/
  if (env->globenvs & (1L << level)) /* level is GLOBENV (or a copy of it) */
    {
      return false;
      /* do not record any debugging information for global variables */
    }
  else
    {
      ExprPtr i_l,index;
      Boolean found;
      ENTER;

      i_l = KINT(rank);
      if (level!=0)
	i_l = KONS(i_l,KINT(level));
      index = MapGet(i_l,CE_VARS(env),&found);
      if (found) /* variable already on stack */
	{
	  bcdPut(DBG_ACCESS,f);
	  bcdPut((OPCODE)(IVAL(index)),f);
	}
      else
	{
	  COUNTER i;

	  i = level;
	  do {
	    i_l = KONS(NIL,KINT(i));
	    index = MapGet(i_l,CE_VARS(env),&found);
	    if (found)
	      break;
	  } while (i++!=env->level);
	  /* level should always be >=0 here! */
	  bcdPut(DBG_VARREF,f);
	  bcdPut((OPCODE)(IVAL(index)),f);
	  bcdPut((OPCODE)(i-level),f);
	  bcdPut((OPCODE)rank,f);
	}
      QUIT;
      return true;
    }
}
#endif

PRIVATE void compileSwDrop(f,excess)
IN bcFile *f;
IN INT excess;
{
  if (excess>1)
    {
      bcPut(BC_SWDROP,f);
      bcPut((OPCODE)-excess,f);
      bcPush(1-excess,f);
    }
}

#ifdef HAS_PROTOTYPES
PRIVATE void compileGoto(IN bcFile *f, IN OPCODE opcode,
			 INOUT ExprPtr gotop, IN COUNTER drop)
#else
PRIVATE void compileGoto(f,opcode,gotop,drop)
IN bcFile *f;
IN OPCODE opcode;
INOUT ExprPtr gotop;
IN COUNTER drop;
#endif
{
  ExprPtr g;
  INT excess,addr;
  Boolean patch = false;
  OPCODE offset;

  bcPush(-drop,f);
  if (f->dead)
    return;
  g = CAR(gotop);
  switch (g->kind)
    {
    case E_VARREF: /* backward jump */
      excess = f->stackDepth-VARLEVEL(g);
      if (excess<0)
	berror(message(609));
      if (excess==0)
	patch = gotoOp0(&opcode); /* special case for some opcodes, when they do not drop
				     any element from the stack. */
      bcPut(opcode,f);
      bcPut((OPCODE)excess,f);
      addr = bcTell(f);
      offset = (OPCODE)(VARRANK(g)-addr);
      if (addr+offset!=VARRANK(g))
	berror(message(94));
      if (patch)
	offset++; /* because pc will be 1 too low when taking the branch in that case */
      bcPut(offset,f);
      break;
    case E_NIL: case E_CONS: /*case E_TUPLE_2:*/ /* forward jump */
      {
	ExprPtr aux;
	ENTER;

	bcPut(opcode,f);
	bcPut((OPCODE)-1,f); /* dummy, to be patched later on. */
	aux = Kvarref(bcTell(f),f->stackDepth);
	(void) KEEP(aux);
	g = Kons(aux,g);
	SETCAR(gotop,g);
	QUIT;
	bcPut((OPCODE)-1,f); /* dummy, to be patched later on. */
      }
      break;
    default:
      kinderror(message(616),g->kind);
      /*NOTREACHED*/
      break;
    }
  if (opcode==BC_GOTO || opcode==BC_GOTO_0)
    f->dead = 1;
}

PRIVATE ExprPtr beginForward(f)
IN bcFile *f;
{
  ExprPtr gotop;

  gotop = MakePair(E_CONS,NIL,(f->dead?trueValue:falseValue));
  (void) KEEP(gotop);
  return gotop;
}

PRIVATE void endForward0(f,gotop)
IN bcFile *f;
INOUT ExprPtr gotop;
{
  ExprPtr ll,vr;
  INT excess;

  for (ll = CAR(gotop); ll!=NIL; ll=CDR(ll))
    {
      vr = CAR(ll);
      switch (ll->kind)
	{
	case E_CONS: /* normal goto */
	  excess = VARLEVEL(vr)-f->stackDepth;
	  if (excess<0)
	    berror(message(608));
	  bcPatchJump(VARRANK(vr),excess,f);
	  break;
	case E_TUPLE_2: /* switch table entry */
	  bcPatchSwitchEntry(VARRANK(vr),f);
	  break;
	}
    }
}

PRIVATE void endForward(f,gotop)
IN bcFile *f;
INOUT ExprPtr gotop;
{
  endForward0(f,gotop);
  if (CAR(gotop)!=NIL) /* in which case this is coming from (potentially) live code */
    { /* Indeed, compileGoto's in dead code are disabled, so only goto's coming from
	 live code are recorded. */
      if (f->stackDepth>f->maxStackDepth)
	berror(message(643));
      f->dead = 0;
      SETCAR(gotop,tuple0); /* to signal compileGoto() that it is now illegal to jump to gotop. */
#ifdef INSERT_MERGE_NODES
      bcPut(BC_MERGE,f);
#endif
    }
}

#define isForward(gotop) (CAR(gotop)->kind==(CHAR)E_CONS)

#if DEBUGGING!=0
PRIVATE INT beginDebug(f,st)
IN bcFile *f;
OUT bcdState *st;
{
  INT pc;
  INT patch;

  if (!debugEnabled)
    return 0;
  bcdGetState(f,st);
  st->dz = currentDebugZone;
  if (f->dead)
    return -1;
  pc = bcTell(f);
  if ((INT)(OPCODE)pc == pc)
    {
      bcdPut(DBG_BEGIN,f);
      bcdPut((OPCODE)pc,f);
    }
  else
    {
      bcdPut(DBG_BEGIN_BIG,f);
      bcdPutInt(pc,f);
    }
  patch = bcdTell(f);
  bcdPut((OPCODE)-1,f); /* unknown offset */
  return patch;
}

PRIVATE void endDebug(patch,f,st)
IN bcFile *f;
IN INT patch;
IN CONST bcdState *st;
{
  INT pc;

  if (!debugEnabled)
    return;
  currentDebugZone = st->dz;
  if (patch==-1)
    return;
  if (patch==bcdTell(f)-1) /* no debugging information recorded since beginDebug() */
    { /* then erase the BEGIN or BEGIN_BIG. */
      bcdRewindState(f,st);
    }
  else
    {
      bcdPatch(patch,f);
      pc = bcTell(f);
      if ((INT)(OPCODE)pc == pc)
	{
	  bcdPut(DBG_END,f);
	  bcdPut((OPCODE)pc,f);
	}
      else
	{
	  bcdPut(DBG_END_BIG,f);
	  bcdPutInt(pc,f);
	}
    }
}
#endif

PRIVATE void setStackDepth(f,level)
IN bcFile *f;
IN COUNTER level;
{
  if (f->dead || level==f->stackDepth)
    f->stackDepth = level;
  else
    berror(message(615));
}

PRIVATE void compileRaise(f,env)
IN bcFile *f;
IN CONST compEnv *env;
{
  ExprPtr handlers, h;
  COUNTER mlevel;

  compileDebugZone(true,f);
  handlers = CE_HANDLERS(env);
  if (handlers!=NIL)
    { /* compile a goto when the label for the handler is known */
      COUNTER savelevel;

      savelevel = f->stackDepth;
      handlers = CAR(handlers);
      mlevel = COUNTEDN(handlers);
      h = COUNTEDVAL(handlers);
      compileMapDrop(f,mlevel);
      compileSwDrop(f,f->stackDepth-COUNTEDN(h));
      compileGoto(f,BC_GOTO,COUNTEDVAL(h),0);
      f->stackDepth = savelevel;
    }
  else
    {
      compileCodePointOp(BC_RAISE,f);
    }
}

PRIVATE void compileRaiseImm(f,env,idx)
IN bcFile *f;
IN CONST compEnv *env;
IN CONST INT idx;
{
  compileCGlobal(f,idx);
  compileRaise(f,env);
}

PRIVATE void compileBindings P((IN bcFile *f,IN CONST compEnv *env,
				IN CONST ExprPtr bindings,INOUT ExprPtr errp));

MAPCOLLECTFUNC(do_restricttolevel)
{
  INT level;

  level = MAPDATA(INT);
  if ((level==0 && x->kind!=(CHAR)E_CONS) ||
      (level!=0 && x->kind==(CHAR)E_CONS && IVAL(CDR(x))==level))
    return MapletKons(x,y);
  return NewMap();
}

PRIVATE MapPtr restrictToLevel(oldmap,newmap,level)
IN CONST MapPtr oldmap,newmap;
IN INT level;
{
  MapPtr map;
  ENTER;

  map = MapDomRestrBy(newmap,oldmap);
  map = MapCollect(map,do_restricttolevel,(char *)level);
  map = MapMerge(oldmap,map);
  QUIT;
  return map;
}

#if DEBUGGING
MAPSWEEPFUNC(do_resetdbgnames)
{
  compEnv *env;
  ExprPtr name;

  env = MAPDATA(compEnv *);
  name = LETBIND(y);
  y = MakePair ((ExprKind)y->kind, x, LETBODY(y));
  (void) KEEP (y);
  CE_PARENT_DBGNAMES(env) = MapPut (name, y, CE_PARENT_DBGNAMES(env));
  return false;
}
#endif

PRIVATE void resetDbgNames (env)
INOUT compEnv *env;
{
#if DEBUGGING
  if (debugEnabled)
    {
      ENTER;

      (void) MapSweep (CE_DBGNAMES(env), do_resetdbgnames, env);
      KEEPONLY(CE_PARENT_DBGNAMES(env));
    }
#endif
  CE_DBGNAMES(env) = NewMap();
}

PRIVATE void compileLet1(f,env,bindings,body,errp,localp/*=true*/)
IN bcFile *f;
IN CONST compEnv *env;
IN CONST ExprPtr bindings;
IN CONST ExprPtr body;
INOUT ExprPtr errp;
IN Boolean localp;
{
  COUNTER n,n0;
  bindCompFun comp;
#if DEBUGGING!=0
  INT dstart;
  bcdState st;
#endif
  compEnv *newenv = (compEnv *)alloctmp(sizeof(compEnv));

  comp = /*localp?*/compileBindings/*:compileExprFun*/;
  *newenv = *env;
  CE_LETDELAY(newenv) = NewMap();
  resetDbgNames (newenv);
  newenv->_comp.next = theHeapFrames;
  theHeapFrames = (gcStackView *)&newenv->_comp;

  n = n0 = IVAL(CAR(bindings));
  if (n<0) /* then this is an imported module (an interface file, normally),
	      and the evaluator then first allocates a fake copy of the toplevel
	      environment GLOBENV below the current one. */
    {
      newenv->level++;
      newenv->globenvs |= (1L << newenv->level);
      n = -1-n;
    }
  else berror (message(930));
  newenv->level++;
  /* If n0<0, we do not disable the NEEDENV feature, the reason is the following:
     if n0<0, then this is not just a let or a local, but in fact an imported module.
     Such modules may be huge, and the only reasonable way to import them is with
     CEF_NEEDENV | CEF_SPARESTACK: this way, every binding is compiled as:
     <compute value> <get components of value as specified by the pattern>
     <store components in heap-allocated environment> <drop temporary data on stack>.
     Otherwise, every list of bindings is compiled as:
     <compute value> <get components>, possibly followed by <store components>
     (if CEF_NEEDENV is set), and leaving all components on the stack, which
     would be troublesome in the case of modules since the stack has a fixed
     size (BC_STACK_SIZE=512 by default).
     This can still be troublesome in case of big let- or local-expressions,
     or when importing a module inside a let, but I hope this won't happen.
     */
  if (newenv->flags & CEF_NEEDENV)
    {
      compileCurEnv(f,env); /* --- env (if needenv); --- (if !needenv) */
      /* notice that we compileCurEnv() on env, not newenv: we want to get the
	 old environment. */
      bcPut(BC_FAKEENV,f);
      bcPut(BC_CONSENV,f);
      bcPut((OPCODE)n,f);
      newCurEnv(newenv,f->stackDepth-1);
    }
  /* --- env (if needenv); --- (if !needenv) */
#if DEBUGGING!=0
  dstart = beginDebug(f,&st);
#endif
  newenv->ilevel = newenv->level;
  compileBindings(f,newenv,CDR(bindings),errp);
  newenv->ilevel = env->ilevel;
  newenv->flags = env->flags;
  (*comp)(f,newenv,body,errp); /* --- <...> res */
  /*
  if (localp)
    {
  */
      CE_VARS(env) = restrictToLevel(CE_VARS(env),CE_VARS(newenv),env->ilevel);
      CE_INLINES(env) = restrictToLevel(CE_INLINES(env),CE_INLINES(newenv),env->ilevel);
      /*
    }
      */
#if DEBUGGING!=0
  endDebug(dstart,f,&st);
#endif

  theHeapFrames = newenv->_comp.next;
}

MAPSWEEPFUNC(do_dumpletstores)
{
  struct dumpletstores *data;

  data = MAPDATA(struct dumpletstores *);
  bcPut((OPCODE)(IVAL(y)-data->f->stackDepth),data->f);
  bcPut((OPCODE)(IVAL(x)),data->f);
  return false;
}

PRIVATE void dumpLetStores(f,envlevel,set)
IN bcFile *f;
IN INT envlevel;
IN CONST MapPtr set;
{
  struct dumpletstores *data;

  if (set==NewMap())
    return;
  data = (struct dumpletstores *)alloctmp(sizeof(struct dumpletstores));
  if (set->kind==(CHAR)E_MAPLET)
    {
      bcPut(BC_LET1,f);
      bcPut((OPCODE)(envlevel-f->stackDepth),f);
      data->f = f;
      data->envlevel = envlevel;
      (void) do_dumpletstores(MAPLETX(set),MAPLETY(set),(char *)data);
    }
  else
    {
      bcPut(BC_LET,f);
      bcPut((OPCODE)(envlevel-f->stackDepth),f);
      data->f = f;
      data->envlevel = envlevel;
      (void) MapSweep(set,do_dumpletstores,data);
      bcPut((OPCODE)0,f);
    }
}

#if DEBUGGING!=0
MAPSWEEPFUNC(do_flushdbgnames)
{
  struct flushdbgnames *data;
  COUNTER level,rank;

  data = MAPDATA(struct flushdbgnames *);
  if (x->kind==E_INT)
    {
      rank = IVAL(x);
      level = 0;
    }
  else
    {
      rank = IVAL(CAR(x));
      level = IVAL(CDR(x));
    }
  if (debugVarref(data->f,data->env,level,rank))
    {
      INT bcupos,n,i;
      CHAR *namebuf;
      ExprPtr name;
      OPCODE op;

      name = LETBIND(y);
      n = SLEN(name);
      namebuf = (CHAR *)alloctmp(n+1);
      safememcpy(namebuf,STEXT(name),n);
      namebuf[n] = 0;
      bcupos = bcuTell(data->f);
      bcuPut(namebuf,n+1,1,data->f);
      switch (y->kind)
	{
	case E_APPL: op = DBG_DEFCON; break;
	case E_GOTO: op = DBG_DEFEXC; break;
	default: /* case E_LET: */ op = DBG_DEFNAME; break;
	}
      bcdPut(op,data->f);
      bcdPut((OPCODE)bcupos,data->f);
      i = getConstant(data->f,LETBODY(y));
      bcdPut((OPCODE)i,data->f);
    }
  return false;
}

PRIVATE void flushDbgNames(f,env)
IN bcFile *f;
IN CONST compEnv *env;
{
  struct flushdbgnames *data;
  MapPtr dbgnames;

  if (!debugEnabled)
    return;
  dbgnames = CE_DBGNAMES(env);
  if (dbgnames==NewMap())
    return;
  data = (struct flushdbgnames *)alloctmp(sizeof(struct flushdbgnames));
  data->f = f;
  data->env = env;
  (void) MapSweep(dbgnames,do_flushdbgnames,data);
  resetDbgNames (env);
}
#else
#define flushDbgNames(f,env)
#endif

PRIVATE ExprPtr compilePattern P((IN bcFile *f,IN CONST compEnv *env,IN CONST ExprPtr pat,
				  INOUT ExprPtr errp,IN CONST ExprPtr choicelist,
				  IN Boolean notagcheck));

PRIVATE void addLetDelay(env,rank,index)
INOUT compEnv *env;
IN INT rank, index;
{
  ExprPtr r,i,i_l;
#ifdef DEBUG
  ExprPtr ii;
  Boolean found;
#endif
  ENTER;

  r = KINT(rank);
  i = KINT(index);
#ifdef DEBUG
  ii = MapGet(r,CE_LETDELAY(env),&found);
  if (found && ii!=i)
    berror(message(610));
#endif
  setMapPut(r,i,&CE_LETDELAY(env));
  i_l = r;
  if (env->ilevel!=0)
    i_l = KONS(i_l,KINT(env->ilevel));
  setMapPut(i_l,i,&CE_VARS(env));
  QUIT;
}

#if DEBUGGING!=0
PRIVATE void addDbgName(env,rank,data)
INOUT compEnv *env;
IN INT rank;
IN CONST ExprPtr data;
{
  ExprPtr i_l;
  ENTER;

  if (debugEnabled)
    {
      i_l = KINT(rank);
      if (env->ilevel!=0)
	i_l = KONS(i_l,KINT(env->ilevel));
      setMapPut(i_l,data,&CE_DBGNAMES(env));
    }
  QUIT;
}
#else
#define addDbgName(env,rank,data)
#endif

PRIVATE void compileDrop(f,n)
IN bcFile *f;
IN COUNTER n;
{
  if (n!=0)
    {
      bcPut(BC_DROP,f);
      bcPut((OPCODE)n,f);
      bcPush(-n,f);
    }
}

PRIVATE void compilePushDrop(f,level,push)
IN bcFile *f;
IN COUNTER level;
IN Boolean push;
{
  if (push)
    compileSwDrop(f,f->stackDepth-level);
  else
    compileDrop(f,f->stackDepth-level);
}

PRIVATE void compileMapDrop(f,maplevel)
IN bcFile *f;
IN COUNTER maplevel;
{
  if (f->mapStackDepth!=maplevel)
    {
      bcPut(BC_MAPDROP,f);
      bcPut((OPCODE)(f->mapStackDepth-maplevel),f);
    }
}

DEF_PRIVATE_ACTION_1ARG(compilepattern)
{
  struct compilepattern *data;

  data = (struct compilepattern *)ENV_UNDERLYING(env);
  return compilePattern(data->f,data->env,data->pat,data->errp,data->choicelist,
			data->notagcheck);
}
ACTION_ENV

PRIVATE ExprPtr compilePattern1(f,env,pat,errp,choicelist,notagcheck)
IN bcFile *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
IN Boolean notagcheck;
{
  struct compilepattern *data = (struct compilepattern *)alloctmp(sizeof(struct compilepattern));

  data->f = f;
  data->env = env;
  data->pat = pat;
  data->errp = errp;
  data->choicelist = choicelist;
  data->notagcheck = notagcheck;
  return CallCC(ACTION_NAME(compilepattern),(ExprPtr)data,true);
}

PRIVATE ExprPtr compilePattern(f,env,pat,errp,choicelist,notagcheck)
IN bcFile *f;
INOUT compEnv *env;
IN CONST ExprPtr pat;
INOUT ExprPtr errp;
IN CONST ExprPtr choicelist;
IN Boolean notagcheck;
{
  ENTER;

  if (STKFULL(pat))
    {
      return compilePattern1(f,env,pat,errp,choicelist,notagcheck);
    }
again:
  switch (pat->kind)
    {
    case E_DEBUGNAME:
      addDbgName(env,COUNTEDN(DNINFO(pat)),COUNTEDVAL(DNINFO(pat)));
      pat = DNPAT(pat);
      goto again;
    case E_VARINTRO:
      addLetDelay(env,VARRANK(pat),f->stackDepth-1);
      break;
    default:
      kinderror(message(648),pat->kind);
      break;
    }
  KEEPONLY(choicelist);
  return choicelist;
}

PRIVATE void compileBinding P((IN bcFile *f,IN CONST compEnv *env,
			       IN CONST ExprPtr binding, INOUT ExprPtr errp));

DEF_PRIVATE_ACTION_1ARG(compilebinding)
{
  struct compilebinding *data;

  data = (struct compilebinding *)ENV_UNDERLYING(env);
  (void) compileBinding(data->f,data->env,data->binding,data->errp);
  return NIL;
}
ACTION_END

PRIVATE void compileBinding1(f,env,binding,errp)
IN bcFile *f;
IN CONST compEnv *env;
IN CONST ExprPtr binding;
INOUT ExprPtr errp;
{
  struct compilebinding *data = (struct compilebinding *)alloctmp(sizeof(struct compilebinding));

  data->f = f;
  data->env = env;
  data->binding = binding;
  data->errp = errp;
  (void) CallCC(ACTION_NAME(compilebinding),(ExprPtr)data,true);
}

PRIVATE void compileBinding(f,env,binding,errp)
IN bcFile *f;
IN CONST compEnv *env;
IN CONST ExprPtr binding;
INOUT ExprPtr errp;
{
  ENTER;

  switch (binding->kind)
    {
    case E_LET:
      /* local bindings */
      if (STKFULL(binding))
	{
	  compileBinding1(f,env,binding,errp);
	  return;
	}
      compileLet1(f,env,LETBIND(binding),LETBODY(binding),errp,true);
      break;
    case E_DEBUGDECL:
      {
#if DEBUGGING!=0
	if (debugEnabled)
	  {
	    currentDebugZone = REFVAL(binding);
	    compileDebugZone(false,f);
	    compileCodePoint(f);
	  }
#endif
	break;
      }
    default:
      {
	ExprPtr pat,e,savevar = NOEXPR,saveinline = NOEXPR;
	COUNTER savelevel = 0;
	INT saveflags;
	Boolean spare;

	pat = CAR(binding);
	e = CDR(binding);
	spare = (env->flags & (CEF_NEEDENV | CEF_SPARESTACK))==(CEF_NEEDENV | CEF_SPARESTACK);
	if (spare)
	  {
	    savevar = CE_VARS(env);
	    (void) KEEP(savevar);
	    saveinline = CE_INLINES(env);
	    (void) KEEP(saveinline);
	    savelevel = f->stackDepth;
	  }
	saveflags = env->flags;
	if ((env->flags & CEF_NEEDENV) && !needEnv(e,env))
	  env->flags &= ~CEF_NEEDENV;
	switch (e->kind)
	  {
	  case E_PROFILE:
	    {
	      ExprPtr saveprof;

	      saveprof = CE_PROF(env);
	      CE_PROF(env) = CDR(e);
	      e = CAR(e);
	      compileExpr(f,env,e); /* evaluate value */
	      CE_PROF(env) = saveprof;
	    }
	    break;
	  default:
	    {
	      ExprPtr saveprof;

	      saveprof = CE_PROF(env);
	      CE_PROF(env) = defaultProf;
	      compileExpr(f,env,e); /* evaluate value */
	      CE_PROF(env) = saveprof;
	    }
	    break;
	  }
	env->flags = saveflags;
	(void) compilePattern(f,env,pat,errp,NIL,false);
	/* we can ignore the result choicelist, which should
	   simply be NIL itself, since we do not include choices in
	   Let bindings. */
	if (env->flags & CEF_NEEDENV)
	  {
	    dumpLetStores(f,getCurIEnv(env),CE_LETDELAY(env));
	    if (spare)
	      {
		CE_VARS(env) = savevar;
		CE_INLINES(env) = saveinline;
		compileDrop(f,f->stackDepth-savelevel);
	      }
	  }
	CE_LETDELAY(env) = NewMap();
	/* there is no letconsenv trickery here, as we assume that when
	   compiling bindings (typically in a BC_LET), the current env
	   has already been allocated by BC_CONSENV. */
	flushDbgNames(f,env);
      }
    }
  QUIT;
}

PRIVATE void compileBindings(f,env,bindings,errp)
IN bcFile *f;
IN CONST compEnv *env;
IN CONST ExprPtr bindings;
INOUT ExprPtr errp;
{
#if DEBUGGING!=0
  ExprPtr savedz = currentDebugZone;
#endif

  while (bindings!=NIL)
    {
      compileBinding(f,env,CAR(bindings),errp);
      bindings = CDR(bindings);
    }
#if DEBUGGING!=0
  currentDebugZone = savedz;
#endif
}

PRIVATE void compileExpr(f,env,e)
IN bcFile *f;
IN CONST compEnv *env;
IN CONST ExprPtr e;
{
  ENTER;

again:
  switch (e->kind)
    {
    case E_VARREF:
      compileVarref(f,env,VARLEVEL(e),VARRANK(e));
      break;
    case E_DEBUGZONE:
#if DEBUGGING!=0
      if (debugEnabled)
	{
	  ExprPtr savedz = currentDebugZone;

	  currentDebugZone = DZINFO(e);
	  compileExpr(f,env,DZCODE(e));
	  currentDebugZone = savedz;
	  break;
	}
#endif
      e = DZCODE(e);
      goto again;
    case E_PROFILE:
      {
	ExprPtr saveprof;

	saveprof = CE_PROF(env);
	CE_PROF(env) = CDR(e);
	e = CAR(e);
	compileExpr (f,env,e);
	CE_PROF(env) = saveprof;
      }
    break;
    default:
      kinderror(message(623),e->kind);
      /*NOTREACHED*/
      break;
    }
  QUIT;
}

PUBLIC ExprPtr Compile(e,level,bindp)
IN CONST ExprPtr e;
IN COUNTER level;
IN Boolean bindp;
{
  bcFile *f;
  ExprPtr errp;
  compEnv *env = (compEnv *)alloctmp(sizeof(compEnv));

#if DEBUGGING!=0
  currentDebugZone = NULL;
#endif
  env->level = env->ilevel = level;
  env->globenvs = 0x1L; /* only level 0 holds GLOBENV */
  env->flags = CEF_NEEDENV | CEF_SPARESTACK;
  env->_comp.n = CE_N;
  env->_comp.next = theHeapFrames;
  CE_VARS(env) = NewMap();
  CE_INLINES(env) = NewMap();
  CE_HANDLERS(env) = NIL;
  CE_LETDELAY(env) = NewMap();
  CE_DZ(env) = NIL;
  CE_PROF(env) = defaultProf;
  CE_FV(env) = NewMap();
  CE_DBGNAMES(env) = NewMap();
  CE_PARENT_DBGNAMES(env) = NewMap();
  /*defineMe(env,NIL);*/
  theHeapFrames = (gcStackView *)&env->_comp;

  f = bcOpenD(env);

  newCurEnv(env,0); /* environment will be at position 0 of the stack, initially */
  errp = beginForward(f);
  if (bindp)
    {
      compileBindings(f,env,e,errp);
      compileConstant(f,falseValue); /* everything ok */
      if (isForward(errp))
	{
	  ExprPtr finp;

	  finp = beginForward(f);
	  compileGoto(f,BC_GOTO,finp,0);
	  setStackDepth(f,f->stackDepth-1);
	  endForward(f,errp);
	  compileConstant(f,trueValue);
	  endForward(f,finp);
	}
    }
  else
    compileExpr(f,env,e);
  compileDebugZone(false,f);
  compileCodePointOp(BC_RETURN,f);
  bcPut((OPCODE)(f->stackDepth-1),f);

  theHeapFrames = env->_comp.next;
  return bcCloseD (f);
}
#endif
