/* Definitions for compiler to interpreted form.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define ORIGINAL_CONS   0
#define ORIGINAL_REF	1
#define	ORIGINAL_GET	2
#define	ORIGINAL_SET	3
#define ORIGINAL_APPEND_OF  4
#define ORIGINAL_UNION_OF   5
#define ORIGINAL_MERGE_OF   6
#define ORIGINAL_MERGE_UNDER_OF 7
#define ORIGINAL_INTER_OF   8
#define ORIGINAL_ID 9
#define ORIGINAL_IT 10
#define ORIGINAL_TRUE   11
#define ORIGINAL_FALSE  12
#define ORIGINAL_NIL	13
#define ORIGINAL_BIND_EXC  14
#define ORIGINAL_MATCH_EXC  15
#define ORIGINAL_EMPTY_EXC  16
#define ORIGINAL_MAPGET_EXC 17
#define ORIGINAL_MAP_EXC	18
#define ORIGINAL_NOMEM_EXC  19
#define ORIGINAL_STACK_EXC  20
#define ORIGINAL_INCOMP_EXC 21
#define ORIGINAL_NTH_EXC	22
#define ORIGINAL_RANGE_EXC  23
#define ORIGINAL_ASCII_EXC  24
#define ORIGINAL_STRNTH_EXC 25
#define ORIGINAL_NONINV_EXC 26
#define ORIGINAL_SUBSCRIPT_EXC  27
#define ORIGINAL_ARITH_EXC  28
#define ORIGINAL_IO_EXC 29
#define ORIGINAL_RE_EXC 30
#define ORIGINAL_UFIND_EXC 31
#define ORIGINAL_CATCH_EXC 32
#define ORIGINAL_OVERWRITE  33
#define ORIGINAL_FORCE  34
#define ORIGINAL_MAPAPP 35
#define ORIGINAL_NONSENSE 36
#define ORIGINAL_FUTURE 37
#define ORIGINAL_EQUAL 38
#define ORIGINAL_NEQUAL 39
#define ORIGINAL_NOT 40
#define NORIGINALS 41

/* ids : map from ids to couple (i,level) or i if level=0; level is from the top
   structs : map from structure and functor ids to records (structures)
	     or lambdas (functors)
   sigs : map from signature identifiers to curtailing (record) pattern
	  (for structures, this is the curtailing pattern to apply to the structure;
	  for functors, this is the curtailing pattern to compose with the functor)
   level : level (from the top) for varrefs
   ilevel : level (from the top) for varintros
   nlocals : number of already allocated varintros (during pattern evaluation)
   cons : map from constructors at (i,level) (or i if level=0)
	  to constructor tag (or a set if no tag)
	  con = set of possible int(kind)s if no tag (datatype/abstype con, unboxed) [obsolete]
	  con = E_DATA(ttm,NIL,magic=index-of-constructor-in-ttm-array)
	        if datatype/abstype con (con is the constructor)
          con = E_DATACON(E_DATA(...as above...), flags)
                in most actual datatype/abstype cases.
	  con = exception(string) if exception con (always boxed)
	  con = abstract(con',env) if quotient type con (con' is a set or a string). [obsolete]
	  (env is shared amongst all objects of the same tyname)
	Except for:
	  con = polyfun (or clr_profile or profile naming the function)
	        when this is a polyfun that was deemed short enough to be inlinable)
	  con = string 'con' if omitted datatype constructor (datatype with
	        only one, non-constant constructor).
   tystate : type state (used only when compiling)
   tytags : map from type names to tytag maps.  A tytag map is an
            ttm = data (magic=-m, array [con(-m), ..., con(n)], nil)
	    of constructors, m>=0, n>=0 (with AVAL(CAR(ttm))->n = m+n+1).
	    If C is con(-i), i>0, then C takes no argument, and
	    the data C itself is represented as E_DATA(ttm, NIL, magic=-i);
	    if C is con(i), i>=0, then C takes an argument, and
	    C(arg) is represented as E_DATA(ttm, arg, magic=i).
	    All this except when the datatype has exactly one constructor,
	    which is non-constant; in this case, we map the type name to
	    the string of its unique constructor.
   imports : (reversed) list of dummy bindings (couples (pattern,variable name)
 	  to patch at link-time)
   dbgcnt : counter of debugzone blocks (only if debugging mode)
   */

#define C_NODEBUG (0x1)
#define C_NODBGNAME (0x2)
#define C_PROFILE (0x4)

struct compile_assum {
  CHAR kind; /*  = COMPILE_ASSUM_KIND	*/
  CHAR mark;
  CHAR flags;
  struct compile_assum *next;
  ExprPtr ids;
#if MODULE_SYSTEM & STD_ML_MODULE
  ExprPtr structs;
  ExprPtr sigs;
#endif
  COUNTER level;
  COUNTER ilevel;
  COUNTER nlocals;
  ExprPtr cons;
  ExprPtr tystate;
  MapPtr tytags;
  ExprPtr imports;
  ExprPtr fundef; /* name of current function */
  COUNTER dbgcnt;
};

#define C_FLAGS(assum) ((struct compile_assum *)assum)->flags

#define C_IDS(assum) ((struct compile_assum *)assum)->ids
#if MODULE_SYSTEM & STD_ML_MODULE
#define C_STRUCTS(assum) ((struct compile_assum *)assum)->structs
#define C_SIGS(assum) ((struct compile_assum *)assum)->sigs
#else
#define C_STRUCTS(assum) NIL
#define C_SIGS(assum) NIL
#endif
#define C_LEVEL(assum) ((struct compile_assum *)assum)->level
#define C_ILEVEL(assum) ((struct compile_assum *)assum)->ilevel
	/*  the level is coded as l;
	the level of varrefs is always (l & ~C_LEVEL_INC_MASK),
	the level of varintros is
	(l & C_LEVEL_INC_MASK)?((l & ~C_LEVEL_INC_MASK)+1):l
	*/
#define C_NLOCALS(assum) ((struct compile_assum *)assum)->nlocals
#define C_CONS(assum) ((struct compile_assum *)assum)->cons
#define C_TYSTATE(assum) ((struct compile_assum *)assum)->tystate
#define	C_TYTAGS(assum) ((struct compile_assum *)assum)->tytags
#define	C_IMPORTS(assum) ((struct compile_assum *)assum)->imports
#define	C_FUNDEF(assum) ((struct compile_assum *)assum)->fundef
#if DEBUGGING!=0
#define	C_DBGCNT(assum) ((struct compile_assum *)assum)->dbgcnt
#endif

#if MODULE_SYSTEM & STD_ML_MODULE
#define MAKE_TABLE(ids,structs,sigs,level,ilevel,nlocals,cons,tystate,tytags,imports,fundef,dbgcnt) \
	MakeCompAssum(ids,structs,sigs,level,ilevel,nlocals,cons,tystate,tytags,imports,fundef,dbgcnt)
IMPORT ExprPtr MakeCompAssum P((ExprPtr ids,ExprPtr structs,ExprPtr sigs,
				COUNTER level,COUNTER ilevel,COUNTER nlocals,ExprPtr cons,
				ExprPtr tystate,MapPtr tytags,ExprPtr imports,
				ExprPtr fundef,COUNTER dbgcnt));
#else
#define MAKE_TABLE(ids,structs,sigs,level,ilevel,nlocals,cons,tystate,tytags,imports,fundef,dbgcnt) \
	MakeCompAssum(ids,level,ilevel,nlocals,cons,tystate,tytags,imports,fundef,dbgcnt)
IMPORT ExprPtr MakeCompAssum P((ExprPtr ids,COUNTER level,COUNTER ilevel,COUNTER nlocals,
				ExprPtr cons,ExprPtr tystate,MapPtr tytags,ExprPtr imports,
				ExprPtr fundef,COUNTER dbgcnt));
#endif
IMPORT ExprPtr DupCompAssum P((ExprPtr assum));
IMPORT Boolean hasChoice P((IN CONST ExprPtr pat));
IMPORT MapPtr getDbgNames P((IN CONST ExprPtr pat, IN CONST MapPtr res));

EXTERN COUNTER minDebugLevel;
