/* Shared structure allocation functions.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include <assert.h>
#include "portable.h"
#include "mem.h"
#include "eval.h"
#include "expr.h"
#include "special.h"
#include "cons.h"
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#ifdef HAS_PROTOTYPES
#include "scons.h"
#endif

#define NDEBUG
#undef COUNT

#define __SEG__ EXPR

EXTERN ExprPtr zzero,two
#if FLOAT_FORMAT==IEEE754
	       ,pinf,minf
#endif
	;

#ifdef COUNT
INT nNewKtuple2 = 0, nNewKons = 0, nNewMapletKons = 0, nNewMapKons = 0,
  nNewKcomplex = 0, nNewKtypeAppl = 0,
  nNewKtypeFx = 0, nNewKrecord1 = 0;
INT nKtuple2 = 0, nKons = 0, nMapletKons = 0, nMapKons = 0, nKcomplex = 0, nKtypeAppl = 0,
  nKtypeFx = 0, nKrecord1 = 0;
INT nNewTuples = 0;
INT nKTuples = 0;
#endif

PUBLIC ExprPtr *Ktuple2Table=(ExprPtr *)0L;
PUBLIC ExprPtr *KonsTable=(ExprPtr *)0L;
PUBLIC ExprPtr *MapletKonsTable=(ExprPtr *)0L;
PUBLIC ExprPtr *MapKonsTable=(ExprPtr *)0L;
PUBLIC ExprPtr *KcomplexTable=(ExprPtr *)0L;
PUBLIC ExprPtr *KtypeApplTable=(ExprPtr *)0L;
PUBLIC ExprPtr *KtypeFxTable=(ExprPtr *)0L;
PUBLIC ExprPtr *Krecord1Table=(ExprPtr *)0L;
PUBLIC ExprPtr *DataTable=(ExprPtr *)0L;
PUBLIC ExprPtr *IntTable=(ExprPtr *)0L;
PUBLIC ExprPtr *RealTable=(ExprPtr *)0L;
PUBLIC ExprPtr *StringTable=(ExprPtr *)0L;
PUBLIC ExprPtr *VarTable=(ExprPtr *)0L;
PUBLIC ExprPtr *RefTable=(ExprPtr *)0L;
PUBLIC ExprPtr *ExtTable=(ExprPtr *)0L;
PUBLIC ExprPtr *ArrayTable=(ExprPtr *)0L;

EXTERN doExtDesc doExtFree;

PUBLIC unsigned long nSharedKtuple2 = 0L;
PUBLIC unsigned long nSharedKons = 0L;
PUBLIC unsigned long nSharedMapletKons = 0L;
PUBLIC unsigned long nSharedMapKons = 0L;
PUBLIC unsigned long nSharedKcomplex = 0L;
PUBLIC unsigned long nSharedKtypeAppl = 0L;
PUBLIC unsigned long nSharedKtypeFx = 0L;
PUBLIC unsigned long nSharedKrecord1 = 0L;
PUBLIC unsigned long nSharedData = 0L;
PUBLIC unsigned long nSharedInts = 0L;
PUBLIC unsigned long nSharedReals = 0L;
PUBLIC unsigned long nSharedArrays = 0L;
PUBLIC unsigned long nSharedStrings = 0L;

PUBLIC ExprPtr Ktuple2(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKtuple2++;
#endif
  RequireOneCell();
  epp0 = &Ktuple2Table[ktuple2_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedKtuple2++;
#ifdef COUNT
  nNewKtuple2++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_TUPLE_2;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr Kons(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKons++;
#endif
  RequireOneCell();
  epp0 = &KonsTable[kons_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedKons++;
#ifdef COUNT
  nNewKons++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_CONS;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr MapletKons(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nMapletKons++;
#endif
  RequireOneCell();
  epp0 = &MapletKonsTable[mapletkons_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedMapletKons++;
#ifdef COUNT
  nNewMapletKons++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_MAPLET;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr MapKons(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nMapKons++;
#endif
  RequireOneCell();
  epp0 = &MapKonsTable[mapkons_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedMapKons++;
#ifdef COUNT
  nNewMapKons++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_MAP;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr Kcomplex(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKcomplex++;
#endif
  RequireOneCell();
  epp0 = &KcomplexTable[kcomplex_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedKcomplex++;
#ifdef COUNT
  nNewKcomplex++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_COMPLEX;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr KtypeAppl(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKtypeAppl++;
#endif
  RequireOneCell();
  epp0 = &KtypeApplTable[ktypeappl_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedKtypeAppl++;
#ifdef COUNT
  nNewKtypeAppl++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_TYPE_APPL;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr KtypeFx(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKtypeFx++;
#endif
  RequireOneCell();
  epp0 = &KtypeFxTable[ktypefx_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedKtypeFx++;
#ifdef COUNT
  nNewKtypeFx++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_TYPEFX;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PUBLIC ExprPtr Krecord1(car,cdr)
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKrecord1++;
#endif
  RequireOneCell();
  epp0 = &Krecord1Table[krecord1_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;
#ifdef i386
    ExprPtr ecar, ecdr, elink;
#endif

    if (cdr->mark & MARK_IN_CDR ||
	car->mark & MARK_IN_CAR)
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=447,639 D1mr=332,788
		     when ep = *epp0 is put before the MARK_* tests,
		     Dr=428,987 D1mr=317,004 now. */
#ifdef i386
      for (; ep!=NOEXPR; ep = elink)
	{
	  /* fill entire cache line at once by loading all fields in one fell swoop.
	   We must hope the compiler knows how to optimizer this for Intel architectures.
	  (Normally, fetch ep->kind as a 32-bit word then keep only one byte, then
	  fetch all other fields.) */
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	  ep = elink; /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  elink = ELINK(ep); ecar = CAR(ep); ecdr = CDR(ep);
	  if (cdr==ecdr && car==ecar)
	    return ep;
	}
#else
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep))
	    return ep;
	}
#endif
    }
  }
  car->mark |= MARK_IN_CAR;
  cdr->mark |= MARK_IN_CDR;
  nSharedKrecord1++;
#ifdef COUNT
  nNewKrecord1++;
#endif
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_RECORD_1;
    ep->mark = MARK_SHARED;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

PRIVATE unsigned long gcd (i, j)
     IN unsigned long i, j;
{
  unsigned long k, r;

  if (i<j)
    {
      k = i;
      i = j;
      j = k;
    }
  while (j!=0)
    {
      r = i % j;
      i = j;
      j = r;
    }
  return i;
}

#define DEF_EXPAND_TABLE(name,table,oldsize,hashcode) \
PUBLIC Boolean name(size) \
/*INOUT ExprPtr *table;*/ \
/*INOUT unsigned long oldsize;*/ \
     IN unsigned long size; \
{ \
  ExprPtr *newPairTable; \
  ExprPtr *src_ap, *app, ap; \
  unsigned long h1, j, h2; \
  unsigned long d, p2; \
  ExprPtr **epp; \
 \
  d = gcd (size, oldsize); \
  p2 = size/d; \
  epp = (ExprPtr **) alloctmp (p2 * sizeof (ExprPtr *)); \
  newPairTable = (ExprPtr *)SAFEMALLOC(size*sizeof(ExprPtr)); \
  if (newPairTable==NULL) \
    { \
      freetmp (epp); \
      return false; \
    } \
  for (h1=0, src_ap = table; h1<oldsize; h1++, src_ap++) \
    { \
      /* Old hash code was h1 = a % oldsize; \
	 New hash code is h2 = a % size. \
	 In particular, a = oldsize * q1 + h1 \
	 a = size * q2 + h2. \
	 Let d = gcd (oldsize, size), p1 = oldsize/d, p2 = size/d. \
	 Then a = d p1 q1 + h1 = d p2 q2 + h2, \
	 so h2 = d (p1 q1 - p2 q2) + h1.  In particular, h2 differs from h1 by \
	 a multiple of d, and therefore h2 can take up to p2 different values \
	 (h1, h1+d, ..., h1+(p2-1)d; all this modulo size). \
 \
	 Although this code is meant to work for any values of oldsize and size, \
	 it has two limitations: \
	 - first, oldsize and size should have a large common factor, otherwise \
	 the epp table will start to grow too large; \
	 - second, size should really be larger than oldsize, otherwise \
	 we cannot guarantee that the lists in each slot will be copied \
	 in such a way that the order is preserved.  Preserving order is \
	 important, because the gc relies on having lists where elements \
	 in the new generations occur before elements of the old generations. \
      */ \
      for (j=0, h2=h1; j<p2; j++) \
	{ \
	  epp [j] = &newPairTable[h2]; /* h2 = h1+j d mod size */ \
	  h2 += d; \
	  if (h2>size) \
	    h2 -= size; \
	} \
      for (ap = *src_ap; ap!=NOEXPR; ap=ELINK(ap)) \
	{ \
	  h2 = hashcode(ap,size); \
	  if (h2<h1) \
	    { j = p2 - (h1-h2)/d; assert (h1==h2+(p2-j)*d); } /* Not (h2-h1)/d + p2: h2-h1 is *unsigned*! */ \
	  else { j = (h2-h1)/d; assert(h2==h1+d*j); }			\
	  *epp [j] = ap; \
	  epp [j] = &ELINK(ap); \
	} \
      for (j=0; j<p2; j++) \
	*epp[j] = NOEXPR; \
    } \
  MARKFREE(table); \
  table = newPairTable; \
  oldsize = size; \
  freetmp (epp); \
  return true; \
}

#ifdef OBSOLETE
#define PAIRHASH(ap,size) pair_hashcode1(CAR(ap),CDR(ap),size)

DEF_EXPAND_TABLE(ExpandKtuple2Table,Ktuple2Table,ktuple2_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandKonsTable,KonsTable,kons_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandMapletKonsTable,MapletKonsTable,mapletkons_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandMapKonsTable,MapKonsTable,mapkons_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandKcomplexTable,KcomplexTable,kcomplex_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandKtypeApplTable,KtypeApplTable,ktypeappl_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandKtypeFxTable,KtypeFxTable,ktypefx_hash_size,PAIRHASH)
DEF_EXPAND_TABLE(ExpandKrecord1Table,Krecord1Table,krecord1_hash_size,PAIRHASH)
#endif

PRIVATE unsigned long nextSize (size)
     IN unsigned long size;
{
  static unsigned long smallprimes[] = { 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 0 };
  unsigned long p;

  if (size % 2==0)
    size++;
  do {
    unsigned long *pp;

    for (pp=smallprimes; (p = *pp++)!=0; )
      if (size % p==0)
	{
	  size += 2;
	  break;
	}
  } while (p!=0);
  return size;
}

PUBLIC Boolean ResizePairTable(PairTablep, pair_hash_sizep, size)
     INOUT ExprPtr **PairTablep;
     INOUT unsigned long *pair_hash_sizep;
     IN CONST unsigned long size;
{
  ExprPtr *newPairTable;
  ExprPtr *epp, *eppend, ep, elink, car, cdr;
  ExprPtr *tgtp, tgt;

  /*
  if (size % *pair_hash_sizep==0)
    return ExpandPairTable (size);
  */
  size = nextSize (size);
  newPairTable = (ExprPtr *)SAFEMALLOC(size*sizeof(ExprPtr));
  if (newPairTable==NULL)
    return false;
  for (tgtp = newPairTable, eppend = tgtp+size; tgtp<eppend;)
    *tgtp++ = NOEXPR;
  for (epp = *PairTablep, eppend = epp + *pair_hash_sizep; epp<eppend; epp++)
    {
      for (ep = *epp; ep!=NOEXPR; ep = elink)
	{
	  elink = ELINK(ep);
	  car = CAR(ep);
	  cdr = CDR(ep);
	  tgtp = &newPairTable[pair_hashcode1(car,cdr,size)];
	  ELINK(ep) = *tgtp;
	  *tgtp = ep;
	}
    }
  MARKFREE(*PairTablep);
  *PairTablep = newPairTable;
  *pair_hash_sizep = size;
  return true;
}

PUBLIC Boolean ResizeKtuple2Table(size)
     IN unsigned long size;
{
  return ResizePairTable (&Ktuple2Table, &ktuple2_hash_size, size);
}

PUBLIC Boolean ResizeKonsTable(size)
     IN unsigned long size;
{
  return ResizePairTable (&KonsTable, &kons_hash_size, size);
}

PUBLIC Boolean ResizeMapletKonsTable(size)
     IN unsigned long size;
{
  return ResizePairTable (&MapletKonsTable, &mapletkons_hash_size, size);
}

PUBLIC Boolean ResizeMapKonsTable(size)
     IN unsigned long size;
{
  return ResizePairTable (&MapKonsTable, &mapkons_hash_size, size);
}

PUBLIC Boolean ResizeKcomplexTable(size)
     IN unsigned long size;
{
  return ResizePairTable (&KcomplexTable, &kcomplex_hash_size, size);
}

PUBLIC Boolean ResizeKtypeApplTable(size)
     IN unsigned long size;
{
  return ResizePairTable (&KtypeApplTable, &ktypeappl_hash_size, size);
}

PUBLIC Boolean ResizeKtypeFxTable(size)
     IN unsigned long size;
{
  return ResizePairTable (&KtypeFxTable, &ktypefx_hash_size, size);
}

PUBLIC Boolean ResizeKrecord1Table(size)
     IN unsigned long size;
{
  return ResizePairTable (&Krecord1Table, &krecord1_hash_size, size);
}

#ifdef HAS_PROTOTYPES
PUBLIC ExprPtr Kdata(IN REGISTER SHORT magic,
		     IN REGISTER CONST ExprPtr car,
		     IN REGISTER CONST ExprPtr cdr)
#else
PUBLIC ExprPtr Kdata(magic,car,cdr)
IN REGISTER SHORT magic;
IN REGISTER CONST ExprPtr car;
IN REGISTER CONST ExprPtr cdr;
#endif
{
  REGISTER ExprPtr *epp0;

  RequireOneCell();
  epp0 = &DataTable[data_hashcode(magic,car,cdr)];
  {
    REGISTER ExprPtr ep;

    if (/*cdr==NIL || */(cdr->mark & MARK_IN_DATA))
    {
      ep = *epp0; /* Note that this gets us out of the data L1 cache fairly often.
		     According to valgrind --tool=cachegrind, Dr=16,796   D1mr=8,852
		     if ep = *epp0 is done before the MARK_* test,
		     Dr=13,764   D1mr=6,371 now. */
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep) && magic==ep->magic)
	    /* ep->kind must be E_DATA, so we don't check it */
	    return ep;
	  ep = ELINK (ep); /* unroll loop */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep) && magic==ep->magic)
	    /* ep->kind must be E_DATA, so we don't check it */
	    return ep;
	  ep = ELINK (ep); /* unroll loop */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep) && magic==ep->magic)
	    /* ep->kind must be E_DATA, so we don't check it */
	    return ep;
	  ep = ELINK (ep); /* unroll loop */
	  if (ep==NOEXPR)
	    break;
	  if (cdr==CDR(ep) && car==CAR(ep) && magic==ep->magic)
	    /* ep->kind must be E_DATA, so we don't check it */
	    return ep;
	}
    }
  }
  /*if (cdr!=NIL)
    {*/
  cdr->mark |= MARK_IN_DATA;
  /*}*/
  nSharedData++;
  {
    REGISTER ExprPtr ep;

    QALLOC_HOMO(ep);
    ep->kind = (CHAR)E_DATA;
    ep->mark = MARK_SHARED;
    ep->magic = magic;
    /*ep->gcmark = 0;*/
    ELINK(ep) = *epp0; /* always insert at beginning of slot */
    CAR(ep) = car;
    CDR(ep) = cdr;
    *epp0 = ep;
    return ep;
  }
}

#define DATAHASH(ap,size) data_hashcode1(ap->magic,CAR(ap),CDR(ap),size)
DEF_EXPAND_TABLE(ExpandDataTable,DataTable,data_hash_size,DATAHASH)

#ifdef OBSOLETE
PUBLIC ExprPtr SharePair(e)
IN REGISTER ExprPtr e;
{
  REGISTER ExprKind kind;
  REGISTER CONST ExprPtr car;
  REGISTER CONST ExprPtr cdr;
  REGISTER ExprPtr *epp0;

  RequireOneCell();
  kind = e->kind;
  car = CAR(e);
  cdr = CDR(e);
  epp0 = &PairTable[pair_hashcode(car,cdr)];
  {
    REGISTER ExprPtr ep;

    ep = *epp0;
    if (/*cdr==NIL || */(cdr->mark & MARK_IN_CDR) ||
	/*car==NIL || */(car->mark & MARK_IN_CAR))
    {
      for (; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (cdr==CDR(ep) && car==CAR(ep) && kind==ep->kind)
	    return ep;
	}
    }
  }
  /*if (car!=NIL)
    {*/
  car->mark |= MARK_IN_CAR;
  /*}*/
  /*if (cdr!=NIL)
    {*/
  cdr->mark |= MARK_IN_CDR;
  /*}*/
  e->mark = MARK_SHARED;
  /*e->gcmark = 0;*/
  ELINK(e) = *epp0; /* always insert at beginning of slot */
  *epp0 = e;
  return e;
}
#endif

PUBLIC ExprPtr KONS(a,b)
IN CONST ExprPtr a,b;
{
  ExprPtr c;

  c = Kons(a,b);
  return KEEP(c);
}

PUBLIC ExprPtr Kint(i)
IN REGISTER INT i;
{
  REGISTER ExprPtr *epp0;

  RequireOneCell();
  epp0 = &IntTable[int_hashcode(i)];
  {
    REGISTER ExprPtr ep;

    for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
      {
	if (IVAL(ep)==i)
	  return ep;
	ep = ELINK(ep); /* unroll loop */
	if (ep==NOEXPR)
	  break;
	if (IVAL(ep)==i)
	  return ep;
	ep = ELINK(ep); /* unroll loop */
	if (ep==NOEXPR)
	  break;
	if (IVAL(ep)==i)
	  return ep;
	ep = ELINK(ep); /* unroll loop */
	if (ep==NOEXPR)
	  break;
	if (IVAL(ep)==i)
	  return ep;
      }
  }
  nSharedInts++;
  {
    REGISTER ExprPtr new;

    QALLOC_HOMO(new);
    new->kind = (CHAR)E_INT;
    new->mark = MARK_SHARED;
    /*new->gcmark = 0;*/
    ELINK(new) = *epp0; /* always insert at beginning of slot */
    IVAL(new) = i;
    INEXT(new) = NOEXPR;
    *epp0 = new;
    return new;
  }
}

#define INTHASH(ap,size) int_hashcode1(IVAL(ap),size)
DEF_EXPAND_TABLE(ExpandIntTable,IntTable,int_hash_size,INTHASH)

PUBLIC ExprPtr KINT(i)
IN CONST INT i;
{
  ExprPtr c;

  c = Kint(i);
  return KEEP(c);
}

#if FLOAT_FORMAT==IEEE754
PUBLIC ExprPtr *NaNs=(ExprPtr *)0L;
#endif

PUBLIC ExprPtr Kreal(x)
IN REGISTER CONST REAL *x;
{
#if FLOAT_FORMAT==IEEE754
  switch (classifyD(x))
    {
    case C_ZERONUM:
    case -C_ZERONUM: /* identify 0.0 and -0.0 */
      return zzero;
    case C_INFINITE:
      return pinf;
    case -C_INFINITE:
      return minf;
    case C_QNAN:
    case C_SNAN:
    case -C_SNAN:
    {
      int i;

      i = NaNcodeD(x);
      if (i<-MAX_NAN || i>MAX_NAN)
	{
	  berror(message(423),i);
	  i = 0;
	}
      return NaNs[i+MAX_NAN];
    }
    }
#else
  if (*x==zero)
    return zzero;
#endif
    {
      REGISTER ExprPtr *epp0;

      RequireOneCell();
      epp0 = &RealTable[real_hashcode(x)];
      {
	REGISTER ExprPtr ep;

	for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
	  {
	    if (XVAL(ep) == *x)
	      return ep;
	    ep = ELINK(ep); /* unroll loop. */
	    if (ep==NOEXPR)
	      break;
	    if (XVAL(ep) == *x)
	      return ep;
	    ep = ELINK(ep); /* unroll loop. */
	    if (ep==NOEXPR)
	      break;
	    if (XVAL(ep) == *x)
	      return ep;
	    ep = ELINK(ep); /* unroll loop. */
	    if (ep==NOEXPR)
	      break;
	    if (XVAL(ep) == *x)
	      return ep;
	  }
      }
      nSharedReals++;
      {
	REGISTER ExprPtr new;

	QALLOC_HOMO(new);
	new->kind = (CHAR)E_REAL;
	new->mark = MARK_SHARED;
	/*new->gcmark = 0;*/
	ELINK(new) = *epp0; /* always insert at beginning of slot */
	XVAL(new) = *x;
	*epp0 = new;
	return new;
      }
    }
}

#define REALHASH(ap,size) real_hashcode1(&XVAL(ap),size)
DEF_EXPAND_TABLE(ExpandRealTable,RealTable,real_hash_size,REALHASH)

PUBLIC COUNTER string_hashcode1(len,name,size)
IN REGISTER INT len;
IN REGISTER CONST CHAR *name;
IN unsigned long size;
{
  REGISTER CHAR *end;
  REGISTER unsigned long total;

/* we compare on at most 4 characters at the beginning and 4 characters
   at the end; this makes hashing fast. */
  end = name+len;
  total = len;
  switch (len)
    {
      case 8: def: total += end[-4]; /*fallthrough*/
      case 7: total += name[3]; /*fallthrough*/
      case 6: total += end[-3]; /*fallthrough*/
      case 5: total += name[2]; /*fallthrough*/
      case 4: total += end[-2]; /*fallthrough*/
      case 3: total += name[1]; /*fallthrough*/
      case 2: total += end[-1]; /*fallthrough*/
      case 1: total += name[0]; /*fallthrough*/
      case 0: break;
      default: goto def;
    }
  return (COUNTER)(total % size);
}

/* the string table has slots of the form:
	a binary tree, using the middle character c to decide,
	with leaves as unordered lists of strings (through
	the link field). This binary tree is realized with
	E_SLOT_BRANCH nodes: the mark field identifies a
	character, if it is equal to c, we traverse the list
	in the link field; if c is less, we try the branch
	in the car; otherwise, the branch in the cdr.
		This tree is not balanced, because we hope
	it will be almost balanced on average.
*/
PUBLIC ExprPtr Kstring1(len,name,copy)
IN REGISTER INT len;
IN REGISTER CHAR *name;
IN Boolean copy;
{
  switch (len)
    {
    case 0:
      if (!copy)
	VECFREE(name,0);
      return voidString;
    case 1:
      {
	ExprPtr e;

	e = charStrings[*name];
	if (!copy)
	  VECFREE(name,1);
	return e;
      }
    default:
    {
      REGISTER ExprPtr *epp0;

      RequireOneHeteroCell();
      epp0 = &StringTable[string_hashcode(len,name)];
      {
	REGISTER ExprPtr ep;

	for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
	  {
	    if (len!=SLEN(ep))
	    {
proceed:
	      continue;
	    }
	    if (memcmp (name, STEXT(ep), len)==0)
	      {
		if (!copy)
		  VECFREE(name, len);
		return ep;
	      }
#ifdef OBSOLETE
	    {
	      REGISTER CHAR *s,*ss;
	      REGISTER INT i;

	      for (i=0, s=name, ss=(CHAR *)STEXT(ep); i<len; i++)
		if (*s++ != *ss++)
	          goto proceed;
	      if (!copy)
		VECFREE(name, len);
	      return ep;
	    }
#endif
	  }
      }
      nSharedStrings++;
      {
	REGISTER ExprPtr new;

	QALLOC_HETERO(new);
	new->kind = (CHAR)E_STRING;
	new->mark = MARK_SHARED;
	/*new->gcmark = 0;*/
	ELINK(new) = *epp0; /* always insert at beginning of slot */
	SLEN(new) = MAX_SIGNED_LONG; /* to prevent stop-and-copy gc from moving name,
					which is normally allocated on the heap */
	STEXT(new) = name;
	*epp0 = new;
	if (copy)
	  {
	    BEGINQUICKKEEP();
	    QUICKKEEP(new);
#if STOP_AND_COPY_VECTORS
	    scAlloc(*(char **)&STEXT(new),len);
#else
	    VECALLOC(STEXT(new),len);
#endif
	    memcpy(STEXT(new),name,len);
	    ENDQUICKKEEP();
	  }
	SLEN(new) = len;
	return new;
      }
    }
    }
}

#define STRINGHASH(ap,siz) string_hashcode1(SLEN(ap),STEXT(ap),siz)
DEF_EXPAND_TABLE(ExpandStringTable,StringTable,string_hash_size,STRINGHASH)

PUBLIC ExprPtr Kstr(name)
IN CONST CHAR *name;
{
  return Kstring(strlen((char *)name),name);
}

PUBLIC ExprPtr Kvar(kind,i,level)
IN ExprKind kind;
IN REGISTER COUNTER i,level;
{
  REGISTER ExprPtr *epp0;

  RequireOneCell();
  epp0 = &VarTable[var_hashcode(i,level)];
  {
    REGISTER ExprPtr ep;

    for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
      {
	if (i==VARRANK(ep) && level==VARLEVEL(ep) && ep->kind==kind)
	  return ep;
	ep = ELINK(ep); /* unroll loop. */
	if (ep==NOEXPR)
	  break;
	if (i==VARRANK(ep) && level==VARLEVEL(ep) && ep->kind==kind)
	  return ep;
	ep = ELINK(ep); /* unroll loop. */
	if (ep==NOEXPR)
	  break;
	if (i==VARRANK(ep) && level==VARLEVEL(ep) && ep->kind==kind)
	  return ep;
	ep = ELINK(ep); /* unroll loop. */
	if (ep==NOEXPR)
	  break;
	if (i==VARRANK(ep) && level==VARLEVEL(ep) && ep->kind==kind)
	  return ep;
      }
  }
  {
    REGISTER ExprPtr cons;

    QALLOC_HOMO(cons);
    cons->kind = (CHAR)kind;
    cons->mark = MARK_SHARED;
    /*cons->gcmark = 0;*/
    ELINK(cons) = *epp0; /* always insert at beginning of slot */
    VARRANK(cons) = i;
    VARLEVEL(cons) = level;
    *epp0 = cons;
    return cons;
  }
}

PUBLIC ExprPtr Kcounted(kind,expr,n)
IN ExprKind kind;
IN CONST ExprPtr expr;
IN INT n;
{
  REGISTER ExprPtr *epp0;

  RequireOneCell();
  epp0 = &RefTable[ref_hashcode(expr)];
  {
    REGISTER ExprPtr ep;

    if (/*expr==NIL || */(expr->mark & MARK_IN_REF))
      for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (REFVAL(ep)==expr && ep->kind==kind && COUNTEDN(ep)==n)
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (REFVAL(ep)==expr && ep->kind==kind && COUNTEDN(ep)==n)
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (REFVAL(ep)==expr && ep->kind==kind && COUNTEDN(ep)==n)
	    return ep;
	  ep = ELINK(ep); /* unroll loop. */
	  if (ep==NOEXPR)
	    break;
	  if (REFVAL(ep)==expr && ep->kind==kind && COUNTEDN(ep)==n)
	    return ep;
	}
  }
  /*if (expr!=NIL)
    {*/
  expr->mark |= MARK_IN_REF;
  /*}*/
  {
    REGISTER ExprPtr new;

    QALLOC_HOMO(new);
    new->kind = (CHAR)kind;
    new->mark = MARK_SHARED;
    /*new->gcmark = 0;*/
    ELINK(new) = *epp0; /* always insert at beginning of slot */
    REFVAL(new) = expr;
    COUNTEDN(new) = n;
    *epp0 = new;
    return new;
  }
}

PUBLIC ExprPtr KCOUNTED(expr,n)
IN CONST ExprPtr expr;
IN INT n;
{
  expr = Kcounted(E_COUNTED,expr,n);
  return KEEP(expr);
}


PUBLIC ExprPtr Kexternal(desc,data)
IN REGISTER CONST extDescPtr desc;
IN REGISTER CONST char *data;
{
  {
    REGISTER ExprPtr *epp0;
    REGISTER CHAR c;

    RequireOneHeteroCellAndOneSlot();
    epp0 = &ExtTable[(*desc->hash)(data) % EXT_HASH_SIZE];
    c = desc->name[0];
    {
      REGISTER ExprPtr ep;

      for (; (ep = *epp0)!=NOEXPR;)
	{
	  if (ep->kind==(CHAR)E_SLOT_BRANCH)
	    {
	      if (c==ep->mark)
		{
		  epp0 = &ELINK(ep);
		  break;
		}
	      if (c<ep->mark)
		epp0 = &CAR(ep);
	      else
		epp0 = &CDR(ep);
	    }
	  else
	    {
	      REGISTER CHAR cc;

	      cc = EXTDESC(ep)->name[0];
	      if (cc!=c) /* create a E_SLOT_BRANCH node */
		{
		  REGISTER ExprPtr cons;

		  SLOTALLOC(cons);
		  *epp0 = cons;
		  cons->mark = cc;
		  ELINK(cons) = ep;
		  CAR(cons) = CDR(cons) = NOEXPR;
		  if (c<cc)
		    epp0 = &CAR(cons);
		  else
		    epp0 = &CDR(cons);
		}
	      break;
	    }
	}
      for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
	{
	  if (EXTDESC(ep)==desc && (*desc->eq)(EXTDATA(ep),data))
	    {
	      (*desc->sweep) (data, &doExtFree);
	      return ep;
	    }
	}
    }
    {
      ExprPtr new;

      QALLOC_HETERO(new);
      new->kind = (CHAR)E_EXTERNAL;
      new->mark = MARK_SHARED;
      /*new->gcmark = 0;*/
      ELINK(new) = *epp0; /* always insert at beginning of slot */
      EXTDESC(new) = desc;
      EXTDATA(new) = data;
      *epp0 = new;
      return new;
    }
  }
}

PUBLIC COUNTER array_hashcode1(n,exprs,kind,shape,size)
IN INT n;
IN CONST REGISTER ExprPtr *exprs;
IN ExprKind kind;
IN CONST ExprPtr shape;
IN unsigned long size;
{
  REGISTER unsigned long total;
  REGISTER ExprPtr *epp;

  /* we only use the first 4 and the last 4 entries of the array to
     compute the hash function, for speed. */
  total = n + (unsigned long)kind + (((unsigned long)shape) >> NLOWBITS);
  epp = exprs+n;
  switch (n)
    {
      case 8: def: total += (unsigned long)epp[-4]; /*fallthrough*/
      case 7: total += (unsigned long)exprs[3]; /*fallthrough*/
      case 6: total += (unsigned long)epp[-3]; /*fallthrough*/
      case 5: total += (unsigned long)exprs[2]; /*fallthrough*/
      case 4: total += (unsigned long)epp[-2]; /*fallthrough*/
      case 3: total += (unsigned long)exprs[1]; /*fallthrough*/
      case 2: total += (unsigned long)epp[-1]; /*fallthrough*/
      case 1: total += (unsigned long)exprs[0]; /*fallthrough*/
      case 0: break;
      default: goto def;
    }
  return (COUNTER)(total % size);
}

PUBLIC ExprPtr Karray(kind,shape,n,exprs)
IN REGISTER ExprKind kind;
IN CONST REGISTER ExprPtr shape;
IN REGISTER INT n;
IN CONST ExprPtr *exprs;
{
  REGISTER ExprPtr *epp0;

#ifdef COUNT
  nKTuples++;
#endif
  RequireOneHeteroCell();
  epp0 = &ArrayTable[array_hashcode(n,exprs,kind,shape)];
  {
    REGISTER ExprPtr ep;

    for (ep = *epp0; ep!=NOEXPR; ep = ELINK(ep))
      {
	if (shape!=ASHAPE(ep))
	  continue;
	if (kind!=ep->kind)
	  continue;
	{
	  REGISTER VecPtr vpp;
	  
	  vpp = AVAL(ep);
	  if (vpp->n!=n)
	    {
proceed:
	      continue;
	    }
	  if (memcmp (exprs, vpp->expr, n*sizeof(ExprPtr))==0)
	      return ep;
#ifdef OBSOLETE
	  {
	    REGISTER ExprPtr *elp,*elpp;
	    REGISTER COUNTER i;
	    
	    for (i=0, elp=exprs, elpp=vpp->expr; i<n; i++)
	      {
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
		i++; /* unroll loop. */
		if (i>=n)
		  break;
		if (*elp++ != *elpp++)
		  goto proceed;
	      }
	    return ep;
	  }
#endif
	}
      }
  }
  nSharedArrays++;
#ifdef COUNT
  nNewTuples++;
#endif
  {
    REGISTER ExprPtr expr;
    VecPtr vp;

    scAlloc(*(char **)&vp,VECSIZE(n));
    vp->n = n;
    memcpy(vp->expr,exprs,n*sizeof(ExprPtr));
    MAKEKRECORD_NOGC(kind,shape,vp,expr);
    ELINK(expr) = *epp0; /* insert at start of slot */
    *epp0 = expr;
    expr->mark |= MARK_SHARED;
    return expr;
  }
}

#ifdef OBSOLETE
#define ARRAYHASH(ap,size) array_hashcode1(AVAL(ap)->n,AVAL(ap)->expr,ap->kind,ASHAPE(ap),size)
DEF_EXPAND_TABLE(ExpandArrayTable,ArrayTable,array_hash_size,ARRAYHASH)
#endif

PUBLIC Boolean ResizeKarrayTable(size)
     IN CONST unsigned long size;
{
  ExprPtr *newKarrayTable;
  ExprPtr *epp, *eppend, ep, elink;
  ExprPtr *tgtp, tgt;

  size = nextSize (size);
  newKarrayTable = (ExprPtr *)SAFEMALLOC(size*sizeof(ExprPtr));
  if (newKarrayTable==NULL)
    return false;
  for (tgtp = newKarrayTable, eppend = tgtp+size; tgtp<eppend;)
    *tgtp++ = NOEXPR;
  for (epp = ArrayTable, eppend = epp + array_hash_size; epp<eppend; epp++)
    {
      for (ep = *epp; ep!=NOEXPR; ep = elink)
	{
	  elink = ELINK(ep);
	  tgtp = &newKarrayTable[array_hashcode1(AVAL(ep)->n,AVAL(ep)->expr,ep->kind,ASHAPE(ep),size)];
	  ELINK(ep) = *tgtp;
	  *tgtp = ep;
	}
    }
  MARKFREE(ArrayTable);
  ArrayTable = newKarrayTable;
  array_hash_size = size;
  return true;
}
