/* Initialization for shared structure allocation.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include "eval.h"
#include "expr.h"
#include "special.h"
#include "cons.h"
#include "conproto.h"
#ifdef HAS_PROTOTYPES
#include "scnsinit.h"
#endif

#define __SEG__ INIT

PUBLIC ExprPtr patany,
#ifdef OBSOLETE
  array0,
#endif
  tuple0,record0;
PUBLIC ExprPtr zzero,one,mone,two,half,
#if FLOAT_FORMAT==IEEE754
	pinf,minf,
#endif
	zpi,izero,ione,itwo;
EXTERN ExprPtr *NaNs;

PUBLIC CHAR *theChars;
PUBLIC ExprPtr voidString;
PUBLIC ExprPtr *charStrings=(ExprPtr *)0L;

PUBLIC unsigned long ktuple2_hash_size = KTUPLE2_HASH_SIZE;
PUBLIC unsigned long kons_hash_size = KONS_HASH_SIZE;
PUBLIC unsigned long mapletkons_hash_size = MAPLETKONS_HASH_SIZE;
PUBLIC unsigned long mapkons_hash_size = MAPKONS_HASH_SIZE;
PUBLIC unsigned long kcomplex_hash_size = KCOMPLEX_HASH_SIZE;
PUBLIC unsigned long ktypeappl_hash_size = KTYPEAPPL_HASH_SIZE;
PUBLIC unsigned long ktypefx_hash_size = KTYPEFX_HASH_SIZE;
PUBLIC unsigned long krecord1_hash_size = KRECORD1_HASH_SIZE;
PUBLIC unsigned long data_hash_size = DATA_HASH_SIZE;
PUBLIC unsigned long int_hash_size = INT_HASH_SIZE;
PUBLIC unsigned long real_hash_size = REAL_HASH_SIZE;
PUBLIC unsigned long string_hash_size = STRING_HASH_SIZE;
PUBLIC unsigned long array_hash_size = ARRAY_HASH_SIZE;

PUBLIC unsigned long gc_hash_cost_offset[2];

PRIVATE void InitPairTable()
{
  COUNTER i;
  ExprPtr *ap;

  Ktuple2Table = (ExprPtr *)SAFEMALLOC(ktuple2_hash_size*sizeof(ExprPtr));
  for (i=0, ap=Ktuple2Table; i<ktuple2_hash_size; i++)
    *ap++ = NOEXPR;
  KonsTable = (ExprPtr *)SAFEMALLOC(kons_hash_size*sizeof(ExprPtr));
  for (i=0, ap=KonsTable; i<kons_hash_size; i++)
    *ap++ = NOEXPR;
  MapletKonsTable = (ExprPtr *)SAFEMALLOC(mapletkons_hash_size*sizeof(ExprPtr));
  for (i=0, ap=MapletKonsTable; i<mapletkons_hash_size; i++)
    *ap++ = NOEXPR;
  MapKonsTable = (ExprPtr *)SAFEMALLOC(mapkons_hash_size*sizeof(ExprPtr));
  for (i=0, ap=MapKonsTable; i<mapkons_hash_size; i++)
    *ap++ = NOEXPR;
  KcomplexTable = (ExprPtr *)SAFEMALLOC(kcomplex_hash_size*sizeof(ExprPtr));
  for (i=0, ap=KcomplexTable; i<kcomplex_hash_size; i++)
    *ap++ = NOEXPR;
  KtypeApplTable = (ExprPtr *)SAFEMALLOC(ktypeappl_hash_size*sizeof(ExprPtr));
  for (i=0, ap=KtypeApplTable; i<ktypeappl_hash_size; i++)
    *ap++ = NOEXPR;
  KtypeFxTable = (ExprPtr *)SAFEMALLOC(ktypefx_hash_size*sizeof(ExprPtr));
  for (i=0, ap=KtypeFxTable; i<ktypefx_hash_size; i++)
    *ap++ = NOEXPR;
  Krecord1Table = (ExprPtr *)SAFEMALLOC(krecord1_hash_size*sizeof(ExprPtr));
  for (i=0, ap=Krecord1Table; i<krecord1_hash_size; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitDataTable()
{
  COUNTER i;
  ExprPtr *ap;

  DataTable = (ExprPtr *)SAFEMALLOC(data_hash_size*sizeof(ExprPtr));
  for (i=0, ap=DataTable; i<data_hash_size; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitIntTable()
{
  COUNTER i;
  ExprPtr *ap;

  IntTable = (ExprPtr *)SAFEMALLOC(int_hash_size*sizeof(ExprPtr));
  for (i=0, ap=IntTable; i<int_hash_size; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitRealTable()
{
  COUNTER i;
  ExprPtr *ap;

  RealTable = (ExprPtr *)SAFEMALLOC(real_hash_size*sizeof(ExprPtr));
  for (i=0, ap=RealTable; i<real_hash_size; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitStringTable()
{
  COUNTER i;
  ExprPtr *ap;
  ExprPtr *ep,e;

  StringTable = (ExprPtr *)SAFEMALLOC(string_hash_size*sizeof(ExprPtr));
  for (i=0, ap=StringTable; i<string_hash_size; i++)
    *ap++ = NOEXPR;
  theChars = (CHAR *)SAFEMALLOC(256L);
  for (i=0; i<256; i++)
    theChars[i] = (CHAR)i;
  voidString = NIL;
  charStrings = (ExprPtr *)SAFEMALLOC(256L*sizeof(ExprPtr));
  for (i=0; i<256; i++)
    charStrings[i] = NIL;
  
  MAKESTRING(0,(CHAR *)"",voidString);
  voidString->mark |= MARK_SHARED;
  for (i=0, ep=charStrings; i<256; i++, ep++)
    {
      MAKESTRING(1,&theChars[i],e);
      e->mark |= MARK_SHARED;
      *ep = e;
    }
}

PRIVATE void InitVarTable()
{
  COUNTER i;
  ExprPtr *ap;

  VarTable = (ExprPtr *)SAFEMALLOC(VAR_HASH_SIZE*sizeof(ExprPtr));
  for (i=0, ap=VarTable; i<VAR_HASH_SIZE; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitRefTable()
{
  COUNTER i;
  ExprPtr *ap;

  RefTable = (ExprPtr *)SAFEMALLOC(REF_HASH_SIZE*sizeof(ExprPtr));
  for (i=0, ap=RefTable; i<REF_HASH_SIZE; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitExtTable()
{
  COUNTER i;
  ExprPtr *ap;

  ExtTable = (ExprPtr *)SAFEMALLOC(EXT_HASH_SIZE*sizeof(ExprPtr));
  for (i=0, ap=ExtTable; i<EXT_HASH_SIZE; i++)
    *ap++ = NOEXPR;
}

PRIVATE void InitArrayTable()
{
  COUNTER i;
  ExprPtr *ap;

  ArrayTable = (ExprPtr *)SAFEMALLOC(array_hash_size*sizeof(ExprPtr));
  for (i=0, ap=ArrayTable; i<array_hash_size; i++)
    *ap++ = NOEXPR;
}

PUBLIC void InitAllocTables()
{
  REAL z;
  int i;

  InitPairTable();
  InitDataTable();
  InitIntTable();
  InitRealTable();
  InitStringTable();
  InitVarTable();
  InitRefTable();
  InitExtTable();
  InitArrayTable();
  gc_hash_cost_offset[0] = (unsigned long)(0.8*(EXTENDED)(ktuple2_hash_size+
							  kons_hash_size+
							  mapletkons_hash_size+
							  mapkons_hash_size+
							  kcomplex_hash_size+
							  ktypeappl_hash_size+
							  ktypefx_hash_size+
							  krecord1_hash_size+
							  int_hash_size
							  +real_hash_size));
  gc_hash_cost_offset[1] = (unsigned long)(0.8*(EXTENDED)(string_hash_size
							  +array_hash_size));
  EALLOC_HOMO(patany);
  patany->kind = (CHAR)E_PAT_ANY;
  patany->mark |= MARK_SHARED;
#ifdef OBSOLETE
  EALLOC_HOMO(array0);
  array0->kind = (CHAR)E_ARRAY_0;
  array0->mark |= MARK_SHARED;
  CGlobals[I_ARRAY0] = array0;
#endif
  EALLOC_HOMO(tuple0);
  tuple0->kind = (CHAR)E_TUPLE_0;
  tuple0->mark |= MARK_SHARED;
  CGlobals[I_TUPLE0] = tuple0;
  EALLOC_HOMO(record0);
  record0->kind = (CHAR)E_RECORD_0;
  record0->mark |= MARK_SHARED;
  CGlobals[I_RECORD0] = record0;
  MAKEREAL(0.0,zzero);
  zzero->mark |= MARK_SHARED;
  CGlobals[I_ZZERO] = zzero;
#if FLOAT_FORMAT==IEEE754
  z = 0.0; /* to prevent Sun cc from complaining that we are dividing
	      by zero in the next lines */
  MAKEREAL(1.0/z,pinf);
  pinf->mark |= MARK_SHARED;
  CGlobals[I_PINF] = pinf;
  MAKEREAL(-1.0/z,minf);
  minf->mark |= MARK_SHARED;
  CGlobals[I_MINF] = minf;
  NaNs = (ExprPtr *)SAFEMALLOC((2*MAX_NAN+1)*sizeof(ExprPtr));
  for (i=-MAX_NAN; i<=MAX_NAN; i++)
    NaNs[i+MAX_NAN] = NIL;
  for (i=-MAX_NAN; i<=MAX_NAN; i++)
    {
      ExprPtr e;

      makeNaND(i,&z);
      MAKEREAL(z,e);
      e->mark |= MARK_SHARED;
      NaNs[i+MAX_NAN] = e;
    }
#endif
  z = 1.0;
  CGlobals[I_ONE] = one = Kreal(&z);
  one->mark |= MARK_SHARED;
  z = -1.0;
  CGlobals[I_MONE] = mone = Kreal(&z);
  mone->mark |= MARK_SHARED;
  z = 2.0;
  CGlobals[I_TWO] = two = Kreal(&z);
  two->mark |= MARK_SHARED;
  z = 0.5;
  CGlobals[I_HALF] = half = Kreal(&z);
  half->mark |= MARK_SHARED;
  z = 3.14159265358979324847905;
  CGlobals[I_ZPI] = zpi = Kreal(&z);
  zpi->mark |= MARK_SHARED;
  CGlobals[I_IZERO] = izero = Kint(0);
  izero->mark |= MARK_SHARED;
  CGlobals[I_IONE] = ione = Kint(1);
  ione->mark |= MARK_SHARED;
  CGlobals[I_ITWO] = itwo = Kint(2);
  itwo->mark |= MARK_SHARED;
}
