/* AIX/RS6000 hackery.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/errno.h>
#else
#include <errno.h>
#endif

/* hack to patch executables on DPX20 machines. Does the following:
   - replace sequences:
start:     81A3000C (l r13, 0xc(r3))
	   7C016800 (cmp cr0,r1,r13)
	   4081001C (ble start+0x24)
     by sequences ending with:
	   4800001C (bl start+0x24)

	   These sequences can be found at the start of the longjmp()
and siglongjmp() codes in the standard libraries. The patch is meant
to suppress the check for a longjmp() to a lower stack.
*/

char blech[] = {
  0x81, 0xA3, 0x00, 0x0C,
  0x7C, 0x01, 0x68, 0x00,
  0x40, 0x81, 0x00, 0x1C
};

int blechlength = 3*4;

char yippee[] = {
  0x81, 0xA3, 0x00, 0x0C,
  0x7C, 0x01, 0x68, 0x00,
  0x48, 0x00, 0x00, 0x1C
};

main(argc,argv)
int argc;
char *argv[];
{
  FILE *in;
  long pos;
  int c;
  char *blechp;
  int blechgot;

  if (argc!=2)
    usage(argv);
  in = fopen(argv[1],"rb+");
  if (in==NULL)
    {
      fprintf(stderr,"%s: cannot open file %s.\n",argv[0],argv[1]);
      exit(1);
    }
  for (pos=0, blechgot=0, blechp=blech; (c=getc(in))!=-1;)
    {
      pos++;
      if (((char) c)==(*blechp))
	{
	  blechp++;
	  if (++blechgot==blechlength)
	    {
	      safefseek(argv,in,pos-blechlength);
	      fprintf(stderr,"Patching %s at position %d.\n",argv[1],pos-blechlength);
	      for (blechgot=0; blechgot<blechlength; blechgot++)
		{
		  putc(yippee[blechgot],in);
		}
	      safefseek(argv,in,pos);
	      blechp = blech;
	      blechgot = 0;
	    }
	}
      else
	{
	  blechp = blech; /* this is correct only if blech cannot be superposed to itself
			     (it is the case here) */
	  blechgot = 0;
	}
    }
  fclose(in);
}

safefseek(argv,in,pos)
char *argv[];
FILE *in;
long pos;
{
  if (fseek(in,pos,0))
    {
      fprintf(stderr,"%s: fseek failed for position %d (errno=%d).\n",argv[0],pos,errno);
      exit(1);
    }
}

usage(argv)
char *argv[];
{
  fprintf(stderr,"Usage: %s <file-to-patch>.\n",argv[0]);
  exit(1);
}
