/* Installation of the HimML primitives in the initial toplevel environment.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "eval.h"
#include "compile.h"
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#endif
#include <time.h>
#ifdef OBSOLETE
/* to include time.h portably is tricky, so we do the following
   (adapted from "Notes on Writing Portable Programs in C", nov. 1990,
   A.Dolenc, A.Lemmke, D.Keppel, G.V.Reilly)
*/
#if OS==MAC || OS==AMIGA_OS
#include <time.h>
#endif
#if OS==UNIX_SYS_V
#include <time.h>
#else
#if OS==UNIX_BSD
#include <sys/time.h>
#endif
#endif
#endif
#include "expr.h"
#include "execinit.h"
#include "ex1proto.h"
#include "exeproto.h"
#include "regproto.h"
#include "exgproto.h"
#include "grmproto.h"
#include "typproto.h"
#include "cmpproto.h"
#include "prtproto.h"
#if TRANSLATOR
#include "bcproto.h"
#endif
#include "type.h"
#ifdef HAS_PROTOTYPES
#include "sexcinit.h"
#endif

#if CODE_STATUS==STATUS_ALPHA
#define STATUS "alpha"
#else
#if CODE_STATUS==STATUS_BETA
#define STATUS "beta"
#else
#if CODE_STATUS==STATUS_GAMMA
#define STATUS "gamma"
#else
#define STATUS "release"
#endif
#endif
#endif

#if OS==UNIX_SYS_V || OS==UNIX_BSD
#define OPT_EXPANDED
#else
#ifdef THINK_C
#undef OPT_EXPANDED
#else
#define OPT_EXPANDED
#endif
#endif

#if OS==AMIGA_OS
#define VERS1(v1,v2,s,r,d) "$VER: himml " #v1 "." #v2 #r " (" d ")\n"
#define VERS(v1,v2,s,r,d) VERS1(v1,v2,s,r,d)
/* this double macro VERS/VERS1 is needed to expand the arguments before
   stringifying them. */
static char *version = VERS(MAJOR_VERSION,MINOR_VERSION,STATUS,REVISION,AMIDATE);
/* for identification by the version command. */
#endif

#ifdef HIMMLX
EXTERN CHAR *funNames[MAX_GLOBALS];
#endif

PUBLIC char *codeStatus = STATUS;

PUBLIC char maintenance[] = MAINTENANCE;
PUBLIC char creator[] = "Jean Goubault-Larrecq";

PUBLIC ExprPtr stdinval,stdoutval,stderrval;

PUBLIC ExprPtr features,featuresType;
PUBLIC ExprPtr numclassttm,optttm;

EXTERN SymPtr betaeq,gama,gamaeq,numalpha,numbeta;
EXTERN ExprPtr num_alpha,num_beta;
#ifdef OLDDATA
EXTERN ExprPtr stringTags;
#endif
EXTERN ExprPtr leftMarginRef,rightMarginRef,maxLinesRef,floatFormatRef;
EXTERN COUNTER theRmargin;
#if TRANSLATOR
PUBLIC ExprPtr maxInlineRef;
PUBLIC INT theInlineLimit=30;
PUBLIC ExprPtr theArgs;
#ifndef HIMMLRUN
EXTERN ExprPtr primCompileMap;
#endif
#endif
EXTERN ExprPtr usePaths;

EXTERN int theArgc,theArgi;
EXTERN char **theArgv;

PRIVATE void makeFeatures P((void));
PRIVATE ExprPtr makeArgs P((void));

PUBLIC ExprPtr _minf,_mnorm,_mden,_zero,_pden,_pnorm,_pinf,none;
EXTERN ExprPtr intType;

EXTERN extDesc outfileExtDesc,infileExtDesc,outstringExtDesc,instringExtDesc,
  iarrayExtDesc,tableExtDesc;
EXTERN extDesc regexpExtDesc;
EXTERN extDesc inprocessExtDesc,outprocessExtDesc;
EXTERN ExprPtr noneSym,someSym;

EXTERN ExprPtr basisEnv;

#define __SEG__ INIT

/* there is no garbage collection during initialization,
	so KEEP() is a noop */
#undef KEEP
#define KEEP(x) x
#undef KONS
#define KONS Kons

PRIVATE ExprPtr KSYMBOL(n)
IN COUNTER n;
{
  ExprPtr sym;

  sym = Ksymbol(message(n));
  return KEEP(sym);
}

PUBLIC void addTypeAction(sym,type,action)
IN CONST SymPtr sym;
IN CONST ExprPtr type;
IN CONST compiledAction action;
{
  addIdType(sym,type);
  addVal(sym,MakeCompiled(E_COMPILED_CLOSURE,action,NIL),NOEXPR);
}

PUBLIC void addFun(n,type,action,name)
IN COUNTER n;
IN CONST ExprPtr type;
IN CONST compiledAction action;
IN CONST CHAR *name;
{
  addTypeAction(KSYMBOL(n),type,action);
#ifdef HIMMLX
  if (name!=NULL)
    {
      VecPtr vp = EVARS(basisEnv);
      COUNTER n = vp->n-1;

      funNames[n] = name;
    }
#endif
}

#ifdef HIMMLRUN
#undef ADDBCFUN
#define ADDBCFUN(n,type,action,bcaction) ADDFUN(n,type,action)
#else

PUBLIC void addBCFun(n,type,action,bcaction,name)
IN COUNTER n;
IN CONST ExprPtr type;
IN CONST compiledAction action,bcaction;
IN CONST CHAR *name;
{
  ExprPtr sym,f;

  sym = KSYMBOL(n);
  addTypeAction(sym,type,action);
#ifdef HIMMLX
  if (name!=NULL)
    {
      VecPtr vp = EVARS(basisEnv);
      COUNTER n = vp->n-1;

      funNames[n] = name;
    }
#endif
  f = MakeCompiled(E_COMPILED_CLOSURE,bcaction,sym);
  sym = Kint(EVARS(basisEnv)->n-1);
  primCompileMap = MapAdd(sym,f,primCompileMap);
}
#endif

PUBLIC ExprPtr addTypeRef(namei,rgnnamei,type,val)
IN COUNTER namei,rgnnamei;
IN CONST ExprPtr type,val;
{
  ExprPtr valref,sym;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  ExprPtr aux,rgnname,f;

  rgnname = KtypeAppl(KSYMBOL(rgnnamei),NIL);
  (void) KEEP(rgnname);
  aux = MakePair(E_ARRAY_2,rgnSym,Singleton(rgnname));
  (void) KEEP(aux);
  f = KtypeFx(refSym,aux);
  (void) KEEP(f);
  type = KtypeAppl(f,KONS(type,NIL));
#else
  type = KtypeAppl(refSym,KONS(type,NIL));
#endif
  (void) KEEP(type);
  sym = KSYMBOL(namei);
  addIdType(sym,type);
  MAKEREF(val,valref);
  addVal(sym,valref,NOEXPR);
  return valref;
}

#ifdef OBSOLETE
PRIVATE ExprPtr Cons(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return z;
}

PRIVATE ExprPtr CONS(x,y)
IN CONST ExprPtr x,y;
{
  ExprPtr z;

  MAKECONS(x,y,z);
  return KEEP(z);
}
#endif

void InitExec()
{
#ifdef OBSOLETE
  ExprPtr nancodesym,nancodeSym,nancodeType,nancodecons;
#endif
  ExprPtr sym,aux,numclassType,listalpha;
  ExprPtr type,strstrlType;
  ENTER;

#if TRANSLATOR
#ifndef HIMMLRUN
  primCompileMap = NewMap();

  {
    ExprPtr f = MakeCompiled(E_COMPILED_CLOSURE,ACTION_NAME(bc_overwrite),overwriteSym);
    ExprPtr i = Kint(ORIGINAL_OVERWRITE);
    primCompileMap = MapAdd(i,f,primCompileMap);
  }
#endif
#endif

  registerExtDesc(&outfileExtDesc);
  registerExtDesc(&infileExtDesc);
  registerExtDesc(&outstringExtDesc);
  registerExtDesc(&instringExtDesc);
  registerExtDesc(&regexpExtDesc);
  registerExtDesc(&inprocessExtDesc);
  registerExtDesc(&outprocessExtDesc);
  registerExtDesc(&iarrayExtDesc);
  registerExtDesc(&tableExtDesc);
  listalpha = listType(alpha);
  strstrlType = coupleType(stringType,listType(stringType));
#ifdef OBSOLETE
  /* datatype NaNCode = NaN0 | NaN1 | NaN2 | NaN3 | NaN4 | NaN5 | NaN6 */
  nan0 = KSYMBOL(242);
  nan1 = KSYMBOL(243);
  nan2 = KSYMBOL(244);
  nan3 = KSYMBOL(245);
  nan4 = KSYMBOL(246);
  nan5 = KSYMBOL(247);
  nan6 = KSYMBOL(248);
  nan7 = KSYMBOL(440);
  setConstructor(nan0);
  setConstructor(nan1);
  setConstructor(nan2);
  setConstructor(nan3);
  setConstructor(nan4);
  setConstructor(nan5);
  setConstructor(nan6);
  setConstructor(nan7);
  nancodesym = KSYMBOL(249);
  nancodeSym = KSYMBOL(250);
  nancodeType = KtypeAppl(nancodeSym,NIL);
  (void) KEEP(nancodeType);
  nancodecons = MapMerge(MapMerge(MapPair(nan0,nancodeType,
					  nan1,nancodeType),
				  MapPair(nan2,nancodeType,
					  nan3,nancodeType)),
			 MapMerge(MapPair(nan4,nancodeType,
					  nan5,nancodeType),
				  MapPair(nan6,nancodeType,
					  nan7,nancodeType)));
  addDataType(nancodesym,NIL,nancodeType,nancodecons);
#ifdef OLDDATA
  addConstructor(nancodeSym,nan0,nan0,NewMap());
  addConstructor(nancodeSym,nan1,nan1,NewMap());
  addConstructor(nancodeSym,nan2,nan2,NewMap());
  addConstructor(nancodeSym,nan3,nan3,NewMap());
  addConstructor(nancodeSym,nan4,nan4,NewMap());
  addConstructor(nancodeSym,nan5,nan5,NewMap());
  addConstructor(nancodeSym,nan6,nan6,NewMap());
  addConstructor(nancodeSym,nan7,nan7,NewMap());
#else
  nanttm = addConstructors(nancodeSym,-8,-1,
			   nan0,nan1,nan2,nan3,nan4,nan5,nan6,nan7);
#endif
  QUIT;
#endif
  {
    ExprPtr numclasscons,numclasssym,numclassSym,
      c_mnan,c_minf,c_mnorm,c_mden,c_zero,c_pden,c_pnorm,c_pinf,c_pnan;

    /* datatype numClass = MNaN of int | MInf | MNormal | MDenormal | Zero |
       PDenormal | PNormal | PInf | PNaN of int */
    c_mnan = KSYMBOL(251);
    c_minf = KSYMBOL(252);
    c_mnorm = KSYMBOL(253);
    c_mden = KSYMBOL(254);
    c_zero = KSYMBOL(255);
    c_pden = KSYMBOL(256);
    c_pnorm = KSYMBOL(257);
    c_pinf = KSYMBOL(258);
    c_pnan = KSYMBOL(259);
    setConstructor(c_mnan);
    setConstructor(c_minf);
    setConstructor(c_mnorm);
    setConstructor(c_mden);
    setConstructor(c_zero);
    setConstructor(c_pden);
    setConstructor(c_pnorm);
    setConstructor(c_pinf);
    setConstructor(c_pnan);
    numclasssym = KSYMBOL(260);
    numclassSym = KSYMBOL(261);
    numclassType = KtypeAppl(numclassSym,NIL);
    (void) KEEP(numclassType);
    numclasscons = MapMerge(MapMerge(MapMerge(MapPair(c_mnan,funType(intType,numclassType),
						      c_minf,numclassType),
					      MapPair(c_mnorm,numclassType,
						      c_mden,numclassType)),
				     MapMerge(MapPair(c_zero,numclassType,
						      c_pden,numclassType),
					      MapPair(c_pnorm,numclassType,
						      c_pinf,numclassType))),
			    MapSingleton(c_pnan,funType(intType,numclassType)));
    addDataType(numclasssym,NIL,numclassType,numclasscons);
#ifdef OLDDATA
    addConstructor(numclassSym,_mnan,_mnan,stringTags);
    addConstructor(numclassSym,_minf,_minf,NewMap());
    addConstructor(numclassSym,_mnorm,_mnorm,NewMap());
    addConstructor(numclassSym,_mden,_mden,NewMap());
    addConstructor(numclassSym,_zero,_zero,NewMap());
    addConstructor(numclassSym,_pden,_pden,NewMap());
    addConstructor(numclassSym,_pnorm,_pnorm,NewMap());
    addConstructor(numclassSym,_pinf,_pinf,NewMap());
    addConstructor(numclassSym,_pnan,_pnan,stringTags);
#else
    numclassttm = addConstructors(numclassSym,-7,1,
				  c_minf,c_mnorm,c_mden,c_zero,
				  c_pden,c_pnorm,c_pinf,c_mnan,c_pnan);
    _minf = Kdata(-7,numclassttm,NIL);
    _mnorm = Kdata(-6,numclassttm,NIL);
    _mden = Kdata(-5,numclassttm,NIL);
    _zero = Kdata(-4,numclassttm,NIL);
    _pden = Kdata(-3,numclassttm,NIL);
    _pnorm = Kdata(-2,numclassttm,NIL);
    _pinf = Kdata(-1,numclassttm,NIL);
#endif
    QUIT;
  }
  {
    ExprPtr optType,optcons;

    /* datatype 'a option = NONE | SOME of 'a */
    setConstructor(noneSym);
    setConstructor(someSym);
    optType = KtypeAppl(optSym,Kons(alpha,NIL));
    (void) KEEP(optType);
    optcons = MapPair(noneSym,optType,
		      someSym,funType(alpha,optType));
    addDataType(optsym,APPLARG(optType),optType,optcons);
#ifdef OLDDATA
    addConstructor(optSym,none,none,NewMap());
    addConstructor(optSym,some,some,NIL);
#else
    optttm = addConstructors(optSym,-1,0,noneSym,someSym);
    none = Kdata(-1,optttm,NIL);
#endif
    QUIT;
  }
  {
    ExprPtr Vd,vd,vdtype,vdcons;

    /* datatype void = VOID of void */
    Vd = KSYMBOL(499);
    vd = KSYMBOL(501);
    setConstructor(Vd);
    vdtype = KtypeAppl(vd,NIL);
    (void) KEEP(vdtype);
    vdcons = MapSingleton(Vd,funType(vdtype,vdtype));
    addDataType(KSYMBOL(500),NIL,vdtype,vdcons);
#ifdef OLDDATA
    addConstructor(vd,Vd,Vd,NIL);
#else
    (void) addConstructors(vd,0,0,Vd);
#endif
    QUIT;
  }
  ADDFUN(262,funType(num_alpha,coupleType(numclassType,numclassType)),classify);
  QUIT;
  type = funType(coupleType(num_alpha,num_alpha),num_alpha);
  ADDFUN(263,type,add);
  ADDFUN(264,type,sub);
  ADDBCFUN(588,type,nummin,bc_rmin);
  ADDBCFUN(589,type,nummax,bc_rmax);
  QUIT;
  aux = KnumType(MapPair(numalpha,one,numbeta,one));
  (void) KEEP(aux);
  ADDFUN(265,funType(coupleType(num_alpha,num_beta),aux),mul);
  QUIT;
  aux = KnumType(MapPair(numalpha,one,numbeta,mone));
  (void) KEEP(aux);
  ADDFUN(266,funType(coupleType(num_alpha,num_beta),aux),div);
  QUIT;
  ADDFUN(267,funType(num_alpha,num_alpha),neg);
  QUIT;
  sym = KSYMBOL(268);
  addIdType(sym,numType);
  addVal(sym,zpi,NOEXPR);
  QUIT;
  addTypeAction(powerSym,funType(coupleType(numType,numType),numType),ACTION_NAME(pow));
  QUIT;
  aux = KnumType(MapSingleton(numalpha,two));
  (void) KEEP(aux);
  ADDFUN(269,funType(num_alpha,aux),sqr);
  QUIT;
  aux = KnumType(MapSingleton(numalpha,half));
  (void) KEEP(aux);
  ADDFUN(270,funType(num_alpha,aux),sqrt);
  QUIT;
  type = funType(num_alpha,num_alpha);
  ADDFUN(271,type,abs);
  ADDFUN(272,type,re);
  ADDFUN(273,type,im);
  ADDFUN(274,type,conj);
  QUIT;
  type = funType(coupleType(num_alpha,num_alpha),boolType);
  ADDBCFUN(278,type,less,bc_rlt);
  ADDBCFUN(279,type,greater,bc_rgt);
  ADDBCFUN(280,type,leq,bc_rle);
  ADDBCFUN(281,type,greq,bc_rge);
  QUIT;
  type = funType(numType,numType);
  ADDFUN(282,type,exp);
  ADDFUN(283,type,exp1);
  ADDFUN(284,type,log);
  ADDFUN(285,type,log1);
  ADDFUN(286,type,sin);
  ADDFUN(287,type,cos);
  ADDFUN(288,type,tan);
  ADDFUN(289,type,asin);
  ADDFUN(290,type,acos);
  ADDFUN(291,type,atan);
  ADDFUN(292,type,sh);
  ADDFUN(293,type,ch);
  ADDFUN(294,type,th);
  ADDFUN(295,type,ash);
  ADDFUN(296,type,ach);
  ADDFUN(297,type,ath);
  ADDFUN(298,type,arg);
  QUIT;
  ADDFUN(299,funType(mapType(alphaeq,beta),setType(alphaeq)),dom);
  QUIT;
  ADDFUN(300,funType(mapType(alphaeq,betaeq),setType(betaeq)),rng);
  QUIT;
  ADDBCFUN(301,funType(coupleType(funType(beta,gama),funType(alpha,beta)),
		       funType(alpha,gama)),compose,bc_compose);
  QUIT;
  ADDFUN(302,funType(coupleType(mapType(betaeq,gama),mapType(alphaeq,betaeq)),
		     mapType(alphaeq,gama)),mapcompose);
  QUIT;
  ADDFUN(303,funType(mapType(alphaeq,betaeq),mapType(betaeq,alphaeq)),inv);
  QUIT;
  ADDFUN(304,funType(listalpha,listalpha),rev);
  QUIT;
  type = listalpha;
  type = funType(coupleType(type,type),type);
  ADDFUN(305,type,rev1);
  ADDFUN(306,type,append);
  QUIT;
  ADDFUN(307,funType(coupleType(stringType,stringType),stringType),concat);
  QUIT;
  type = funType(listType(stringType),stringType);
  ADDFUN(308,type,concatof);
  ADDFUN(309,type,concatof);
  QUIT;
  ADDFUN(310,funType(num_alpha,boolType),real);
  QUIT;
  type = funType(numType,boolType);
  ADDFUN(311,type,integer);
  ADDFUN(312,type,natural);
  QUIT;
  type = funType(numType,numType);
  ADDBCFUN(313,type,floor,bc_floor);
  ADDBCFUN(314,type,ceil,bc_ceil);
  QUIT;
  ADDFUN(315,funType(coupleType(num_alpha,num_alpha),numType),quotient);
  QUIT;
  ADDFUN(316,funType(coupleType(num_alpha,num_alpha),num_alpha),remainder);
  QUIT;
  ADDFUN(317,funType(coupleType(num_alpha,num_alpha),
		     coupleType(numType,num_alpha)),quotrem);
  QUIT;
  type = funType(coupleType(mapType(alphaeq,gama),mapType(alphaeq,beta)),
		 mapType(alphaeq,beta));
  ADDBCFUN(318,type,domrestrto,bc_domrestrto);
  ADDBCFUN(319,type,domrestrby,bc_domrestrby);
  QUIT;
  type = funType(coupleType(mapType(alphaeq,betaeq),mapType(betaeq,gama)),
		 mapType(alphaeq,betaeq));
  ADDFUN(320,type,rngrestrto);
  ADDFUN(321,type,rngrestrby);
  QUIT;
  type = funType(coupleType(setType(alphaeq),setType(alphaeq)),
		 setType(alphaeq));
  ADDBCFUN(322,type,union,bc_overwrite);
  ADDBCFUN(324,type,diff,bc_diff);
  ADDBCFUN(323,type,inter,bc_domrestrto);
  QUIT;
  type = funType(coupleType(mapType(alphaeq,beta),mapType(alphaeq,gama)),
		 boolType);
  ADDFUN(622,type,interp);
  QUIT;
  type = mapType(alphaeq,beta);
  ADDBCFUN(325,funType(coupleType(type,type),type),delta,bc_delta);
  QUIT;
  type = funType(coupleType(mapType(alphaeq,beta),
			    mapType(alphaeq,gama)),
		 boolType);
  ADDFUN(326,type,subset);
  QUIT;
  type = mapType(alphaeq,betaeq);
  type = funType(coupleType(type,type),boolType);
  ADDFUN(327,type,submap);
  QUIT;
  ADDBCFUN(328,funType(coupleType(alphaeq,mapType(alphaeq,beta)),boolType),inset,bc_inset);
  QUIT;
  ADDFUN(329,funType(coupleType(coupleType(alphaeq,betaeq),mapType(alphaeq,betaeq)),
		     boolType),inmap);
  QUIT;
  ADDFUN(330,funType(mapType(alphaeq,beta),intType),card);
  QUIT;
  ADDFUN(331,funType(listalpha,intType),len);
  QUIT;
  ADDFUN(332,funType(listalpha,setType(intType)),inds);
  QUIT;
  ADDFUN(333,funType(listType(alphaeq),setType(alphaeq)),elems);
  QUIT;
  ADDFUN(334,funType(listalpha,mapType(intType,alpha)),listmap);
  QUIT;
  type = coupleType(listalpha,intType);
  ADDFUN(335,funType(type,alpha),nth);
  ADDFUN(336,funType(type,listalpha),nthtail);
  QUIT;
  ADDFUN(337,funType(coupleType(intType,intType),setType(intType)),range);
  QUIT;
  ADDBCFUN(338,funType(stringType,intType),size,bc_size);
  QUIT;
  ADDFUN(339,funType(intType,stringType),chr);
  QUIT;
  ADDBCFUN(340,funType(stringType,intType),ord,bc_ord);
  QUIT;
  ADDBCFUN(341,funType(coupleType(stringType,intType),intType),ordnth,bc_ordof);
  QUIT;
  ADDFUN(342,funType(stringType,listType(stringType)),explode);
  QUIT;
  ADDBCFUN(345,funType(KONS(stringType,coupleType(intType,intType)),stringType),substr,bc_substr);
  QUIT;
  ADDFUN(346,funType(mapType(alphaeq,beta),alphaeq),chooseindom);
  QUIT;
  ADDFUN(347,funType(mapType(alphaeq,beta),beta),chooseinrng);
  QUIT;
  type = mapType(alphaeq,beta);
  type = funType(coupleType(coupleType(alphaeq,beta),type),type);
  ADDFUN(348,type,mapadd);
  ADDFUN(349,type,mapaddunder);
  QUIT;
  type = mapType(alphaeq,beta);
  type = funType(coupleType(alphaeq,type),type);
  ADDFUN(350,type,mapremove);
  QUIT;
  type = mapType(alphaeq,beta);
  type = funType(type,coupleType(type,type));
  ADDFUN(351,type,mapsplit);
  QUIT;
  type = funType(NIL,NIL);
  ADDFUN(353,type,gc);
  QUIT;
  type = funType(stringType,NIL);
  ADDFUN(762,type,gcfull);
  QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  {
    ExprPtr sigma,initsigma,f,atype,fx;
    
    /* array : int * 'a ->[fx] 'a array[rho]
       with fx >= 'a init[rho] */
    f = KtypeFx(arraySym,genRgn);
    (void) KEEP(f);
    atype = KtypeAppl(f,KONS(alpha,NIL));
    (void) KEEP(atype);
    initsigma = MakePair(E_ARRAY_2,effectSym,Singleton(initFx(genRgn,alpha)));
    (void) KEEP(initsigma);
    type = funTypeFx(coupleType(intType,_alpha),atype,initsigma);
#else
    type = KtypeAppl(arraySym,KONS(_alpha,NIL));
    (void) KEEP(type);
    type = funType(coupleType(intType,_alpha),type);
#endif
    sym = arraysym;
    addTypeAction(sym,type,ACTION_NAME(array));
    QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* sub : 'a array[rho] * int ->[fx] 'a with fx >= read[rho] */
    fx = readFx(genRgn);
    sigma = MakePair(E_ARRAY_2,effectSym,fx);
    type = funTypeFx(coupleType(atype,intType),alpha,sigma);
#else
    type = KtypeAppl(arraySym,KONS(alpha,NIL));
    (void) KEEP(type);
    type = funType(coupleType(type,intType),alpha);
#endif
    ADDBCFUN(354,type,aref,bc_aref);
    QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* update : 'a array[rho] * int * 'a ->[fx] unit
       with fx >= write[rho] */
    fx = writeFx(genRgn);
    sigma = MakePair(E_ARRAY_2,effectSym,fx);
    type = funTypeFx(KONS(atype,coupleType(intType,alpha)),NIL,sigma);
#else
    type = KtypeAppl(arraySym,KONS(alpha,NIL));
    (void) KEEP(type);
    type = funType(KONS(type,coupleType(intType,alpha)),NIL);
#endif
    ADDBCFUN(355,type,aset,bc_aset);
    QUIT;
    /* length : 'a array -> int */
    type = KtypeAppl(arraySym,KONS(alpha,NIL));
    (void) KEEP(type);
    type = funType(type,intType);
    ADDFUN(356,type,length);
    QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* arrayoflist : 'a list ->[fx] 'a array[rho] with fx >= 'a init[rho] */
    type = funTypeFx(listalpha,atype,initsigma);
  }
#else
  type = KtypeAppl(arraySym,KONS(_alpha,NIL));
  type = funType(listType(_alpha),type);
#endif
  ADDFUN(357,type,listarray);
  QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  {
    ExprPtr sigma,initsigma,f,atype,fx;
    
    /* iarray : int * int ->[fx] intarray[rho]
       with fx >= int init[rho] */
    f = KtypeFx(iarraySym,genRgn);
    (void) KEEP(f);
    atype = KtypeAppl(f,NIL);
    (void) KEEP(atype);
    initsigma = MakePair(E_ARRAY_2,effectSym,Singleton(initFx(genRgn,intType)));
    (void) KEEP(initsigma);
    type = funTypeFx(coupleType(intType,intType),atype,initsigma);
#else
    type = funType(coupleType(intType,intType),iarrayType);
    ADDFUN(674,type,iarray);
#endif
    QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* isub : intarray[rho] * int ->[fx] int with fx >= read[rho] */
    fx = readFx(genRgn);
    sigma = MakePair(E_ARRAY_2,effectSym,fx);
    type = funTypeFx(coupleType(atype,intType),intType,sigma);
#else
    type = iarrayType;
    type = funType(coupleType(type,intType),intType);
#endif
    ADDBCFUN(675,type,iaref,bc_iaref);
    QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* update : intarray[rho] * int * int ->[fx] unit
       with fx >= write[rho] */
    fx = writeFx(genRgn);
    sigma = MakePair(E_ARRAY_2,effectSym,fx);
    type = funTypeFx(KONS(atype,coupleType(intType,intType)),NIL,sigma);
#else
    type = iarrayType;
    type = funType(KONS(type,coupleType(intType,intType)),NIL);
#endif
    ADDBCFUN(676,type,iaset,bc_iaset);
    QUIT;
    /* ilength : intarray -> int */
    type = funType(iarrayType,intType);
    ADDFUN(677,type,ilength);
    QUIT;
    type = listType(intType);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* arrayoflist : int list ->[fx] intarray[rho] with fx >= int init[rho] */
    type = funTypeFx(type,atype,initsigma);
  }
#else
  type = funType(type,iarrayType);
#endif
  ADDFUN(678,type,ilistarray);
  QUIT;
  type = funType(coupleType(stringType,stringType),boolType);
  ADDFUN(358,type,strless);
  QUIT;
  type = funType(coupleType(alpha,beta),alpha);
  ADDBCFUN(359,type,before,bc_before);
  QUIT;
  type = funType(listalpha,boolType);
  ADDBCFUN(360,type,null,bc_null);
  QUIT;
  type = funType(mapType(alphaeq,beta),boolType);
  ADDBCFUN(361,type,empty,bc_empty);
  QUIT;
  type = funType(NIL,alpha);
  ADDFUN(362,type,abort);
  QUIT;
  type = funType(intType,alpha);
  ADDFUN(363,type,quit);
  QUIT;
  type = MapPair(writesym,funType(stringType,NIL),REST_TAG,alpha);
  type = funType(type,funType(dynType,NIL));
  ADDFUN(435,type,print);
  ADDFUN(436,type,pretty);
  QUIT;
  type = MapPair(writesym,funType(stringType,NIL),flushsym,funType(NIL,NIL));
  sym = KSYMBOL(437);
  addIdType(sym,type);
  {
#define N 2
    field fields[N];

    fields[0].label = writesym;
    fields[0].val = makePrimitive(prtstdout);
    fields[1].label = flushsym;
    fields[1].val = makePrimitive(flushstdout);
    sortFieldArray(N,fields);
    addVal(sym,MakeRecordGnrl(N,fields,fieldsToShape(N,fields)),NOEXPR);
    stdoutval = sym;
#undef N
  }
  sym = KSYMBOL(438);
  addIdType(sym,type);
  {
#define N 2
    field fields[N];

    fields[0].label = writesym;
    fields[0].val = makePrimitive(prtstderr);
    fields[1].label = flushsym;
    fields[1].val = makePrimitive(flushstderr);
    sortFieldArray(N,fields);
    addVal(sym,MakeRecordGnrl(N,fields,fieldsToShape(N,fields)),NOEXPR);
    stderrval = sym;
#undef N
  }
  QUIT;
  type = MapMerge(MapMerge(MapPair(writesym,funType(stringType,NIL),
				   flushsym,funType(NIL,NIL)),
			   MapPair(tellsym,funType(NIL,intType),
				   seeksym,funType(intType,NIL))),
		  MapMerge(MapPair(suspendsym,funType(NIL,NIL),
				   truncatesym,funType(NIL,NIL)),
			   MapPair(advancesym,funType(intType,NIL),
				   seekendsym,funType(intType,NIL))));
  type = funType(stringType,type);
  ADDFUN(439,type,outfile);
  ADDFUN(453,type,appendfile);
  QUIT;
  type = MapPut(killsym,funType(NIL,NIL),
		MapPair(writesym,funType(stringType,NIL),
			flushsym,funType(NIL,NIL)));
  type = funType(strstrlType,type);
  ADDFUN(495,type,outprocess);
  QUIT;
  type = MapMerge(MapMerge(MapPair(writesym,funType(stringType,NIL),
				   convertsym,funType(NIL,stringType)),
			   MapPair(tellsym,funType(NIL,intType),
				   seeksym,funType(intType,NIL))),
		  MapPut(truncatesym,funType(NIL,NIL),
			 MapPair(advancesym,funType(intType,NIL),
				 seekendsym,funType(intType,NIL))));
  type = funType(stringType,type);
  ADDFUN(816,type,outstring);
  QUIT;
  type = MapPair(readsym,funType(intType,stringType),
		 readlinesym,funType(NIL,stringType));
  sym = KSYMBOL(454);
  addIdType(sym,type);
  {
#define N 2
    field fields[N];

    fields[0].label = readsym;
    fields[0].val = makePrimitive(readstdin);
    fields[1].label = readlinesym;
    fields[1].val = makePrimitive(lstdin);
    sortFieldArray(N,fields);
    addVal(sym,MakeRecordGnrl(N,fields,fieldsToShape(N,fields)),NOEXPR);
    stdinval = sym;
#undef N
  }
  QUIT;
  type = MapMerge(MapMerge(MapPair(readsym,funType(intType,stringType),
				   readlinesym,funType(NIL,stringType)),
			   MapPair(tellsym,funType(NIL,intType),
				   seeksym,funType(intType,NIL))),
		  MapPut(suspendsym,funType(NIL,NIL),
			 MapPair(advancesym,funType(intType,NIL),
				 seekendsym,funType(intType,NIL))));
  type = funType(stringType,type);
  ADDFUN(455,type,infile);
  QUIT;
  type =  MapPut(killsym,funType(NIL,NIL),
		 MapPair(readsym,funType(intType,stringType),
			   readlinesym,funType(NIL,stringType)));
  type = funType(strstrlType,type);
  ADDFUN(494,type,inprocess);
  QUIT;
  type =  MapPut(killsym,funType(NIL,NIL),
		 MapMerge(MapPair(writesym,funType(stringType,NIL),
				  flushsym,funType(NIL,NIL)),
			  MapPair(readsym,funType(intType,stringType),
				  readlinesym,funType(NIL,stringType))));
  type = funType(strstrlType,type);
  ADDFUN(496,type,inoutprocess);
  QUIT;
  type = MapMerge(MapMerge(MapPair(readsym,funType(intType,stringType),
				   readlinesym,funType(NIL,stringType)),
			   MapPair(tellsym,funType(NIL,intType),
				   seeksym,funType(intType,NIL))),
		  MapPair(advancesym,funType(intType,NIL),
			  seekendsym,funType(intType,NIL)));
  type = funType(stringType,type);
  ADDFUN(813,type,instring);
  QUIT;
  type = funType(stringType,setType(stringType));
  ADDFUN(527,type,dir);
  ADDFUN(535,type,filebits);
  QUIT;
  type = funType(stringType,NIL);
  ADDFUN(528,type,cd);
  ADDFUN(531,type,unlink);
  ADDFUN(532,type,rmdir);
  ADDFUN(533,type,mkdir);
  QUIT;
  type = funType(coupleType(stringType,stringType),NIL);
  ADDFUN(534,type,rename);
  QUIT;
  type = funType(stringType,intType);
  ADDFUN(93,type,system);
  QUIT;
  type = funType(NIL,stringType);
  ADDFUN(529,type,pwd);
  QUIT;
  type = funType(intType,stringType);
  ADDFUN(530,type,errormsg);
  QUIT;
  type = funType(NIL,numType);
  ADDFUN(465,type,rrand);
  ADDFUN(466,type,zrand);
  QUIT;
  type = funType(NIL,intType);
  ADDFUN(489,type,rand);
  QUIT;
  sym = KSYMBOL(490);
  addIdType(sym,intType);
  addVal(sym,KINT(MIN_SIGNED_LONG),NOEXPR);
  sym = KSYMBOL(491);
  addIdType(sym,intType);
  addVal(sym,KINT(MAX_SIGNED_LONG),NOEXPR);
  sym = KSYMBOL(507);
  addIdType(sym,intType);
  addVal(sym,KINT(NBITS_IN_LONG),NOEXPR);
  QUIT;
  type = funType(NIL,boolType);
  ADDFUN(68,type,maybe);
  QUIT;
  type = funType(NIL,coupleType(timeType,timeType));
  ADDFUN(620,type,time);
  QUIT;
  leftMarginRef = addTypeRef(636,635,intType,izero);
  rightMarginRef = addTypeRef(638,637,intType,KINT(theRmargin));
  maxLinesRef = addTypeRef(640,639,intType,KINT(100));
  floatFormatRef = addTypeRef(517,516,stringType,KSYMBOL(515));
  QUIT;
#if TRANSLATOR
  maxInlineRef = addTypeRef(525,524,intType,KINT(theInlineLimit));
  QUIT;
#endif
  type = listType(stringType);
  usePaths = addTypeRef(519,518,type,NIL);
  QUIT;
#if CONTINUATION_MODEL & (CALLCC|CATCH)
  {
    ExprPtr aux2;
    ExprPtr alphaarg;

    alphaarg = KONS(alpha,NIL);
#if IMPERATIVE_TYPES==EFFECT_TYVAR
    aux2 = KtypeFx(contSym,genRgn);
    (void) KEEP(aux2);
    aux2 = KtypeAppl(aux2,alphaarg);
    (void) KEEP(aux2);
    aux3 = MakePair(E_ARRAY_2,
		    effectSym,
		    Pair(pureGenFx,gotoFx(genRgn)));
    (void) KEEP(aux3);
    aux3 = funTypeFx(alpha,beta,aux3);
    type = funTypeFx(aux2,aux3,pureGenFx);
#else
    aux2 = KtypeAppl(contSym,alphaarg);
    (void) KEEP(aux2);
    type = funType(aux2,funType(alpha,beta));
#endif
  }
  ADDFUN(218,type,throw);
  QUIT;
  {
    ExprPtr aux2;

#if IMPERATIVE_TYPES==EFFECT_TYVAR
    /* callcc : (alpha cont[rho] ->[fx] alpha) ->[fx'] alpha,
       fx' >= fx U exn(rho,alpha) U from(rho) */
    aux2 = KtypeFx(contSym,genRgn);
    (void) KEEP(aux2);
    aux2 = KtypeAppl(aux2,KONS(alpha,NIL));
    (void) KEEP(aux2);
    aux2 = funTypeFx(aux2,alpha,pureGenFx);
    aux3 = MakePair(E_ARRAY_2,effectSym,
		    SetAdd(pureGenFx,
			   Pair(exnFx(genRgn,alpha),
				fromFx(genRgn))));
    (void) KEEP(aux3);
    type = funTypeFx(aux2,alpha,aux3);
#else
    aux2 = KtypeAppl(contSym,KONS(_alpha,NIL));
    (void) KEEP(aux2);
    type = funType(funType(aux2,_alpha),_alpha);
#endif
  }
#endif
#if CONTINUATION_MODEL & CALLCC
  ADDFUN(352,type,callcc);
#endif
#if CONTINUATION_MODEL & CATCH
  ADDFUN(483,type,catch); /* catch has the same type as callcc */
  QUIT;
#endif
  type = funType(numType,intType);
  ADDBCFUN(237,type,int,bc_int);
  QUIT;
  type = funType(intType,numType);
  ADDBCFUN(492,type,num,bc_num);
  QUIT;
  type = funType(intType,intType);
  ADDBCFUN(592,type,iabs,bc_iabs);
  ADDBCFUN(593,type,ineg,bc_ineg);
  ADDBCFUN(594,type,isqr,bc_isqr);
  ADDBCFUN(514,type,inot,bc_inot);
  QUIT;
  type = funType(coupleType(intType,intType),boolType);
  ADDBCFUN(579,type,ilt,bc_ilt);
  ADDBCFUN(581,type,ile,bc_ile);
  ADDBCFUN(580,type,igt,bc_igt);
  ADDBCFUN(582,type,ige,bc_ige);
  QUIT;
  type = funType(coupleType(intType,intType),intType);
  ADDBCFUN(538,type,imin,bc_imin);
  ADDBCFUN(539,type,imax,bc_imax);
  ADDBCFUN(583,type,iplus,bc_iplus);
  ADDBCFUN(584,type,iminus,bc_iminus);
  ADDBCFUN(222,type,itimes,bc_itimes);
  ADDBCFUN(585,type,idiv,bc_idiv);
  ADDBCFUN(586,type,imod,bc_imod);
  ADDBCFUN(508,type,ior,bc_ior);
  ADDBCFUN(509,type,ixor,bc_ixor);
  ADDBCFUN(510,type,iand,bc_iand);
  ADDBCFUN(511,type,ilsl,bc_ilsl);
  ADDBCFUN(512,type,ilsr,bc_ilsr);
  ADDBCFUN(513,type,iasr,bc_iasr);
  QUIT;
  type = coupleType(intType,intType);
  type = funType(type,type);
  ADDFUN(587,type,idivmod);
  QUIT;
#if IMPERATIVE_TYPES==EFFECT_TYVAR
  type = writeFx(genRgn);
  type = MakePair(E_ARRAY_2,effectSym,type);
  type = funTypeFx(KtypeAppl(refSym,Kons(intType,NIL)),NIL,type);
#else
  type = funType(KtypeAppl(refSym,Kons(intType,NIL)),NIL);
#endif
  ADDFUN(650,type,inc);
  ADDFUN(651,type,dec);
  QUIT;
  type = funType(NIL,NIL);
  ADDFUN(808,type,stampworld);
  QUIT;
  type = funType(funType(alpha,beta),funType(listalpha,listType(beta)));
  ADDFUN(826,type,map);
  QUIT;
  type = funType(funType(alpha,beta),funType(listalpha,NIL));
  ADDFUN(827,type,app);
  QUIT;
  /* fold : ('a * 'b -> 'b) -> 'b -> 'a list -> 'b */
  type = funType(funType(coupleType(alpha,beta),beta),
		 funType(beta,funType(listalpha,beta)));
  ADDFUN(828,type,fold);
  ADDFUN(829,type,revfold);
  QUIT;
  type = funType(listalpha,alpha);
  ADDBCFUN(830,type,hd,bc_hd);
  QUIT;
  type = funType(listalpha,listalpha);
  ADDBCFUN(831,type,tl,bc_tl);
  QUIT;
  type = funType(num_alpha,boolType);
  ADDFUN(832,type,denormal);
  ADDFUN(833,type,rotten);
  QUIT;
  type = funType (coupleType (num_alpha, intType), num_alpha);
  ADDFUN(759,type,ldexp);
  QUIT;
  type = funType (numType, coupleType (numType, intType));
  ADDFUN(760,type,frexp);
  QUIT;
  type = funType (numType, coupleType (numType, numType));
  ADDFUN(761,type,modf);
  QUIT;
  type = coupleType(intType,intType);
  type = funType(stringType,
		 MapMerge(MapPair(rematchsym,funType(Kons(stringType,type),
						     KtypeAppl(optSym,Kons(type,NIL))),
				  resubstsym,funType(stringType,stringType)),
			  MapPair(rematchfullsym,funType(stringType,
							 KtypeAppl(optSym,Kons(type,NIL))),
				  rematchpsym,funType(stringType,boolType))));
  ADDFUN(560,type,regcomp);
  QUIT;
  type = funType(stringType,KtypeAppl(optSym,Kons(stringType,NIL)));
  ADDFUN(497,type,getenv);
  QUIT;
  type = funType(intType,stringType);
  ADDFUN(561,type,regerrormsg);
  QUIT;
  type = funType(coupleType(alphaeq,alphaeq),boolType);
  ADDFUN(191,type,system_less);
  QUIT;
  theArgs = makeArgs();
  type = funType(NIL, listType(stringType));
  ADDFUN(920,type,getargs);
  QUIT;
  type = funType(stringType, intType);
  ADDFUN(928,type,atoi);
  QUIT;
  type = funType(stringType, numType);
  ADDFUN(929,type,atof);
  QUIT;
  aux = coupleType(alphaeq,beta);
  {
    ExprPtr aux2,tabletype;

    tabletype = KtypeAppl(tableSym,KONS(_alphaeq,KONS(_beta,NIL)));
    ADDFUN(936,funType(NIL,tabletype),table);
    tabletype = KtypeAppl(tableSym,KONS(alphaeq,KONS(beta,NIL)));
    type = funType(tabletype,funType(alphaeq,KtypeAppl(optSym,Kons(beta,NIL))));
    ADDFUN(939,type,table_get);
    type = funType(tabletype,funType(aux,NIL));
    ADDFUN(940,type,table_put);
    ADDFUN(943,type,table_put_behind);
    type = funType(tabletype,funType(alphaeq,NIL));
    ADDFUN(933,type,table_remove);
    type = funType(tabletype,NIL);
    ADDFUN(941,type,table_kill);
    type = funType(tabletype,funType(funType(aux,boolType),boolType));
    ADDFUN(934,type,table_sweep);
    aux2 = mapType(gamaeq,KSYMBOL(942));
    type = funType(tabletype,funType(funType(aux,aux2),aux2));
    ADDFUN(935,type,table_collect);
    QUIT;
  }
  makeFeatures();
  QUIT;
}

PRIVATE ExprPtr KSTR(s)
IN CONST char *s;
{
  ExprPtr aux;

  aux = Kstr((CHAR *)s);
  return KEEP(aux);
}

PRIVATE ExprPtr K(n)
IN CONST COUNTER n;
{
  return KSTR(message(n));
}

PRIVATE ExprPtr makeArgs()
{
  ExprPtr l;
  CHAR *s;
  /*
  ExprPtr type;
  SymPtr sym;
  */
  int i;

  l = NIL;
  for (i=theArgc; --i>=theArgi; )
    {
      s = (CHAR *)theArgv[i];
#ifndef OPT_EXPANDED
      s++
      (void) getString(NULL,NULL,s,false); /* on non-unix systems, the shell does not expand strings */
#endif
      l = KONS (KSTR ((char *)s), l);
    }
  return l;
  /*
    type = listType(stringType);
    sym = KSYMBOL(920);
    addIdType (sym, type);
    addVal (sym, l, NOEXPR);
  */
}

PRIVATE void makeFeatures()
{
  ExprPtr os,reftyping,cont,num,str,prof,debug,floatformat,ver,recomp,start;
  ExprPtr maint,creat;
  ExprPtr strsettype,sym,modset;
  ExprPtr contset,numset,profset,debugset,vertuple;
  ExprPtr v1,v2,v3,v4;
  INT V,v,r;
  time_t today;
  VecPtr vp;

  os = K(364);
  reftyping = K(365);
  cont = K(366);
  num = K(367);
  str = K(368);
  prof = K(369);
  debug = K(370);
  floatformat = K(371);
  ver = K(372);
  recomp = K(373);
  start = K(374);
  maint = K(375);
  creat = K(376);
  strsettype = setType(stringType);
  featuresType =
	MapMerge(MapMerge(MapMerge(MapPair(os,stringType,
					reftyping,stringType),
				MapPair(cont,strsettype,
					num,strsettype)),
			MapMerge(MapPair(str,strsettype,
					prof,strsettype),
				MapPair(debug,strsettype,
					floatformat,stringType))),
		MapMerge(MapMerge(MapPair(ver,KONS(intType,KONS(intType,KONS(stringType,KONS(intType,NIL)))),
					recomp,stringType),
				MapSingleton(start,stringType)),
			MapPair(maint,stringType,
				creat,stringType)));
  sym = KSYMBOL(377);
  addIdType(sym,featuresType);
  V = MAJOR_VERSION;
  v = MINOR_VERSION;
  r = REVISION;
  today = time(NULL);
  contset = CONTINUATION_MODEL_SET;
  numset = NUMERICAL_MODEL_SET;
  profset = PROFILING_SET;
  debugset = DEBUGGING_SET;
  modset = MODULE_SET;
  v1 = KINT(V);
  v2 = KINT(v);
  v3 = KSTR(STATUS);
  v4 = KINT(r);
  CHECKHETEROFREE();
  scAlloc(*(char **)&vp,VECSIZE(4));
  vp->n = 4;
  vp->expr[0] = v1;
  vp->expr[1] = v2;
  vp->expr[2] = v3;
  vp->expr[3] = v4;
  MAKETUPLE_NOGC(vp,vertuple);
  /*KEEP(vertuple); // KEEP is a noop */
  {
#define N 13
    field fields[N];
    ExprPtr shape;

    fields[0].label = os;
    fields[0].val = KSTR(OS_STRING);
    fields[1].label = reftyping;
    fields[1].val = KSTR(IMPERATIVE_TYPES_STRING);
    fields[2].label = cont;
    fields[2].val = contset;
    fields[3].label = num;
    fields[3].val = numset;
    fields[4].label = str;
    fields[4].val = modset;
    fields[5].label = prof;
    fields[5].val = profset;
    fields[6].label = debug;
    fields[6].val = debugset;
    fields[7].label = floatformat;
    fields[7].val = KSTR(FLOAT_FORMAT_STRING);
    fields[8].label = ver;
    fields[8].val = vertuple;
    fields[9].label = recomp;
    fields[9].val = KSTR(DATE);
    fields[10].label = start;
    fields[10].val = KSTR(ctime(&today));
    fields[11].label = maint;
    fields[11].val = KSTR(maintenance);
    fields[12].label = creat;
    fields[12].val = KSTR(creator);
    sortFieldArray(N,fields);
    shape = fieldsToShape(N,fields);
    fieldsToExprs((ExprPtr *)fields,fields,N); /* a bit hackish, but should work */
    features = KrecordGnrl(shape,N,(ExprPtr *)fields);
#undef N
  }
  addVal(sym,features,NOEXPR);
}
