%{
/* Patch scanner file as returned by flex 2.5.2., so as to produce
   scanner files that can be compiled on the Macintosh. (Cures the
   problem of tables that are >32K by putting them in a resource
   file and reading the latter at initialization time.)
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "system.h"

typedef enum { lngs, ints, chrs } type;
char *mactype[] = {"lngs", "ints", "chrs"};
/* First entry may be patched in case of 16 bit ints; see main() */
char *ctype[] = {"int", "short", "char"};

#define NTABLES 32
#define RES_START 128
#define MAXBUF 80

char name[MAXBUF];
int n = 0;
int size;

struct resource {
  char name[MAXBUF];
  char size[MAXBUF];
  type type;
} resources[NTABLES];

FILE *macf;

%}

SPACE [\ \t]
DIGIT [0-9]
LETTER [A-Za-z_]

%x table
%option noyywrap

%%

<INITIAL>static{SPACE}yyconst	{ /* this is the start of a table */
  BEGIN(table);
}
<INITIAL>\"[^\"\n]*\"	|
<INITIAL>.	|
<INITIAL>\n	{ ECHO; }

<table>short({SPACE}+int)?	{ resources[n].type = ints; }
<table>int			|
<table>long({SPACE}+int)?	{ resources[n].type = lngs; }
<table>(unsigned{SPACE}+)?char	{ resources[n].type = chrs; }

<table>{LETTER}+{DIGIT}*\[{DIGIT}+\]	{
  char *s,*t;

  for (s = resources[n].name, t = yytext; (*s++ = *t++)!='['; );
  s[-1] = '\0';
  for (s = resources[n].size; (*s++ = *t++)!=']'; );
  s[-1] = '\0';
}

<table>=		{ /* time to output the tables! */
  fprintf(stdout,"\n#if OS==MAC\nstatic %s *%s = NULL;\n",
	  ctype[resources[n].type],resources[n].name);
  fprintf(stdout,"#else\nstatic yyconst %s %s[%s] =\n",
	  ctype[resources[n].type],resources[n].name,resources[n].size);
  fprintf(macf,"\nresource '%s' (%d,preload,purgeable) { /* %s[%s] */",
	  mactype[resources[n].type],n+RES_START,resources[n].name,resources[n].size);
}

<table>\}{SPACE};	{ /* this is the end of a table */
  ECHO; fputs ("}\n};\n",macf);
  fputs("\n#endif",stdout);
  n++;
  BEGIN(INITIAL);
}
<table>\"[^\"\n]*\"	|
<table>.	|
<table>\n		{ ECHO; fputs(yytext,macf); }

%%

#ifdef HAS_PROTOTYPES
void outmacinit (void)
#else
void outmacinit()
#endif
{
  int i;

  fputs("\n#if OS==MAC\n",stdout);
  for (i=0; i<n; i++)
    {
      fprintf(stdout,"static Handle %s_H = (Handle)0L;\n",resources[i].name);
    }
  fputs("\nvoid flex_mac_end()\n{\n",stdout);
  for (i=0; i<n; i++)
    {
      fprintf(stdout,"  if (%s_H!=(Handle)0L) HUnlock(%s_H);\n",
	      resources[i].name, resources[i].name);
    }
  fputs("}\n\nvoid flex_mac_init()\n{\n",stdout);
  for (i=0; i<n; i++)
    {
      fprintf(stdout,"  if (%s_H==(Handle)0L || %s==NULL)\n    {\n      %s_H = GetResource('%s',%d);\n",
	      resources[i].name, resources[i].name, resources[i].name,
	      mactype[resources[i].type], i+RES_START);
      fprintf(stdout,"      if (%s_H==(Handle)0L) goto error;\n",resources[i].name);
      fprintf(stdout,"      MoveHHi(%s_H);\n",resources[i].name);
      fprintf(stdout,"      %s = (%s *)*%s_H;\n",
	      resources[i].name, ctype[resources[i].type], resources[i].name);
      fprintf(stdout,"    }\n  HLock(%s_H);\n",resources[i].name);
    }
  fputs("  return;\nerror:\n  flex_mac_end();\n",stdout);
  fputs("  YY_FATAL_ERROR(\"Not enough memory to load scanner tables\");\n}\n",stdout);
  fputs("#endif\n",stdout);
}

main()
{
  macf = fopen("lex.r","w");
  if (macf==NULL)
    {
      fputs("Error: Cannot open lex.r for writing.\n",stderr);
      exit(10);
    }
#if OS==MAC
  if (sizeof(int)==sizeof(short))
    mactype[lngs] = mactype[ints];
  else if (sizeof(int)!=sizeof(long))
    {
      fputs("Sorry, the Mac port needs ints of the same size as shorts or longs.\n",stderr);
      fprintf(stderr,"This is not the case (sizeof(int)=%d, but sizeof(short)=%d and sizeof(long)=%d),\n",sizeof(int),sizeof(short),sizeof(long));
      fputs("the only way to correct this is to hack makemaclex.l.\n",stderr);
      exit(10);
    }
#endif
  fputs("/* Flex tables for the Mac, as resources; this is generated by makemaclex. */\n\n",
	macf);
  fputs("type 'lngs' {\n  wide array { longint; };\n};\n",macf);
  fputs("type 'ints' {\n  wide array { integer; };\n};\n",macf);
  fputs("type 'chrs' {\n  wide array { byte; };\n};\n\n",macf);
  yylex();
  outmacinit();
  fclose(macf);
  exit(0);
}
