/* Homemade memory allocator (used only if SPECIAL_MALLOC is set to 1 in OPTIONS).
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "portable.h"
#include "memproto.h"
#if OS==MAC
#ifndef THINK_C
#include <string.h>
#include <memory.h>
#include <resources.h>
#include <files.h>
#include <dialogs.h>
#endif
#endif
#if OS==UNIX_SYS_V || OS==UNIX_BSD
#include <sys/types.h>
#include <sys/stat.h>
#endif
#if OS==AMIGA_OS
/* nothing */
#endif
#include <fcntl.h>
#ifdef HAS_STRING_H
#include <string.h>
#else
#include <strings.h>
#endif
#include "mem.h"
#ifdef HAS_PROTOTYPES
#include "smem.h"
#endif

#define __SEG__ Memory

#undef DEBUG
#undef STATISTICS

void fastmove(from,to,size)
register long *from,*to;
unsigned long size;
{
  register long *endfrom;

  endfrom = (long *)(((char *)from)+size);
  while (from<endfrom) *to++ = *from++;
}

#if SPECIAL_MALLOC

PUBLIC MZPtr theMZone=(MZPtr)0L;
PUBLIC MBPtr mbFreeList=(MBPtr)0L;

void MZcheck P((void));

#define MBreal(mb) ((mb)!=NULL?(mb):((MBPtr)(((char *)theMZone)+MZHDSIZE+theMZone->MZ_block.MB_size)))

void MBwhere(addr,MB1,MB2)
register char * addr;
MBPtr *MB1,*MB2;
{
  register MBPtr mb,next;

  mb = &theMZone->MZ_block;
  do {
    next=mb->MB_next;
    if (addr<(char *)next || next==NULL)
      {
	*MB1 = mb;
	*MB2 = next;
	return;
      }
    mb = next;
  } while(1);
}

#ifdef STATISTICS
INT nAllocs = 0;
INT nFindLoops = 0;
#endif

void MBfindslot(size,prevMB,freeMB)
register unsigned long size;
MBPtr *prevMB,*freeMB;
{
  register MBPtr MB,prev;

#ifdef STATISTICS
  nAllocs++;
#endif
  prev = &theMZone->MZ_block;
  while (MB=prev->MB_next, MB!=NULL)
    {
#ifdef STATISTICS
      nFindLoops++;
#endif
      if (MB->MB_size>=size) break;
      prev = MB;
    }
  *freeMB = MB;
  *prevMB = prev;
}

void MBreserve(prevMB,freeMB,size)
register MBPtr prevMB,freeMB;
register unsigned long size;
{
  register unsigned long dsize;
  register MBPtr newMB;

  dsize = freeMB->MB_size-size;
  if (dsize!=0)
    { /* split block */
      newMB = (MBPtr)(((char *)freeMB)+size);
      newMB->MB_size = dsize;
      freeMB->MB_size = size;
      newMB->MB_next = freeMB->MB_next;
      prevMB->MB_next = newMB;
    }
  else
    {
      prevMB->MB_next = freeMB->MB_next;
    }
}

#define MBmerge1(zzprevMB,zznextMB) \
	((MBnext(zzprevMB)!=zznextMB)? \
	 (zzprevMB->MB_next = zznextMB, zznextMB): \
	 (zzprevMB->MB_next = zznextMB->MB_next, \
	  zzprevMB->MB_size += zznextMB->MB_size, \
	  zzprevMB))

void MBmerge(MB,prevMB,nextMB)
MBPtr MB,prevMB,nextMB;
{
  MBPtr merged1;

  merged1 = MBmerge1(prevMB,MB);
  if (nextMB!=NULL)
    MBmerge1(merged1,nextMB);
  else merged1->MB_next = NULL;
}

#ifdef HAS_PROTOTYPES
PUBLIC char *MallocDefaultHook(unsigned long size)
#else
PUBLIC char *MallocDefaultHook(size)
unsigned long size;
#endif
{
  return (char *)0L;
}

#ifdef STATISTICS
#define MAX_STAT_SIZE 50
INT nMallocs[MAX_STAT_SIZE] = { 0, };
INT nBigMallocs = 0;
#endif

char * Malloc(size)
unsigned long size;
{
  MBPtr prevMB,freeMB;
  register unsigned long rsize;

#ifdef STATISTICS
  if (size>=MAX_STAT_SIZE)
    nBigMallocs++;
  else nMallocs[size]++;
#endif
  rsize = MBrealsize(size);
  MBfindslot(rsize,&prevMB,&freeMB);
  if (freeMB!=NULL)
    {
      /* we have found a slot large enough */
      MBreserve(prevMB,freeMB,rsize);
      return ((char *)freeMB)+HDSIZE;
      /* from this address on, we now have at least size bytes */
    }
  return((*theMZone->MZ_hook)(size));
}

#ifdef STATISTICS
INT nFrees = 0;
#endif

void Mfree(addr)
char *addr;
{
  MBPtr prevMB,nextMB;

#ifdef STATISTICS
  nFrees++;
#endif
#ifdef DEBUG
  if (((char *)addr)<(char *)&theMZone->MZ_block || ((char *)addr)>=(char *)MBnext(&theMZone->MZ_block))
    {
      berror(message(125));
      CRASH();
    }
#endif
  MBwhere(((char *)addr)-HDSIZE,&prevMB,&nextMB);
  MBmerge((MBPtr)(((char *)addr)-HDSIZE),prevMB,nextMB);
}

void mergeFreeLists(mb1,mb2,prevMB)
register MBPtr mb1,mb2,prevMB;
{
  register MBPtr nextMB;

  if (mb1==(MBPtr)0L)
    {
      (void) MBmerge1(prevMB,mb2);
      return;
    }
  if (mb2==(MBPtr)0L)
    {
      (void) MBmerge1(prevMB,mb1);
      return;
    }
again:
  if (mb1<mb2)
    {
      nextMB = mb1->MB_next;
      prevMB = MBmerge1(prevMB,mb1);
      if (nextMB==(MBPtr)0L)
        {
          (void) MBmerge1(prevMB,mb2);
          return;
        }
      mb1 = nextMB;
      goto again;
    }
  nextMB = mb2->MB_next;
  prevMB = MBmerge1(prevMB,mb2);
  if (nextMB==(MBPtr)0L)
    {
      (void) MBmerge1(prevMB,mb1);
      return;
    }
  mb2 = nextMB;
  goto again;
}

MBPtr sortFreeList(mb)
register MBPtr mb;
{
  MBlock mblock;
  register MBPtr midMB,lastMB,nextMB;

  lastMB = mb;
  if (lastMB==(MBPtr)0L)
    return mb;
  midMB = mb;
  lastMB = lastMB->MB_next;
  if (lastMB==(MBPtr)0L)
    return mb;
  lastMB = lastMB->MB_next;
  while (lastMB!=(MBPtr)0L && (nextMB = lastMB->MB_next)!=(MBPtr)0L)
    { /* problem : because this loop is redone each time sortFreeList
    	is called, we may go through its body up to n.log n times,
    	where n=len mbFreeList, instead of n times if we kept the
    	information (but don't forget that mergeFreeLists coalesces
    	blocks too) */
      lastMB = nextMB->MB_next;
      midMB = midMB->MB_next;
    }
  lastMB = midMB->MB_next;
  midMB->MB_next = (MBPtr)0L;
  mblock.MB_size = 0L;
  mblock.MB_next = (MBPtr)0L;
  mergeFreeLists(sortFreeList(mb),sortFreeList(lastMB),&mblock);
  return mblock.MB_next;
}

void commitFree()
{
  MBPtr mb;

  mb = &theMZone->MZ_block;
  mergeFreeLists(sortFreeList(mbFreeList),mb->MB_next,mb);
  mbFreeList = (MBPtr)0L;
}

char * Mrealloc1(addr,newsize)
char * addr;
unsigned long newsize;
{
#define MAX_LONGS 128
#define MAX_BYTES (MAX_LONGS*sizeof(long))
  static char *temp_prolog=
#if OS==UNIX_SYS_V || OS==UNIX_BSD
    "/tmp/HimML$%d$";
#   define ID getpid()
#   undef RETRY_ID
#else
#if OS==AMIGA_OS
  "T:HimML$%d$";
#   define ID (int)FindTask(NULL)
#   undef RETRY_ID
#else
  "HimML$%d$";
#   define ID (int)addr
  /* this is correct since no two processes on the mac
      may have data at the same address */
#   undef RETRY_ID
#endif
#endif
  static char *temp_epilog = ".tmp";
  register unsigned long oldsize;
  register MBPtr MB;
  int fildes;

  MB = (MBPtr)(addr-HDSIZE);
  oldsize = MB->MB_size-HDSIZE;
  if (oldsize<=MAX_BYTES)
    {
      long buffer[MAX_LONGS];

      fastmove(addr,buffer,oldsize);
      markFree(addr); /* it is faster to then commitFree() after a while */
      addr = Malloc(newsize-HDSIZE);
      fastmove(buffer,addr,oldsize);
    }
  else
    {
      char tmp[32];
      char *s;

      strcpy(tmp,temp_prolog);
      s = tmp+strlen(tmp);
      strcpy(s,ltoasc(ID));
      strcat(s,temp_epilog);
#ifdef RETRY_ID
      while (fildes = open(tmp,O_RDONLY), fildes<0)
	{ /* this code is not safe,
	     because another process may create
	     a file of the same name having the
	     same strategy at the same time */
	  close(fildes);
	  RETRY_ID;
	  strcpy(s,ltoasc(ID));
	  strcat(s,temp_epilog);
	}
#endif
      fildes = open(tmp,O_WRONLY);
      if (fildes>=0)
	{
	  (void) write(fildes,addr,oldsize);
	  close(fildes);
	  markFree(addr); /* it is faster to then commitFree() after a while */
	  addr = Malloc(newsize-HDSIZE);
	  if (addr!=NULL)
	    {
	      fildes = open(tmp,O_RDONLY);
	      (void) read(fildes,addr,oldsize);
	      close(fildes);
	    }
	}
      unlink(tmp);
    }
  return(addr);
}

char * Mrealloc(addr,size)
char * addr;
unsigned long size;
{
  register unsigned long oldsize,newsize,dsize;
  register MBPtr MB,nextMB;
  MBPtr MB1,MB2,newfreeMB,newprevMB;

#ifdef DEBUG
  if (addr<(char *)&theMZone->MZ_block || addr>=(char *)MBnext(&theMZone->MZ_block))
    {
      berror(message(126));
      return(addr);
    }
#endif
  newsize = MBrealsize(size);
  MB = (MBPtr)(addr-HDSIZE);
  oldsize = MB->MB_size;
  if (newsize==oldsize)
    return(addr);
  if (newsize<oldsize)
    { /* shrink block; this one is easy */
      dsize = oldsize-newsize;
      MB->MB_size = newsize;
      nextMB = (MBPtr)(((char *)MB)+newsize);
      nextMB->MB_size = dsize;
      MBwhere((char *)nextMB,&MB1,&MB2);
      MBmerge(nextMB,MB1,MB2);
      return(addr);
    }
  /* expand block : there are several cases to consider */
  MBwhere((char *)MB,&MB1,&MB2);
  dsize = newsize-oldsize;
  if (MB2==MBnext(MB))
    { /* we can hope there is enough room just after MB to accomodate the extra memory */
      if (MB2->MB_size>=dsize)
	{
	  /* ok, we can expand with just the memory following MB */
	  MB->MB_size = newsize;
	  MBreserve(MB1,MB2,dsize);
	  return(addr);
	}
    } /* otherwise, we try to reallocate the block, copy it, then free the old block */
  MBfindslot(newsize,&newprevMB,&newfreeMB);
  if (newfreeMB!=NULL)
    { /* memory is large : we can do that way */
      MBreserve(newprevMB,newfreeMB,newsize);
      fastmove(addr,&newfreeMB->MB_next,oldsize-HDSIZE);
      MBwhere((char *)MB,&MB1,&MB2);
      MBmerge(MB,MB1,MB2);
      return ((char *)newfreeMB)+HDSIZE;
    }
  return(Mrealloc1(addr,newsize));
}

void MBcheck(mb)
MBPtr mb;
{
  MBPtr endMB,beginMB;

  beginMB = &theMZone->MZ_block;
  endMB = MBnext(&theMZone->MZ_block);
  if (mb>=endMB || mb<beginMB)
    {
      berror(message(127),mb);
      return;
    }
}

void MFcheck()
{
  MBPtr MB,endMB,beginMB;

  beginMB = &theMZone->MZ_block;
  endMB = MBnext(&theMZone->MZ_block);
  for (MB = mbFreeList; MB!=NULL; MB=MB->MB_next)
    {
      if (MB>=endMB || MB<beginMB)
	{
	  berror(message(127),MB);
	  return;
	}
    }
}

void MZcheck()
{
  MBPtr prevMB,MB,endMB,beginMB,aMB;

  beginMB = &theMZone->MZ_block;
  prevMB = beginMB;
  endMB = MBnext(&theMZone->MZ_block);
  while (MB = prevMB->MB_next, MB!=NULL)
    {
      if (MB>=endMB || MB<beginMB || MB<=prevMB)
	{
	  berror(message(127),prevMB);
	  return;
	}
      for (aMB=(prevMB!=beginMB?(MBnext(prevMB)):(MBPtr)(((char *)theMZone)+MZHDSIZE));
	   aMB>prevMB && aMB<MB; aMB=MBnext(aMB));
      if (aMB!=MB)
	{
	  berror(message(128),aMB);
	  return;
	}
      prevMB = MB;
    }
  MB = endMB;
  for (aMB=(prevMB!=beginMB?(MBnext(prevMB)):(MBPtr)(((char *)theMZone)+MZHDSIZE));
       aMB>prevMB && aMB<MB; aMB=MBnext(aMB));
  if (aMB!=MB)
    {
      berror(message(128),aMB);
      return;
    }
}

#ifdef OBSOLETE
short MZerror()
{
  MBPtr prevMB,MB,endMB,beginMB,aMB;

  beginMB = &theMZone->MZ_block;
  prevMB = beginMB;
  endMB = MBnext(&theMZone->MZ_block);
  while (MB = prevMB->MB_next, MB!=NULL)
    {
      if (MB>=endMB || MB<beginMB || MB<=prevMB)
	{
	  return(1);
	  /* "MemZone free list corrupted at %lx.",prevMB */
	}
      for (aMB=(prevMB!=beginMB?(MBnext(prevMB)):(MBPtr)(((char *)theMZone)+MZHDSIZE));
	   aMB>prevMB && aMB<MB; aMB=MBnext(aMB));
      if (aMB!=MB)
	{
	  return(-1);
	  /* "MemZone allocation map corrupted at %lx.",aMB */
	}
      prevMB = MB;
    }
  MB = endMB;
  for (aMB=(prevMB!=beginMB?(MBnext(prevMB)):(MBPtr)(((char *)theMZone)+MZHDSIZE));
       aMB>prevMB && aMB<MB; aMB=MBnext(aMB));
  if (aMB!=MB)
    {
      return(-1);
      /* "MemZone allocation map corrupted at %lx.",aMB */
    }
  return(0);
}
#endif

unsigned long MaximumMem()
{
  register MBPtr MB;
  register unsigned long maxsize,size;

  maxsize = 0;
  MB = &theMZone->MZ_block;
  while (MB=MB->MB_next, MB!=NULL)
    {
      size = MB->MB_size-HDSIZE;
      if (size>maxsize)  maxsize = size;
    }
  return(maxsize);
}

unsigned long TotalMem()
{
  register MBPtr MB;
  register unsigned long maxsize,size;

  maxsize = 0;
  MB = &theMZone->MZ_block;
  while (MB=MB->MB_next, MB!=NULL)
    {
      size = MB->MB_size-HDSIZE;
      maxsize += size;
    }
  return(maxsize);
}

unsigned long MemSize()
{
  return(theMZone->MZ_block.MB_size);
}

#if OS==UNIX_SYS_V || OS==UNIX_BSD
extern char *sbrk P((int incr));
extern int brk P((char *addr));
#endif

MZPtr newMZone(size)
unsigned long size;
{
  MZPtr mz;
  unsigned long total;
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  char *limit;
#endif

  size = MBrealsize(size);
  total = size+MZHDSIZE;
#if OS==UNIX_SYS_V || OS==UNIX_BSD
  mz = (MZPtr)sbrk(total);
  if (mz==(MZPtr)-1)
    mz = (MZPtr)0L;
#else
  mz = (MZPtr)NEWPTR(total);
#endif
  if (mz!=(MZPtr)0L)
    {
      mz->MZ_block.MB_size = total;
      mz->MZ_block.MB_next = (MBPtr)(((char *)mz)+MZHDSIZE);
      mz->MZ_block.MB_next->MB_size = size;
      mz->MZ_block.MB_next->MB_next = NULL;
      mz->MZ_hook = MallocDefaultHook;
    }
  return(mz);
}

#if OS==UNIX_SYS_V || OS==UNIX_BSD
MZPtr expandMZone(mz,size) /* assumes MZone ends near brk, and nobody uses malloc */
MZPtr mz;
unsigned long size; /* size should be >= 8 */
{
  MBPtr mb,next;
  unsigned long newsize;
  char *oldlimit,*newlimit;

  size = MODMBSTRIDE(size+(MBSTRIDE-1));
  mb = &mz->MZ_block;
  newsize = mb->MB_size+size;
  if (brk(newlimit = ((char *)mz)+newsize)!=0)
    {
      return (MZPtr)0L;
    }
  oldlimit = ((char *)mz)+mz->MZ_block.MB_size;
  if (mb->MB_next==NULL) /* memory zone completely full */
    {
      mz->MZ_block.MB_size = newsize;
      mz->MZ_block.MB_next = (MBPtr)oldlimit;
      mz->MZ_block.MB_next->MB_size = size;
      mz->MZ_block.MB_next->MB_next = NULL;
    }
  else
    {
      for (; (next = mb->MB_next)!=NULL; mb = next); /* find last free block */
      if (MBnext(mb)==(MBPtr)oldlimit) /* last free block must be expanded */
	mb->MB_size += size;
      else /* create new free block */
	{
	  mb = mb->MB_next = (MBPtr)oldlimit;
	  mb->MB_next = NULL;
	  mb->MB_size = size;
	}
      mz->MZ_block.MB_size = newsize;
    }
  return mz;
}

	/* to ensure nobody uses the system malloc, free and realloc primitives,
	we redefine them here */

static MALLOC_T lastfreed=(MALLOC_T)0L; /* for compatibility with old unix versions */

MALLOC_T malloc(size)
size_t size;
{
  unsigned long ssize;
  VOIDPTR addr;

  if (theMZone==(MZone *)0L) /* if mem zone not yet installed, do sbrk */
    {
      addr = (VOIDPTR)sbrk(size);
      return (addr==(VOIDPTR)-1)?((VOIDPTR)0L):addr;
    }
  addr = (VOIDPTR)Malloc((unsigned long)size);
  if (addr==(VOIDPTR)0L)
  {
    ssize = size;
    if (ssize<INC_SIZE)
      ssize = INC_SIZE;
    if (expandMZone(theMZone,ssize+MBSTRIDE)!=(MZPtr)0L)
      {
	addr = (VOIDPTR)Malloc((unsigned long)size);
      }
  }
  return (MALLOC_T)addr;
}

MALLOC_T calloc(nmemb,size)
size_t nmemb,size;
{
  VOIDPTR addr;
  unsigned long rsize;

  rsize = nmemb*size;
  if (theMZone==(MZone *)0L) /* if mem zone not yet installed, do sbrk */
    {
      addr = (VOIDPTR)sbrk(rsize);
      if (addr==(VOIDPTR)-1)
	addr = (VOIDPTR)0L;
    }
  else
    addr = (VOIDPTR)Malloc(rsize);
  if (addr!=(VOIDPTR)0L)
    {
      (void) memset(addr,0,rsize);
    }
  return (MALLOC_T)addr;
}

FREE_T free(addr)
MALLOC_T addr;
{
  if (theMZone==(MZone *)0L) /* if mem zone not yet installed, don't free */
    return;
  if (addr!=(MALLOC_T)0L && addr!=lastfreed)
    {
      if (lastfreed!=(MALLOC_T)0L)
	{
	  if (((char *)lastfreed)<(char *)&theMZone->MZ_block || ((char *)lastfreed)>=(char *)MBnext(&theMZone->MZ_block))
	    ; /* a block outside theMZone is one we allocated before installing theMZone. */
	  else Mfree((char *)lastfreed);
	}
      lastfreed = addr;
    }
#if FREE_T!=void
  return 0;
#endif
}

MALLOC_T realloc(ptr,size)
MALLOC_T ptr;
size_t size;
{
  if (theMZone==(MZone *)0L) /* if mem zone not installed, reallocate with sbrk */
    {
      VOIDPTR new;

      if (size==0)
	return (VOIDPTR)0L;
      new = (VOIDPTR)sbrk(size);
      if (new==(VOIDPTR)-1)
	new = (VOIDPTR)0L;
      if (ptr!=(MALLOC_T)0L)
	fastmove((long *)ptr,(long *)new,size);
      return (MALLOC_T)new;
    }
  if (ptr==(MALLOC_T)0L)
    return (MALLOC_T)Malloc((unsigned long)size);
  if (size==0)
    {
      free(ptr);
      return (MALLOC_T)0L;
    }
  if (ptr==lastfreed)
    {
      lastfreed = (MALLOC_T)0L;
    }
  return (MALLOC_T)Mrealloc((char *)ptr,(unsigned long)size);
}
#endif

void EndMZone(mz)
MZPtr mz;
{
  DISPOSPTR(mz,mz->MZ_block.MB_size);
}
#endif
