/* Tests whether we have register windows, and how many.
   This is necessary in particular for Sparc architectures,
   on which continuation functions do not work unless
   we first flush the register file to the stack.
   Strangely enough, the following code fails to find
   such information, although its strategy is to
   make the HimML stack code core dump until all
   parameters are right.
   Copyright (C) 1992-2000 Jean Goubault-Larrecq and Bull S.A.
   Copyright (C) 2000-2005 Jean Goubault-Larrecq
   and LSV, CNRS UMR 8643 & INRIA Futurs projet Secsi & ENS Cachan.

   This file is part of HimML.

   HimML is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   HimML is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with HimML; see the file COPYING.  If not, write to
   the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>

#ifdef OBSOLETE
extern char *malloc();
#endif

void garble();
int descendingstack1();

/* test and count register windows;
   defined for sparc-based machines,
   may work on others */

int descendingstack()
{
  char	c;

  return descendingstack1(&c);
}

int descendingstack1(cp)
char *cp;
{
  char	c;

  return (&c)<cp;
}

/* the following is adapted from the code in stack.c */

int ascstack;
	/*  number of bytes from initial top of stack to ML top of stack:   */
#define TOP_STACK_OFFSET	16
	/*  number of bytes allocated by default to the ML stack:   */
#define STACK_SIZE  10000
	/*  number of bytes above bottom of stack below which the stack
	pointer must not lie (checked by Eval()):   */
#define SAFETY_SIZE 500

char	*highMark,*lowMark;

jmp_buf topStackBuf;
jmp_buf botStackBuf;

typedef struct StackChunk   {
  jmp_buf jb;
  char	*lowMark;
  char	*highMark;
  char	contents[1];
}   StackChunk;

StackChunk  *theChunk;
jmp_buf theJmpBuf;
int botData;

	/*  the increment by which we progressively deepen the stack:   */
#define STACK_INC   100
	/*  the number of bytes needed to make safe calls to
	establishBotStack() and longjmp(), counting the space
	needed to save registers:   */
#define SECURITY	80

void	establishBotStack()
{
  char	reserve[STACK_INC];

  reserve[0]  =   0;  /*  to prevent optimizers from freeing the space of the array   */
  if (ascstack)
    {
      if  (reserve+STACK_INC+SECURITY<=lowMark)
	{
	  establishBotStack();	/*  recurse until enough stack depth	*/
	  return;
	}
      lowMark =   reserve+STACK_INC-SAFETY_SIZE;
    }
  else
    {
      if  (reserve-STACK_INC-SECURITY>=lowMark)
	{
	  establishBotStack();	/*  recurse until enough stack depth	*/
	  return;
	}
      lowMark =   reserve+SAFETY_SIZE;
    }
  if  ((botData = setjmp(botStackBuf))==0)
    {
      return; /*  ok  */
    }
  else
    {   /*  don't use any locals, and don't simply return ! */
      /*  the reason is that this portion of code is called in
	  an artificially lowered stack (thus no stack is really
	  present yet); see stackJmp()
	  */
      if (ascstack)
	{
	  memcpy(theChunk->highMark,theChunk->contents,
		 theChunk->lowMark-theChunk->highMark);
	}
      else
	{
	  memcpy(theChunk->lowMark,theChunk->contents,
		 theChunk->highMark-theChunk->lowMark);
	}
      /*  smash stack with contents of the chunk  */
      longjmp(theJmpBuf,botData);
    }
}

void	stackJmp(chunk,jb,data)
StackChunk *chunk;
jmp_buf jb;
int data;
{
  if  (chunk==(StackChunk *)0L)
    longjmp(jb,data);
  theChunk	=   chunk;
  memcpy(theJmpBuf,jb,sizeof(jmp_buf));
  longjmp(botStackBuf,data);
  /*  1) pass control to establishBotStack(), where stack is lowered
      to the bottom of the ML stack (lowMark-SAFETY_SIZE);
      2) there, the stack described by the chunk is copied inside
      the processor stack, and control is passed to the
      jmp_buf jb (which are above lowMark)
      */
}

jmp_buf theReturnBuf;
long theResult;
long (*theFun)();
long theArg;

void	establishStack()
{
  char	reserve[TOP_STACK_OFFSET];  /*  to allocate space on the stack  */

  reserve[0]  =   0;  /*  to prevent optimizers from freeing the space of the array   */
  if (ascstack)
    {
      highMark	=   reserve+TOP_STACK_OFFSET;
      lowMark =   reserve+STACK_SIZE;
    }
  else
    {
      highMark	=   reserve;
      lowMark =   reserve-STACK_SIZE;
    }
  if  (setjmp(topStackBuf)==0)
    {
      establishBotStack();
    }
  else
    {   /*  don't use any locals, and don't simply return ! */
      /*  the reason is that this portion of code is called in
	  an artificially lowered stack (thus no stack is really
	  present yet); see dostk()
	  */
      theResult   =   (*theFun)(theArg);
      longjmp(theReturnBuf,1);
    }
}

long dostk(fun,data)
long (*fun)();
long data;
{
  theFun  =   fun;
  theArg  =   data;
  if  (setjmp(theReturnBuf)==0)
    longjmp(topStackBuf,1);
  /*  1) pass control to establishStack(), where stack is lowered
      to the top ML stack (highMark);
      2) there, the function is applied, and a longjmp(theReturnBuf,1)
      is executed, which returns control here, with the stack
      at its original (above highMark) level.
      */
  return theResult;
}

/* the following code is taken from eval.c */

#define SCSIZE ((long)(((char *)((StackChunk *)0L)->contents)-((char *)(StackChunk *)0L)))
void getStackChunk0(chunkp,nwins)
StackChunk **chunkp;
int nwins;
{
  if (ascstack)
    {
      char	*highmark;

      if (nwins!=0)
	{
	  getStackChunk0(chunkp,nwins-1);
	  return;
	}
      highmark = (char *)chunkp;
      *chunkp = (StackChunk *)malloc(SCSIZE+lowMark-highmark);
      (*chunkp)->lowMark = lowMark;
      (*chunkp)->highMark = highmark;
      memcpy((*chunkp)->contents,highmark,lowMark-highmark);
    }
  else
    {
      char	*lowmark;

      if (nwins!=0)
	{
	  getStackChunk0(chunkp,nwins-1);
	  return;
	}
      lowmark = (char *)chunkp;
      *chunkp = (StackChunk *)malloc(SCSIZE+highMark-lowmark);
      (*chunkp)->lowMark = lowmark;
      (*chunkp)->highMark = highMark;
      memcpy((*chunkp)->contents,lowmark,highMark-lowmark);
    }
}

int theNWins;

StackChunk  *getStackChunk()
{
  StackChunk  *chunk;

  if (theNWins!=0)
    getStackChunk0(&chunk,theNWins);
  else if (ascstack)
    {
      char	*highmark;

      highmark = (char *)&chunk;
      chunk = (StackChunk *)malloc(SCSIZE+lowMark-highmark);
      chunk->lowMark = lowMark;
      chunk->highMark = highmark;
      memcpy(chunk->contents,highmark,lowMark-highmark);
    }
  else
    {
      char	*lowmark;

      lowmark = (char *)&chunk;
      chunk = (StackChunk *)malloc(SCSIZE+highMark-lowmark);
      chunk->lowMark = lowmark;
      chunk->highMark = highMark;
      memcpy(chunk->contents,lowmark,highMark-lowmark);
    }
  return chunk;
}

long theValue;

void throw(chunk,val)
StackChunk *chunk;
long val;
{
  theValue = val;
  stackJmp(chunk,chunk->jb,1);
}

long callcc(fun,data)
long (*fun)(/*StackChunk *chunk,char *data*/);
char *data;
{
  StackChunk  *chunk;
  long value;

  chunk   =   getStackChunk();
  if  (setjmp(chunk->jb)==0)
    {
      value = (*fun)(chunk,data);
      /*  normal return   */
    }
  else
    {   /*  continuation has been called	*/
      value   =   theValue;
    }
  return value;
}

jmp_buf albertBuf;

/* the following code was submitted to then IOCCC 92 contest,
   and won the most useful program award; it is used to garble
   the stack, so that we're sure that we dump core if not enough
   register windows are saved. */

/*
Most Useful Program: <albert%bsovax@sun4nl.nluug.nl> Albert van der Horst

	Albert van der Horst
	S P&A R&C
	Oranjestraat 8
	3511 RA   UTRECHT
	The Netherlands


Judges' comments:

    For a canned example:
	make albert
	albert 1234567890123456789
    
    We were impressed with the speed at which it was able to factor
    arbitrarily large numbers consisting of factors that fit into
    a long.


Selected notes from the author:

    The Obfuscated version of the Horst algorithm.

    This program will factor unlimited length numbers and print the 
    factors in ascending order. Numbers of one digit (e.g. 8) 
    are rejected without notice.
    It quits as soon as there is at most one
    factor left, but that factor will not be shown. 
    It accomplishes this efficiently, without resorting to division
    or multiplication, until the candidate factor no longer fits in 
    a signed long. 

    The nicest way is to rename the program into e.g. 4294967297
    if you want to factor Fermat's 4th number. Then just run it.
    Or you may type "prog <some-number>"
    A nice one is also (30 ones)
    albert 111111111111111111111111111111

    Apart from the foregoing there are no special execution instructions.

    To customize the program into a factorizer of a fixed number, use
    cc albert.c -o 4294967297
    or some such.

    There are no data files used, and it is not possible to feed input
    via stdin.

    I think this program is a nice example of algorithmic obfuscation.
    Several times two similar algorithms are merged into one. Then you 
    need quite some tricks to get it running, such as long jumps 
    through recursive subroutines. I felt like a sword smith, 
    welding the sword very long, folding it again to the proper 
    length, keep on hammering till it is tight, then fold again.
    Always keeping it at the proper red hot temperature, but not too
    hot lest the hardness fades.
    The strict naming conventions for subroutines did not make things 
    much clearer after all, but it was not supposed to.

    I would like to draw attention to the robustness of the program
    with respect to error handling and the nice stopping criterion.
    The esthetic appeal of some lines is at the expense of clearness, 
    I apologize.
    Running the program through the c-beautifier reveals nothing,
    it will only destroy some of the lay out.
    Running the program through lint shows the usual remarks for a
    K&R program. Defeating this through casts does not make a program 
    cleaner in my opinion.

    Here are some hints, but they may not be too helpful.
    1. The Horst algorithm is described in the Hobby Computer Club 
	Newsletter, year 82, part 4, a real Dutch treat.
       Assembler, c and Forth version have been around for some years.
    2. Does fractal programming exist after all?
    3. You remember the ToomCook algorithm in Knuth?
       It uses iteration instead of recursion and is quite jumpy.
       This program shares these disadvantages in a modified form.
    4. The Conversion is to be found in Knuth, not so the Observation.
       The Observation: "if it ends in a zero, it is divisible by ten"
*/

#define new(PP) (PP *) malloc((long)sizeof(PP)) 

typedef struct q {
    jmp_buf ppp;
    long qq;
    struct q *P;
    struct q *p;
} 
PP;

PP *P;
int aaaaaa=2;
int aaaaaaa=1;

long qqq;


void aaAaaa(aa,aaa)
char *aa;
char *aaa;
{
    char aaaa = 0;

    &aaaa; &aa;
    if ((((( aaa )))))
    {
        aaaa = *aa;
        *aa=0;
        aa+=strlen(aa+1);
        P =new(PP);
        P->P=P;
        P->p=P;
    }

    if ((((( !setjmp(P->ppp) ))))) 
        {
        if ((((( !isdigit(*aa) )))))
            longjmp(P->ppp,aaaaaaa);
        else {
            P->p->P = new(PP);
            P->p->P->P = P;
            P->p->P->p = P->p;
            P->p = P->p->P;

            P->qq = *aa--;
            P = P->p;
            aaAaaa(aa,0);
        }
    } else {
        if ( !aaaa&&!*aa ) 
           longjmp(P->p->ppp,aaaaaaa);

        if ((((( ((P->qq=aaaa)<10     &&!
		  (isdigit(aaaa)))     ||!
                 (isdigit(*aa)       ||!
                 *aa                        )))))) 
        {
            fprintf(stderr,"Usage %c%s <number>\n",
            (aaa[0]?7:aaaa),aaa+!aaa[0]);
            longjmp(albertBuf,1);
        }
    }
}


void ppPppp(pp,ppp)
PP **pp, *ppp;
{
    int aa;

    &ppp;
    if ((((( !(aa=setjmp(ppp->ppp))||aa==aaaaaa )))))
    {
        if ((((( *pp==ppp )))))
        {
            ppp = (*pp)->p;

            if ( qqq<47 ) return;
            if ( ppp->qq!=48 ) return;

            while ( ppp->qq==48 ) 
            {
                fprintf(stderr,"%ld\n",qqq-45);
                *pp = ppp;
                ppp = ppp->p;
            }
            qqq -= 1;
            longjmp(ppp->ppp,aaaaaaa);
        } else {
            PP *p;

            ppPppp(pp,ppp->p);
            for (p=ppp;p!=*pp;p=p->p)
            {
                int qq=4;
                if ((((( (qqq<47                            &&
                         (qq=0,p->qq+=p->p->qq-96)>=48-qqq) ||
                         (qqq>46                            &&
                         (p->qq-=p->p->qq)<0)                   ))))) 
                {
                    p->qq += qqq+qq;
                    if ( p->p==P && qqq<=46 )
                    {
                        P->p->P = new(PP);
                        P->p->P->P = P;
                        P->p->P->p = P->p;
                        *pp = P = P->p = P->p->P;
                        P->qq = 48;
                    }

                    p->p->qq+=qq==0;
                    p->p->qq-=qq!=0;
                }
                else
                {
                    p->qq += 48;
                }
            }
            if ( ppp->P==P ) longjmp(ppp->ppp,aaaaaaa);
        }
    }
    else
    {
        qqq += 1; 

        while (48==P->qq )
        {
            P->P->p = P->p;
            P = P->p->P = P->P;

        }

        if ( ppp!=ppp->p->p || qqq<47 )
            longjmp(ppp->ppp,aaaaaa);
        else
        {
            fprintf(stderr,"At most one remains\n");
            longjmp(albertBuf,0);
        }
    }
}

void factor(aaa,aaaa)
int aaa;
char **aaaa;
{
    aaAaaa(aaa==aaaaaaa?aaaa[0]:aaaa[1],aaaa[0]);
    qqq = 39;
    ppPppp(&P,P->p);
}


void garble()
{
  char *argv[2];
  int argc=2;

  argv[0] = "albert";
  argv[1] = malloc(100);
  strcpy(argv[1],"1234567890123456789");
  fprintf(stderr,"factorizing %s.\n",argv[1]);
  if (setjmp(albertBuf)==0)
    factor(argc,argv);
}

StackChunk *cc;
jmp_buf topBuf;
long theTopData;

long testfun(chunk,data)
StackChunk *chunk;
long *data;
{
  garble();
  cc = chunk;
  theTopData = *data;
  longjmp(topBuf,1);
}

long test(nreccalls)
int nreccalls;
/* the test is a callcc (fn k => (cc:=k; throw k 1)),
   followed by a throw (!cc) 2 */
{
  long a;

  if (nreccalls!=0)
    {
      a = test(nreccalls-1);
    }
  else
    {
      a = 1L;
      a = callcc(testfun,&a);
    }
  return a;
}

jmp_buf fail_buf;

void fail(n)
int n;
{
  /* something went wrong, we assume this is because the minimum
     number of register windows was not guessed correcly */
  fprintf(stderr,"fail with code %d.\n",n);
  longjmp(fail_buf,1);
}

void fatal(msg,a,b,c,d)
char *msg;
long	a,b,c,d;
{
  fprintf(stderr,msg,a,b,c,d);
  fputc('\n',stderr);
  exit(10);
}

long wholetest(nreccalls)
long nreccalls;
{
  long a;

  if (setjmp(topBuf)==0)
    {
      a = test((int)nreccalls);
      fprintf(stderr,"test() returns\n");
      if (a!=2L)
	fatal("stack garbled (2)");
      return a;
    }
  fprintf(stderr,"caught exception\n");
  if (theTopData!=1L)
    fatal("stack garbled");
  garble();
  throw(cc,2L);
  fatal("throw returns");
  return 0L;
}

int main(argc,argv) /* tstwin <nreccalls> <nwins> */
int argc;
char **argv;
{
  int nreccalls;

  if (setjmp(fail_buf)!=0)
    exit(1); /* something went wrong */
  signal(SIGILL,fail);
  signal(SIGBUS,fail);
  signal(SIGSEGV,fail);
  ascstack = !descendingstack();
  if (argc!=3)
    fatal("wrong number of arguments");
  nreccalls = atoi(argv[1]);
  theNWins = atoi(argv[2]);
  establishStack();
  (void) dostk(wholetest,(long)nreccalls);
  /* test was passed */
  exit(0);
}
